////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     GFUN Module Implementation
///
///     This file implements the GFUN module. This module talks to the GFUN on
///     the blade. This module contains subroutines that set up any memory-
///     mapped registers and descriptor queues that are needed for
///     communication between the Host OS and the GFUN over the PCI bus.
///
///     Descriptor queues and the buffers in which the descriptor queue entries
///     point to are created in here. However, the contents of the buffers are
///     not within the scope of this file.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_gfun.h"
#include "tera_gfun_internal.h"
#include "tera_gfun_reg.h"

// ---- Private Constants and Types -----------------------------------------

// --------------------------------------------------------------------------
//  eREG_OP
///     Register operations
//
//  Description:
///     This enumeration defines register operations. An AND operation is used
///     for masking register values, and an OR operation is used to set values.
///
typedef enum
{
    REG_OP_AND = 0,    ///< Bit-wise AND operation
    REG_OP_OR          ///< Bit-wise OR operation
} eREG_OP;

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

static UINT32 reg_read(
    UINT8 *reg_base,
    UINT32 offset
    );

static void reg_write(
    UINT8 *reg_base,
    UINT32 offset,
    UINT32 value
    );

static void reg_read_write(
    UINT8  *reg_base,
    UINT32  offset,
    UINT32  value,
    eREG_OP op
    );

static sGFUN_DMA_DESC *offset_to_dma_desc(
    sTERA_MEM_DMA *desc_q,
    UINT32         offset
    );

static INT32 desc_q_create(
    sTERA_GFUN_CBLK *cblk,
    UINT32          chan
    );

static void desc_q_destroy(
    sTERA_GFUN_CBLK *cblk,
    UINT32          chan
    );

static void gfun_desc_q_info_write(
    sTERA_GFUN_CBLK *cblk,
    UINT32           chan
    );

static INT32 gfun_device_hw_id_read(
    struct  pci_dev         *dev,
    eTERA_PCOIP_DEVICE_TYPE *dev_type
    );

// ---- Local Functions -----------------------------------------------------

// --------------------------------------------------------------------------
//  reg_read
//      Read a value from one of the memory-mapped registers
//
//  Description:
//      This function reads a value from one of the memory-mapped registers.
//      The value returned is a 32-bit value.
//
static UINT32 reg_read(
    UINT8 *reg_base,      // Registry base
    UINT32 offset         // Offset from base of registry
    )
{
    return ioread32((UINT32 *)(reg_base + offset));
} // reg_read()

// --------------------------------------------------------------------------
//  reg_write
//      Write a value to one of the memory-mapped registers
//
//  Description:
//      This function writes to one of the memory-mapped registers. The value
//      written is 32-bits wide. This function does not do a read-write-modify;
//      it is up to the user to do this.
//
static void reg_write(
    UINT8 *reg_base,      // Registry base
    UINT32 offset,        // Offset from base of registry
    UINT32 value          // Value to write
    )
{
    iowrite32(value, (UINT32 *)(reg_base + offset));
} // reg_write()

// --------------------------------------------------------------------------
//  reg_read_write
//      Perform a read-modify-write to one of the memory-mapped registers
//
//  Description:
//      This function performs a read-modify-write to one of the memory-
//      mapped registers. The value to be written is 32-bits wide.
//
static void reg_read_write(
    UINT8  *reg_base,     // Registry base
    UINT32  offset,       // Offset from base of registry
    UINT32  value,        // Value to write
    eREG_OP op            // Bitwise operator
    )
{
    switch (op)
    {
        case REG_OP_AND:
        {
            reg_write(reg_base, offset, (reg_read(reg_base, offset) & value));
            break;
        }
        case REG_OP_OR:
        {
            reg_write(reg_base, offset, (reg_read(reg_base, offset) | value));
            break;
        }
    }
} // reg_read_write()

// --------------------------------------------------------------------------
//  func_rdy_read
//      Read the value of the FUNC_RDY field
//
//  Description:
//      This function reads the value of the FUNC_RDY field from the memory-
//      mapped registers. This function masks out the field so other values
//      in the register is not returned.
//
static UINT32 func_rdy_read(
    sTERA_GFUN_CBLK *cblk        // Control block
    )
{
    return reg_read(cblk->reg_base, GFUN_REG_GLOB_FUNCSTAT) &
                    GFUN_REG_GLOB_FUNCSTAT_FUNC_RDY_MASK;
} // func_rdy_read()

// --------------------------------------------------------------------------
//  ingress_head_read
//      Read the value of the ingress descriptor queue head pointer
//
//  Description:
//      This function reads the value of the ingress descriptor queue head
//      pointer from the memory-mapped registers.
//
static UINT32 ingress_head_read(
    sTERA_GFUN_CBLK *cblk    // Control block
    )
{
    return (reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRHEADPTR(0, cblk->dev_type)) &
            GFUN_REG_CH_INDESCRHEADPTR_DESCR_HEAD_MASK);
} // ingress_head_read()

// --------------------------------------------------------------------------
//  ingress_head_write
//      Write the value to the ingress descriptor queue head pointer
//
//  Description:
//      This function write the value to the ingress descriptor queue head
//      pointer in the memory-mapped registers.
//
static void ingress_head_write(
    sTERA_GFUN_CBLK *cblk,    // Control block
    UINT32           val      // Value to write
    )
{
    reg_write(cblk->reg_base, GFUN_REG_CH_INDESCRHEADPTR(0, cblk->dev_type),
              (val & GFUN_REG_CH_INDESCRHEADPTR_DESCR_HEAD_MASK));
} // ingress_head_write()

// --------------------------------------------------------------------------
//  ingress_tail_read
//      Read the value of the ingress descriptor queue tail pointer
//
//  Description:
//      This function reads the value of the ingress descriptor queue tail
//      pointer from the memory-mapped registers.
//
static UINT32 ingress_tail_read(
    sTERA_GFUN_CBLK *cblk    // Control block
    )
{
    return (reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRTAILPTR(0, cblk->dev_type)) &
            GFUN_REG_CH_INDESCRTAILPTR_DESCR_TAIL_MASK);
} // ingress_tail_read()

// --------------------------------------------------------------------------
//  ingress_tail_write
//      Write the value to the ingress descriptor queue tail pointer
//
//  Description:
//      This function write the value to the ingress descriptor queue tail
//      pointer in the memory-mapped registers.
//
static void ingress_tail_write(
    sTERA_GFUN_CBLK *cblk,    // Control block
    UINT32           val      // Value to write
    )
{
    reg_write(cblk->reg_base, GFUN_REG_CH_INDESCRTAILPTR(0, cblk->dev_type),
              (val & GFUN_REG_CH_INDESCRTAILPTR_DESCR_TAIL_MASK));
} // ingress_tail_write()

// --------------------------------------------------------------------------
//  egress_head_read
//      Read the value of the egress descriptor queue head pointer
//
//  Description:
//      This function reads the value of the egress descriptor queue head
//      pointer from the memory-mapped registers.
//
static UINT32 egress_head_read(
    sTERA_GFUN_CBLK *cblk    // Control block
    )
{
    return (reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRHEADPTR(0)) &
            GFUN_REG_CH_OUTDESCRHEADPTR_DESCR_HEAD_MASK);
} // egress_head_read()

// --------------------------------------------------------------------------
//  egress_head_write
//      Write the value to the egress descriptor queue head pointer
//
//  Description:
//      This function write the value to the egress descriptor queue head
//      pointer in the memory-mapped registers.
//
static void egress_head_write(
    sTERA_GFUN_CBLK *cblk,    // Control block
    UINT32           val      // Value to write
    )
{
    reg_write(cblk->reg_base, GFUN_REG_CH_OUTDESCRHEADPTR(0),
              (val & GFUN_REG_CH_OUTDESCRHEADPTR_DESCR_HEAD_MASK));
} // egress_head_write()

// --------------------------------------------------------------------------
//  egress_tail_read
//      Read the value of the egress descriptor queue tail pointer
//
//  Description:
//      This function reads the value of the egress descriptor queue tail
//      pointer from the memory-mapped registers.
//
static UINT32 egress_tail_read(
    sTERA_GFUN_CBLK *cblk    // Control block
    )
{
    return (reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRTAILPTR(0)) &
            GFUN_REG_CH_OUTDESCRTAILPTR_DESCR_TAIL_MASK);
} // egress_tail_read()

// --------------------------------------------------------------------------
//  egress_tail_write
//      Write the value to the egress descriptor queue tail pointer
//
//  Description:
//      This function write the value to the egress descriptor queue tail
//      pointer in the memory-mapped registers.
//
static void egress_tail_write(
    sTERA_GFUN_CBLK *cblk,    // Control block
    UINT32           val      // Value to write
    )
{
    reg_write(cblk->reg_base, GFUN_REG_CH_OUTDESCRTAILPTR(0),
              (val & GFUN_REG_CH_OUTDESCRTAILPTR_DESCR_TAIL_MASK));
} // egress_tail_write()

// --------------------------------------------------------------------------
//  offset_to_dma_desc
//      Returns a pointer to the DMA descriptor for a given offset
//
//  Description:
//      This function returns a pointer to the DMA descriptor in the queue at
//      the specified offset. The offset is in bytes.
//
static sGFUN_DMA_DESC *offset_to_dma_desc(
    sTERA_MEM_DMA *desc_q,
    UINT32         offset
    )
{
    UINT8 *buf = desc_q->virt_addr;

    return (sGFUN_DMA_DESC *)&buf[offset];
} // offset_to_dma_desc()

// --------------------------------------------------------------------------
//  desc_q_create
//      Create the ingress and egress descriptor queues
//
//  Description:
//      This function creates common buffers which are used for the ingress
//      and egress descriptor queues. The data buffers for the ingress queue
//      will also be populated. The size of the descriptor buffers is read
//      from the GFUN mailbox registers. If the queues are created successfully,
//      the details of the queues are written into the appropriate GFUN
//      registers, and this function returns TERA_SUCCESS.
//
static INT32 desc_q_create(
    sTERA_GFUN_CBLK *cblk,      // Control block
    UINT32           chan       // GFUN channel
    )
{
    INT32           ret = TERA_SUCCESS;
    UINT32          i;
    sGFUN_DMA_DESC *desc;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);
    ASSERT(chan < GFUN_NUM_DMA_CHAN);

    if (!cblk->initialized)
    {
        mTERA_LOG_ERR("invalid state");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        cblk->mtu_size[chan] = 1400;

        // Allocate a common buffer for the ingress descriptor queue
        cblk->ingress_desc_q[chan] = tera_mem_alloc(cblk->dev,
                                                    INGRESS_DESC_Q_LEN *
                                                    sizeof(sGFUN_DMA_DESC));
        if (cblk->ingress_desc_q[chan] == NULL)
        {
            mTERA_LOG_ERR("failed to allocate common buffer for ingress queue");

            ret = TERA_ERR_NO_MEMORY;
        }
        else
        {
            // Allocate a common buffer for the egress descriptor queue
            cblk->egress_desc_q[chan] = tera_mem_alloc(cblk->dev,
                                                       EGRESS_DESC_Q_LEN *
                                                       sizeof(sGFUN_DMA_DESC));
            if (cblk->egress_desc_q[chan] == NULL)
            {
                mTERA_LOG_ERR("failed to allocate common buffer for egress queue");

                ret = TERA_ERR_NO_MEMORY;
            }
            else
            {
                //
                // Populate the ingress queue with data buffers. We need to do this
                // because the GFUN cannot allocate host OS memory on the fly. We
                // provide empty buffers for the GFUN to use if it needs to when
                // sending messages to the host OS.
                //
                for (i = 0; i < INGRESS_DESC_Q_LEN; i++)
                {
                    cblk->ingress_desc_q_buf[chan][i] = tera_mem_buf_alloc(cblk->dev,
                                                                           cblk->mtu_size[chan]);
                    if (cblk->ingress_desc_q_buf[chan][i] == NULL)
                    {
                        mTERA_LOG_ERR("failed to allocate buffer for ingress descr queue");
                        ret = TERA_ERR_NO_MEMORY;
                        break;
                    }

                    // Fill out the buffer address in the DMA descriptor
                    desc = offset_to_dma_desc(cblk->ingress_desc_q[chan],
                                              i * sizeof(sGFUN_DMA_DESC));
                    desc->buf_addr = cblk->ingress_desc_q_buf[chan][i]->phys_addr;
                    desc->buf_len = cblk->mtu_size[chan] & GFUN_DMA_DESC_BUF_LEN_MASK;
                }

                if (ret == TERA_SUCCESS)
                {
                    //
                    // All the descriptor queues have been created and populated- now write
                    // the details of the queues into the GFUN registers
                    //
                    gfun_desc_q_info_write(cblk, chan);
                }
            }
        }
    }

    if (ret != TERA_SUCCESS)
    {
        // Do cleanup if anything failed
        desc_q_destroy(cblk, chan);
    }

    mTERA_LOG_DBG("end");

    return ret;
} // desc_q_create()

// --------------------------------------------------------------------------
//  desc_q_destroy
//      Destroy the ingress and egress descriptor queues
//
//  Description:
//      This function frees common buffers that are used for the ingress and
//      egress descriptor queues. Any data buffers attached to the queues will
//      also be freed.
//
static void desc_q_destroy(
    sTERA_GFUN_CBLK *cblk,      // Control block
    UINT32           chan       // GFUN channel
    )
{
    UINT32 i;

    if ((cblk != NULL) && cblk->initialized)
    {
        // Free the common buffers for inbound and outbound descriptor queues
        tera_mem_free(cblk->dev, cblk->ingress_desc_q[chan]);
        tera_mem_free(cblk->dev, cblk->egress_desc_q[chan]);

        // Free any inbound descriptor queue buffers
        for (i = 0; i < INGRESS_DESC_Q_LEN; i++)
        {
            tera_mem_buf_free(cblk->dev, cblk->ingress_desc_q_buf[chan][i]);
        }

        // Free any outbound descriptor queue buffers
        for (i = 0; i < EGRESS_DESC_Q_LEN; i++)
        {
            tera_mem_buf_free(cblk->dev, cblk->egress_desc_q_buf[chan][i]);
        }
    }
} // desc_q_destroy()

// --------------------------------------------------------------------------
//  gfun_desc_q_info_write
//      Write the details of the descriptor queues into the GFUN registers
//
//  Description:
//      This function writes the details of the descriptor queues into the
//      appropriate GFUN registers. The descriptor queues MUST be created prior
//      to calling this function.
//
static void gfun_desc_q_info_write(
    sTERA_GFUN_CBLK *cblk,      // Control block
    UINT32           chan       // GFUN channel
    )
{
    ASSERT(cblk != NULL);
    ASSERT(chan < GFUN_NUM_DMA_CHAN);

    // Write the egress descriptor queue to the registers
    reg_write(cblk->reg_base, GFUN_REG_CH_OUTDESCRLEN(chan),
              (EGRESS_DESC_Q_LEN * sizeof(sGFUN_DMA_DESC)) &
              GFUN_REG_CH_OUTDESCRLEN_DESCR_Q_LEN_MASK);
    reg_write(cblk->reg_base, GFUN_REG_CH_OUTDESCRBASEHI(chan),
              (UINT32)((UINT64)cblk->egress_desc_q[chan]->phys_addr >> 32) & 0xFFFFFFFF);
    reg_write(cblk->reg_base, GFUN_REG_CH_OUTDESCRBASELO(chan),
              (UINT32)cblk->egress_desc_q[chan]->phys_addr & 0xFFFFFFFF);

    // Write the ingress descriptor queue to the registers
    reg_write(cblk->reg_base, GFUN_REG_CH_INDESCRLEN(chan, cblk->dev_type),
              (INGRESS_DESC_Q_LEN * sizeof(sGFUN_DMA_DESC)) &
              GFUN_REG_CH_INDESCRLEN_DESCR_Q_LEN_MASK);
    reg_write(cblk->reg_base, GFUN_REG_CH_INDESCRBASEHI(chan, cblk->dev_type),
              (UINT32)((UINT64)cblk->ingress_desc_q[chan]->phys_addr >> 32) & 0xFFFFFFFF);
    reg_write(cblk->reg_base, GFUN_REG_CH_INDESCRBASELO(chan, cblk->dev_type),
              (UINT32)cblk->ingress_desc_q[chan]->phys_addr & 0xFFFFFFFF);

    // Disable writeback on both ingress and egress descriptor queues
    reg_read_write(cblk->reg_base, GFUN_REG_CH_INDMACNTRLSTAT(chan),
                   (UINT32)(~GFUN_REG_CH_INDMACNTRLSTAT_WB_DIS_MASK),
                   REG_OP_AND);
    reg_read_write(cblk->reg_base, GFUN_REG_CH_OUTDMACNTRLSTAT(chan),
                   (UINT32)(~GFUN_REG_CH_OUTDMACNTRLSTAT_WB_DIS_MASK),
                   REG_OP_AND);

    // Egress head and tail to zero to indicate empty queue
    egress_head_write(cblk, 0);
    egress_tail_write(cblk, 0);

    // Ingress head is one descriptor ahead of tail to indicate empty queue
    ingress_head_write(cblk, sizeof(sGFUN_DMA_DESC));
    ingress_tail_write(cblk, 0);

    mTERA_LOG_DBG("GFUN channel %u:", chan);
    mTERA_LOG_DBG("\tGFUN_REG_CH_OUTDESCRLEN=%u",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_OUTDESCRLEN(chan)));
    mTERA_LOG_DBG("\tGFUN_REG_CH_OUTDESCRBASEHI=0x%08x",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_OUTDESCRBASEHI(chan)));
    mTERA_LOG_DBG("\tGFUN_REG_CH_OUTDESCRBASELO=0x%08x",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_OUTDESCRBASELO(chan)));

    mTERA_LOG_DBG("\tegress_head=%u", egress_head_read(cblk));
    mTERA_LOG_DBG("\tegress_tail=%u", egress_tail_read(cblk));

    mTERA_LOG_DBG("\tGFUN_REG_CH_INDESCRLEN=%u",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_INDESCRLEN(chan, cblk->dev_type)));
    mTERA_LOG_DBG("\tGFUN_REG_CH_INDESCRBASEHI=0x%08x",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_INDESCRBASEHI(chan, cblk->dev_type)));
    mTERA_LOG_DBG("\tGFUN_REG_CH_INDESCRBASELO=0x%08x",
                  reg_read(cblk->reg_base,
                  GFUN_REG_CH_INDESCRBASELO(chan, cblk->dev_type)));

    mTERA_LOG_DBG("\tingress_head=%u", ingress_head_read(cblk));
    mTERA_LOG_DBG("\tingress_tail=%u", ingress_tail_read(cblk));
} // gfun_desc_q_info_write()

// --------------------------------------------------------------------------
//  gfun_device_hw_id_read
///     Read the PCoIP Hardware ID
//
//  Description:
///     This function reads the PCI card's vendor and device IDs to determine
///     the type of the Teradici processor installed on the host card.
///
/// Notes:
///     The code does not look at the subsystem vendor and subsystem device
///     IDs. This may become necessary if the driver needs to differentiate
///     between host cards that provide the same vendor and device IDs but
///     different subsystem vendor and device IDs. Currently this situation
///     does not exist.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
static INT32 gfun_device_hw_id_read(
    struct  pci_dev *dev,               ///< Device object
    eTERA_PCOIP_DEVICE_TYPE *dev_type   ///< Device type
    )
{
    INT32 ret = TERA_SUCCESS;

    // if the host card is a TERA1 card
    if ((dev->vendor == TERA_PCI_VENDOR_ID_TERA1) &&
        (dev->device == TERA_PCI_PRODUCT_ID_TERA1))
    {
        *dev_type = TERA_PCOIP_DEV_TERA1;
    }
    // if the host card is a TERA2 card
    else if ((dev->vendor == TERA_PCI_VENDOR_ID_TERA2) &&
             (dev->device == TERA_PCI_PRODUCT_ID_TERA2))
    {
        *dev_type = TERA_PCOIP_DEV_TERA2;
    }
    // unrecognized host card
    else
    {
        //
        // Set the device type equal to TERA1 to ensure the function returns
        // a valid type.
        //
        *dev_type = TERA_PCOIP_DEV_TERA1;
        ret = TERA_ERR_FAILURE;
    }

    if (ret == TERA_SUCCESS)
    {
        mTERA_LOG_DBG("ven=0x%04X, dev=0x%04X, subven=0x%04X, subdev=0x%04X, dev=%d, ret=%d",
                      dev->vendor,
                      dev->device,
                      dev->subsystem_vendor,
                      dev->subsystem_device,
                      *dev_type,
                      ret);
    }
    else
    {
        mTERA_LOG_ERR("ven=0x%04X, dev=0x%04X, subven=0x%04X, subdev=0x%04X, dev=%d, ret=%d",
                      dev->vendor,
                      dev->device,
                      dev->subsystem_vendor,
                      dev->subsystem_device,
                      *dev_type,
                      ret);
    }

    return ret;

} // gfun_device_hw_id_read()

// --------------------------------------------------------------------------
//  tera_pci_select_bars
///     Make BAR mask from the type of resource
//
//  Description:
///     This helper routine makes bar mask from the type of resource.
//
//  Parameters:
///     dev: pointer to the PCI device for which the BAR mask is made
///     flags: resource type to be selected
//
//  Notes:
///     This function was copied from the pci kernel code found in OpenSUSE
///     11.1. It was copied here because the function does not exist in
///     CentOS/RHEL 5.3.
//
static int tera_pci_select_bars(
    struct pci_dev *dev,
    unsigned long flags
    )
{
    int i, bars = 0;
    for (i = 0; i < PCI_NUM_RESOURCES; i++)
    {
        if (pci_resource_flags(dev, i) & flags)
        {
            bars |= (1 << i);
        }
    }
    return bars;
}

// --------------------------------------------------------------------------
//  tera_pci_release_selected_regions
///     Release selected PCI I/O and memory resources
//
//  Description:
///     Release selected PCI I/O and memory resources previously reserved.
///     Call this function only after all use of the PCI regions has ceased.
//
//  Parameters:
///     dev: pointer to the PCI device whose resources were previously reserved
///     bars: Bitmask of BARs to be released
//
//  Notes:
///     This function was copied from the pci kernel code found in OpenSUSE
///     11.1. It was copied here because the function does not exist in
///     CentOS/RHEL 5.3.
//
static void tera_pci_release_selected_regions(
    struct pci_dev *pdev,
    int bars
    )
{
    int i;

    for (i = 0; i < 6; i++)
    {
        if (bars & (1 << i))
        {
            pci_release_region(pdev, i);
        }
    }
}

// --------------------------------------------------------------------------
//  tera_pci_request_selected_regions
///     Reserve selected PCI I/O and memory resources
//
//  Description:
///     This helper routine reserves selected PCI I/O and memory resources
//
//  Parameters:
///     dev: pointer to the PCI device for which the BAR mask is made
///     bars: Bitmask of BARs to be requested
///     res_name: Name to be associated with resource
//
//  Notes:
///     This function was copied from the pci kernel code found in OpenSUSE
///     11.1. It was copied here because the function does not exist in
///     CentOS/RHEL 5.3.
//
static int tera_pci_request_selected_regions(
    struct pci_dev *pdev,
    int bars,
    const char *res_name
    )
{
    int i;

    for (i = 0; i < 6; i++)
    {
        if (bars & (1 << i))
        {
            if(pci_request_region(pdev, i, res_name))
            {
                goto err_out;
            }
        }
    }
    return 0;

err_out:
    while(--i >= 0)
    {
        if (bars & (1 << i))
        {
            pci_release_region(pdev, i);
        }
    }
    return -EBUSY;
}

// ---- Public Functions ----------------------------------------------------

// --------------------------------------------------------------------------
//  tera_gfun_init
///     Initialize the GFUN module
//
//  Description:
///     This function initializes the GFUN module. This function memory maps the
///     GFUN registers. It will also create the control block for this module
///     and attach it to the device object.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG
///
INT32 tera_gfun_init(
    struct  pci_dev *dev,
    char   *driver_name
    )
{
    INT32            ret;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
#ifdef GFUN_DBG
    UINT32 i;
    UINT8 *reg;
#endif

    mTERA_LOG_DBG("start");

    if (driver_name == NULL)
    {
        mTERA_LOG_CRIT("invalid driver name");
        return TERA_ERR_INVALID_ARG;
    }

    // Initialize the control block variables
    cblk->dev = dev;
    cblk->initialized = 0;
    cblk->started = 0;
    cblk->reg_base = NULL;

    // allocate memory mapped resources and enable the device
    cblk->bars = tera_pci_select_bars(dev, IORESOURCE_MEM);
    mTERA_LOG_DBG("bars %d", cblk->bars );
    ret = pci_enable_device(dev);
    if (ret)
    {
        return ret;
    }

    mTERA_LOG_DBG("pci_enable_device OK" );

    // Tell the kernel the device is capable of supporting 32-bit DMA
    // transfers. Kernel version 2.6.18 defined a new macro DMA_BIT_MASK()
    // which replaced the DMA_nnBIT_MASK macro. The 2.6.18 kernel supports
    // the use of both macros. Later versions of the kernel deprecate the
    // DMA_nnBIT_MASK macro, which causes compiler warnings.
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18))
    if (!pci_set_dma_mask(dev, DMA_BIT_MASK(32)) &&
        !pci_set_consistent_dma_mask(dev, DMA_BIT_MASK(32)))
#else
    if (!pci_set_dma_mask(dev, DMA_32BIT_MASK) &&
        !pci_set_consistent_dma_mask(dev, DMA_32BIT_MASK))
#endif
    {
        mTERA_LOG_DBG("pci_set_dma_mask OK");
    }
    else
    {
        mTERA_LOG_CRIT("No suitable DMA available, aborting");
        ret = TERA_ERR_NOT_FOUND;
        goto err_dma;
    }

    ret = tera_pci_request_selected_regions(dev, cblk->bars, driver_name);
    if (ret)
    {
        goto err_pci_req_reg;
    }

    // enable DMA
    pci_set_master(dev);

    // Map the BAR address (the device only supports BAR 0)
    cblk->reg_base = ioremap(pci_resource_start(dev, 0),
                             pci_resource_len(dev, 0));
    cblk->reg_len = pci_resource_len(dev, 0);
    if ((!cblk->reg_base) || (!cblk->reg_len))
    {
        mTERA_LOG_CRIT("reg_base(0x%lX) or length(%d) = 0",
                       (unsigned long)cblk->reg_base, cblk->reg_len);
        ret = TERA_ERR_FAILURE;
        goto err_bad_bar;
    }

    //
    // read the device hardware ID to determine what type of host card
    // is present
    //
    gfun_device_hw_id_read(dev, &cblk->dev_type);

#ifdef GFUN_DBG // Dump the contents of the registers
    reg = cblk->reg_base;

    for (i = 0; i < cblk->reg_len; i += 16)
    {
        mTERA_LOG_DBG("0x%02x: %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x", i,
                      reg[i+3], reg[i+2], reg[i+1], reg[i],
                      reg[i+7], reg[i+6], reg[i+5], reg[i+4],
                      reg[i+11], reg[i+10], reg[i+9], reg[i+8],
                      reg[i+15], reg[i+14], reg[i+13], reg[i+12]);
    }
#endif

    // Everything is ready to go
    cblk->initialized = 1;
    mTERA_LOG_DBG("end");
    return TERA_SUCCESS;

err_bad_bar:
    tera_pci_release_selected_regions(dev, cblk->bars);
err_pci_req_reg:
err_dma:
    pci_disable_device(dev);
    return ret;

} // tera_gfun_init()

// --------------------------------------------------------------------------
//  tera_gfun_exit
///     Clean up the GFUN module
//
//  Description:
///     This function cleans up the GFUN module. This function unmaps the GFUN
///     registers.
///
/// \return
///     None
///
void tera_gfun_exit(
    struct  pci_dev *dev
    )
{
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    mTERA_LOG_DBG("start");

    if ((cblk != NULL) && cblk->initialized)
    {
        tera_pci_release_selected_regions(dev, cblk->bars);
        pci_disable_device(dev);

        cblk->initialized = 0;
    }

    mTERA_LOG_DBG("end");
} // tera_gfun_exit()

// --------------------------------------------------------------------------
//  tera_gfun_start
///     Start the GFUN module
//
//  Description:
///     This function starts the GFUN module. This function will check the
///     FUNC_RDY bit in the memory-mapped registers to see if the GFUN is ready.
///     If the device is not ready, then this function returns
///     STATUS_DEVICE_NOT_READY. If the FUNC_RDY bit is set, then the INIT bit
///     is set to put the GFUN into the initialization state.
///
///     Then the FUNC_CAP register is read to determine if this driver is
///     compatible with the firmware. The value in the FUNC_CAP register is
///     bitwise AND'd with the value from the user. If the resulting value is
///     non-zero, then the GFUN proceeds with creating the descriptor queues
///     and placing the details into the GFUN registers. The INIT bit will then
///     be cleared to tell the GFUN that the descriptor queues are ready, and
///     the function returns STATUS_SUCCESS.
///
/// \return
///     TERA_SUCCESS
///     TERA_ERR_INVALID_STATE
///     TERA_ERR_NOT_READY
///     TERA_ERR_REINIT_NEEDED
///     TERA_ERR_REVISION_MISMATCH
///     TERA_ERR_NO_MEMORY
///
INT32 tera_gfun_start(
    struct  pci_dev *dev,
    UINT32           supported_func_cap
    )
{
    INT32            ret = TERA_SUCCESS;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
    UINT32           i;
    UINT32           try_count=0;

    mTERA_LOG_DBG("start %d: base=0x%08X_%08X, GlobCntlStat=0x%08X, rdy reg=0x%08X, dev type=%d",
                  supported_func_cap,
                  (UINT32)((UINT64)cblk->reg_base >> 32) & 0xFFFFFFFF,
                  (UINT32)((UINT64)cblk->reg_base & 0xFFFFFFFF),
                  reg_read(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT),
                  reg_read(cblk->reg_base, GFUN_REG_GLOB_FUNCSTAT),
                  cblk->dev_type);

    if (!cblk->initialized)
    {
        mTERA_LOG_CRIT("invalid state");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        if (cblk->started)
        {
            mTERA_LOG_DBG("already started");
        }
        else
        {
            //
            // Poll the device every second for the next 15 seconds waiting for
            // firmware to finish initializing.
            //
            while (try_count < 15)
            {
                // Wait for 1 second before trying again
                ssleep(1);
                try_count++;
                mTERA_LOG_DBG("Try %d: reg_base=0x%08X_%08X, rdy reg=0x%08X",
                              try_count,
                              (UINT32)((UINT64)cblk->reg_base >> 32) & 0xFFFFFFFF,
                              (UINT32)((UINT64)cblk->reg_base & 0xFFFFFFFF),
                              reg_read(cblk->reg_base, GFUN_REG_GLOB_FUNCSTAT));

                // if the device is ready exit the loop
                if (func_rdy_read(cblk) == 1)
                {
                    try_count=15;
                }
            }

            // Check if the FUNC_RDY bit is set
            if (func_rdy_read(cblk) == 0)
            {
                mTERA_LOG_ERR("device not ready (FUNC_RDY bit is cleared)");

                ret = TERA_ERR_NOT_READY;
            }
            else
            {
                // Put the GFUN into the initialization state first
                reg_read_write(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT,
                               GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_MASK, REG_OP_OR);

                // Wait up to 1 sec for GFUN to enter initialization state
                mTERA_LOG_DBG("verify GFUN in INIT state");
                for (i = 0; i < 200; i++)
                {
                    if (reg_read(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT) &
                        GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_ALIAS_MASK)
                    {
                        i = 0;
                        break;
                    }

                    // Wait for 5 ms before trying again
                    msleep(5);
                }

                if (i != 0)
                {
                    mTERA_LOG_ERR("GFUN did not enter initialization state!");

                    ret = TERA_ERR_REINIT_NEEDED;
                }
                else
                {
                    // Read the GFUN capabilities from the register
                    cblk->gfun_cap = reg_read(cblk->reg_base, GFUN_REG_GLOB_FUNCCAP);

                    //
                    // Determine if we know how to handle this instance of the GFUN by
                    // comparing the function capabilities reported in the GFUN register
                    // against the support function capabilities
                    //
                    if (!(cblk->gfun_cap & supported_func_cap))
                    {
                        mTERA_LOG_ERR("incompatible GFUN function capabilities");

                        ret = TERA_ERR_REVISION_MISMATCH;
                    }
                    else
                    {
                        mTERA_LOG_INFO("GFUN function capabilities (0x%08x) supported (0x%08x)",
                                       cblk->gfun_cap, supported_func_cap);

                        // Create descriptor queues for the GFUN channels
                        for (i = 0; i < GFUN_NUM_DMA_CHAN; i++)
                        {
                            // Create the ingress and egress descriptor queues
                            if (desc_q_create(cblk, i) != TERA_SUCCESS)
                            {
                                mTERA_LOG_ERR("failed to create descriptor queues");

                                ret = TERA_ERR_NO_MEMORY;
                                break;
                            }
                        }

                        if (ret == TERA_SUCCESS)
                        {
                            // Clear the INIT bit to let the GFUN know the queues are ready
                            reg_read_write(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT,
                                           (UINT32)(~GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_MASK),
                                           REG_OP_AND);

                            // Wait until the INIT_ALIAS bit is cleared or timeout after 1 second
                            for (i = 0; i < 200; i++)
                            {
                                if ((reg_read(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT) &
                                     GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_ALIAS_MASK) == 0)
                                {
                                    i = 0;
                                    break;
                                }

                                // Wait for 5 ms
                                msleep(5);
                            }

                            if (i != 0)
                            {
                                mTERA_LOG_ERR("GFUN did not clear INIT bit");

                                //FIXME Clean up descriptors and put device back into INIT state

                                ret = TERA_ERR_REINIT_NEEDED;
                            }
                            else
                            {
                                cblk->started = 1;
                                ret = TERA_SUCCESS;
                            }
                        }
                    }
                }
            }
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // tera_gfun_start()

// --------------------------------------------------------------------------
//  tera_gfun_stop
///     Stop the GFUN module
///     Stop the descriptor queues for Host OS <-> Tera communication
//
//  Description:
///     This function stops the GFUN module. This function puts the GFUN into
///     the initialization state to prevent it from using the descriptor queues.
///     The queues are also destroyed, effectively dropping any messages in
///     flight.
///
/// \return
///     None
///
void tera_gfun_stop(
    struct  pci_dev *dev
    )
{
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
    UINT32           i;

    mTERA_LOG_DBG("start");

    if ((cblk != NULL) && cblk->started)
    {
        // Reset the state variables for next time
        cblk->started = 0;

        // Put the GFUN back into the initialization state
        reg_read_write(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT,
                       GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_MASK, REG_OP_OR);

        // Wait up to 1 sec for GFUN to enter initialization state
        for (i = 0; i < 200; i++)
        {
            if (reg_read(cblk->reg_base, GFUN_REG_GLOB_GLOBCNTLSTAT) &
                GFUN_REG_GLOB_GLOBCNTLSTAT_INIT_ALIAS_MASK)
            {
                i = 0;
                break;
            }

            // Wait for 5 ms before trying again
            msleep(5);
        }

        if (i != 0)
        {
            mTERA_LOG_CRIT("GFUN did not enter initialization state");
        }

        // Destroy the ingress and egress descriptor queues
        for (i = 0; i < GFUN_NUM_DMA_CHAN; i++)
        {
            desc_q_destroy(cblk, i);
        }
    }

    mTERA_LOG_DBG("end");
} // tera_gfun_stop()

// --------------------------------------------------------------------------
//  tera_gfun_started
///     Returns if the GFUN module has entered the started state
//
//  Description:
///     This function returns if the GFUN module has entered the started state
///     for this device. If it has, this function returns TERA_SUCCESS.
///     Otherwise, this function returns TERA_ERR_INVALID_STATE.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE
///
INT32 tera_gfun_started(
    struct  pci_dev *dev
    )
{
    INT32            ret;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    if ((cblk != NULL) && cblk->started)
    {
        ret = TERA_SUCCESS;
    }
    else
    {
        ret = TERA_ERR_INVALID_STATE;
    }

    return ret;
} // tera_gfun_started()

// --------------------------------------------------------------------------
//  tera_gfun_mbox_read
///     Read a value from a GFUN mailbox slot
//
//  Description:
///     This function reads a 32-bit value from a GFUN mailbox slot. The GFUN
///     mailboxes can only be accessed when the GFUN module is started.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE
///
INT32 tera_gfun_mbox_read(
    struct  pci_dev     *dev,
    eTERA_GFUN_MBOX_SLOT mbox_slot,
    UINT32              *val
    )
{
    INT32            ret = TERA_ERR_INVALID_STATE;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    ASSERT(val != NULL);
    ASSERT(mbox_slot < TERA_GFUN_MBXO_SLOT_MAX);

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("GFUN module not started");
    }
    else
    {
        *val = reg_read(cblk->reg_base, mGFUN_REG_MBOX_SLOT(mbox_slot));
        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_gfun_mbox_read()

// --------------------------------------------------------------------------
//  tera_gfun_mbox_write
///     Write a value to a GFUN mailbox slot
//
//  Description:
///     This function writes a 32-bit value to a GFUN mailbox slot. The GFUN
///     mailboxes can only be accessed when the GFUN module is started.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE
///
INT32 tera_gfun_mbox_write(
    struct  pci_dev     *dev,
    eTERA_GFUN_MBOX_SLOT mbox_slot,
    UINT32               val
    )
{
    INT32            ret = TERA_ERR_INVALID_STATE;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    ASSERT(mbox_slot < TERA_GFUN_MBXO_SLOT_MAX);

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("GFUN module not started");
    }
    else
    {
        reg_write(cblk->reg_base, mGFUN_REG_MBOX_SLOT(mbox_slot), val);
        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_gfun_mbox_write()

// --------------------------------------------------------------------------
//  tera_gfun_cap_get
///     Return the GFUN functions capabilities value
//
//  Description:
///     This function returns the GFUN function capabilities value. This
///     function returns a valid value ONLY if this module has been started.
///
/// \return
///     GFUN function capabilities value
///
UINT32 tera_gfun_cap_get(
    struct  pci_dev *dev
    )
{
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);

    return dev_cblk->gfun_cblk.gfun_cap;
} // tera_gfun_cap_get()

// --------------------------------------------------------------------------
//  tera_gfun_tx_ready
///     Check if egress descriptor queue is able to accept more data
//
//  Description:
///     This function checks if the egress descriptor queue is full or not. This
///     function should be called prior to calling tera_gfun_tx. If the egress
///     queue is not full, i.e., able to accept more data, then this function
///     returns TERA_SUCCESS. If the queue is full, this function returns
///     TERA_ERR_FULL.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_FULL
///
INT32 tera_gfun_tx_ready(
    struct  pci_dev *dev
    )
{
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
    INT32           ret;

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("invalid state");

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        UINT32 head;
        UINT32 tail;

        // Get the current head and tail pointer from register
        head = egress_head_read(cblk);
        tail = egress_tail_read(cblk);

        // Check if buffer is full
        if (((head < tail) &&
             ((tail - head) ==
              ((EGRESS_DESC_Q_LEN - 1) * sizeof(sGFUN_DMA_DESC)))) ||
            ((head > tail) && ((tail + sizeof(sGFUN_DMA_DESC)) == head)))
        {
            ret = TERA_ERR_FULL;
        }
        else
        {
            ret = TERA_SUCCESS;
        }
    }

    return ret;
} // tera_gfun_tx_ready()

#ifdef GFUN_DBG
// --------------------------------------------------------------------------
//  tera_gfun_reg_dump
///     Dump the contents of the GFUN control registers
//
//  Description:
///     This is a debug function that dumps the contents of the GFUN control
///     registers.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE
///
INT32 tera_gfun_reg_dump(
    struct  pci_dev *dev
    )
{
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
    UINT32           reg_base;

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("invalid state");

        return TERA_ERR_INVALID_STATE;
    }
    else
    {
        reg_base = (UINT32)cblk->reg_base;
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDMACNTRLSTAT(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDMACNTRLSTAT(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDMACNTRLSTAT(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDMACHNLDMAEN(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDMACHNLDMAEN(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDMACHNLDMAEN(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDESCRLEN(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDESCRLEN(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRLEN(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDESCRBASELO(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDESCRBASELO(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRBASELO(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDESCRBASEHI(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDESCRBASEHI(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRBASEHI(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDESCRHEADPTR(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDESCRHEADPTR(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRHEADPTR(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_OUTDESCRTAILPTR(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_OUTDESCRTAILPTR(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_OUTDESCRTAILPTR(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDMACNTRLSTAT(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDMACNTRLSTAT(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDMACNTRLSTAT(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDMACHNLDMAEN(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDMACHNLDMAEN(0),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDMACHNLDMAEN(0)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDESCRLEN(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDESCRLEN(0, cblk->dev_type),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRLEN(0, cblk->dev_type)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDESCRBASELO(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDESCRBASELO(0, cblk->dev_type),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRBASELO(0, cblk->dev_type)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDESCRBASEHI(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDESCRBASEHI(0, cblk->dev_type),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRBASEHI(0, cblk->dev_type)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDESCRHEADPTR(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDESCRHEADPTR(0, cblk->dev_type),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRHEADPTR(0, cblk->dev_type)));
        mTERA_LOG_DBG("GFUN_REG_CH_INDESCRTAILPTR(0x%08X) = 0x%08X",
                      reg_base + GFUN_REG_CH_INDESCRTAILPTR(0, cblk->dev_type),
                      reg_read(cblk->reg_base, GFUN_REG_CH_INDESCRTAILPTR(0, cblk->dev_type)));
    }

    return TERA_SUCCESS;
} // tera_gfun_reg_dump()
#endif

// --------------------------------------------------------------------------
//  tera_gfun_tx
///     Send a message to the egress descriptor queue
//
//  Description:
///     This function sends a message to the egress descriptor queue. If the
///     message contains a buffer, then the DMA descriptor is updated to point
///     to that buffer. The user must NOT touch or free this buffer anymore. The
///     buffer is freed on the next time the tail pointer points to the same
///     DMA descriptor.
///
///     If the message was successfully sent to the egress descriptor queue,
///     then this function returns TERA_SUCCESS.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_FULL
///
INT32 tera_gfun_tx(
    struct  pci_dev     *dev,
    sTERA_MEM_BUF       *buf,
    UINT32               buf_len,
    sTERA_GFUN_PKT_CTRL *pkt_ctrl
    )
{
    INT32            ret;
    sGFUN_DMA_DESC  *desc;
    UINT32           tail;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    // Make sure egress queue is not full first
    if ((ret = tera_gfun_tx_ready(dev)) == TERA_SUCCESS)
    {
        tail = egress_tail_read(cblk);

        // Get a pointer to the DMA descriptor
        desc = offset_to_dma_desc(cblk->egress_desc_q[0], tail);

        // Free the existing buffer (if exists from previous tranmit)
        if (desc->buf_addr != 0)
        {
            tera_mem_buf_free(cblk->dev,
                              cblk->egress_desc_q_buf[0][tail / sizeof(sGFUN_DMA_DESC)]);

            cblk->egress_desc_q_buf[0][tail / sizeof(sGFUN_DMA_DESC)] = NULL;
            desc->buf_addr = 0;
        }

        // Fill out the DMA descriptor
        desc->buf_len = buf_len & GFUN_DMA_DESC_BUF_LEN_MASK;
        desc->dma_disable = 0;
        desc->len = buf_len & GFUN_DMA_DESC_LEN_MASK;
        desc->flags = GFUN_DMA_DESC_FLAGS_EOP_MASK | //TODO change this for fragmentation
                      GFUN_DMA_DESC_FLAGS_VALID_MASK;
        desc->pkt_err = 0;

        if (pkt_ctrl != NULL)
        {
            // Copy the packet control to the DMA descriptor
            memcpy(&desc->pkt_ctrl, pkt_ctrl,
                   sizeof(sTERA_GFUN_PKT_CTRL));
        }
        else
        {
            // Zero out the packet control in the DMA descriptor
            memset(&desc->pkt_ctrl, 0, sizeof(sTERA_GFUN_PKT_CTRL));
        }

        if (buf_len > 0)
        {
            // Fill out the buffer address in the DMA descriptor
            desc->buf_addr = buf->phys_addr;

            cblk->egress_desc_q_buf[0][tail / sizeof(sGFUN_DMA_DESC)] = buf;
        }

        // Update the tail pointer in the mem-mapped registers
        tail += sizeof(sGFUN_DMA_DESC);
        tail = tail % (EGRESS_DESC_Q_LEN * sizeof(sGFUN_DMA_DESC));
        egress_tail_write(cblk, tail);
    } // end if

    return ret;
} // tera_gfun_tx()

// --------------------------------------------------------------------------
//  tera_gfun_rx
///     Retrieve next buffer from ingress descriptor queue
//
//  Description:
///     This function retrieves the next buffer from the ingress descriptor
///     queue if one is available. The function will return a pointer to the
///     chunk of DMA'able memory and replenish the descriptor queue with a new
///     buffer.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_EMPTY,
///     TERA_ERR_NO_MEMORY
///
INT32 tera_gfun_rx(
    struct  pci_dev     *dev,
    sTERA_MEM_BUF      **buf,
    UINT32              *buf_len,
    sTERA_GFUN_PKT_CTRL *pkt_ctrl
    )
{
    INT32            ret=TERA_SUCCESS;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;
    UINT32           head;
    UINT32           tail;
    sGFUN_DMA_DESC  *desc;

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("invalid state");

        ret = TERA_ERR_INVALID_STATE;
    }
    else if ((buf == NULL) || (buf_len == NULL))
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // Get the current head and tail pointer from register
        head = ingress_head_read(cblk);
        tail = ingress_tail_read(cblk);

        // Check to if buffer is empty
        if (((head < tail) && ((tail - head) == ((EGRESS_DESC_Q_LEN - 1) * sizeof(sGFUN_DMA_DESC)))) ||
            ((head > tail) && ((tail + sizeof(sGFUN_DMA_DESC)) == head)))
        {
            ret = TERA_ERR_EMPTY;
        }
        else
        {
#ifdef GFUN_DBG
            mTERA_LOG_DBG("head '%u' tail '%u' sizeof(sGFUN_DMA_DESC) '%u'", head, tail, sizeof(sGFUN_DMA_DESC));
#endif

            // Get the location of the next tail pointer
            tail += sizeof(sGFUN_DMA_DESC);
            tail = tail % (INGRESS_DESC_Q_LEN * sizeof(sGFUN_DMA_DESC));

            desc = offset_to_dma_desc(cblk->ingress_desc_q[0], tail);

            if (pkt_ctrl != NULL)
            {
                // Copy the packet control from the DMA descriptor
                memcpy(pkt_ctrl, &desc->pkt_ctrl,
                       sizeof(sTERA_GFUN_PKT_CTRL));
            }

            // Extract the buffer length from the DMA descriptor
            *buf_len = desc->len;

            if (desc->len > 0)
            {
                sTERA_MEM_BUF *new_buf;

                // Allocate a new buffer to replace the one we will take out
                new_buf = tera_mem_buf_alloc(dev, cblk->mtu_size[0]);
                if (new_buf == NULL)
                {
                    mTERA_LOG_ERR("failed to allocate new buffer");

                    ret = TERA_ERR_NO_MEMORY;
                }
                else
                {
                    // Get the virtual address of the descriptor queue buffer
                    *buf = cblk->ingress_desc_q_buf[0][tail / sizeof(sGFUN_DMA_DESC)];

                    // Fill out the buffer address in the DMA descriptor
                    desc->buf_addr = new_buf->phys_addr;

                    // Update the local record of the buffer
                    cblk->ingress_desc_q_buf[0][tail / sizeof(sGFUN_DMA_DESC)] = new_buf;
                }
            }
            else
            {
                *buf = NULL;
            }

            if (ret == TERA_SUCCESS)
            {
                // Clear the valid bit to give control back to GFUN
                desc->flags &= ~GFUN_DMA_DESC_FLAGS_VALID_MASK;

                // Update the tail pointer in the register
                ingress_tail_write(cblk, tail);
            }
        }
    } // end if

    return ret;
} // tera_gfun_rx()

// --------------------------------------------------------------------------
//  tera_gfun_dev_type_read
///     Read the host card device type
//
//  Description:
///     This function reads the device type of the PCoIP processor on the host
///     card.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_EMPTY
///
INT32 tera_gfun_dev_type_read(
    struct  pci_dev *dev,
    UINT32          *type
    )
{
    INT32            ret;
    sTERA_DEV_CBLK  *dev_cblk = pci_get_drvdata(dev);
    sTERA_GFUN_CBLK *cblk = &dev_cblk->gfun_cblk;

    if ((cblk == NULL) || !cblk->started)
    {
        mTERA_LOG_ERR("invalid state");

        ret = TERA_ERR_INVALID_STATE;
    }
    else if (type == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        *type = cblk->dev_type;

        ret = TERA_SUCCESS;
    } // end if

    return ret;
} // tera_gfun_dev_type_read()

