////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     GFUN Module
///
///     This file defines the GFUN module API prototypes and definitions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_mem.h"

// ---- Constants and Types -------------------------------------------------

// --------------------------------------------------------------------------
//  sTERA_GFUN_PKT_CTRL
///     Packet control
//
//  Description:
///     This structure defines the bits of the packet control. Only the first
///     56-bits of the packet control are actually valid. The last eight bits
///     should always be zero.
///
typedef struct sTERA_GFUN_PKT_CTRL_STRUCT
{
    unsigned bit_0_31  : 32;
    unsigned bit_32_55 : 24;
    unsigned unused    : 8;
} sTERA_GFUN_PKT_CTRL;

// --------------------------------------------------------------------------
//  eTERA_GFUN_MBOX_SLOT
///     GFUN mailbox slots
//
//  Description:
///     This enumeration lists all the available GFUN mailbox slots for reading
///     and writing. The mailboxes provide a way of communicating small amounts
///     of information between the firmware and host OS. Each mailbox slot
///     stores a 32-bit value. The information stored in the mailboxes is
///     mutually decided upon by the firmware and host OS drivers.
///
typedef enum
{
    TERA_GFUN_MBOX_SLOT_0   = 0,
    TERA_GFUN_MBOX_SLOT_1   = 1,
    TERA_GFUN_MBOX_SLOT_2   = 2,
    TERA_GFUN_MBOX_SLOT_3   = 3,
    TERA_GFUN_MBOX_SLOT_4   = 4,
    TERA_GFUN_MBOX_SLOT_5   = 5,
    TERA_GFUN_MBOX_SLOT_6   = 6,
    TERA_GFUN_MBOX_SLOT_7   = 7,
    TERA_GFUN_MBOX_SLOT_8   = 8,
    TERA_GFUN_MBOX_SLOT_9   = 9,
    TERA_GFUN_MBOX_SLOT_10  = 10,
    TERA_GFUN_MBOX_SLOT_11  = 11,
    TERA_GFUN_MBOX_SLOT_12  = 12,
    TERA_GFUN_MBOX_SLOT_13  = 13,
    TERA_GFUN_MBOX_SLOT_14  = 14,
    TERA_GFUN_MBOX_SLOT_15  = 15,
    TERA_GFUN_MBOX_SLOT_16  = 16,
    TERA_GFUN_MBOX_SLOT_17  = 17,
    TERA_GFUN_MBOX_SLOT_18  = 18,
    TERA_GFUN_MBOX_SLOT_19  = 19,
    TERA_GFUN_MBOX_SLOT_20  = 20,
    TERA_GFUN_MBOX_SLOT_21  = 21,
    TERA_GFUN_MBOX_SLOT_22  = 22,
    TERA_GFUN_MBOX_SLOT_23  = 23,
    TERA_GFUN_MBOX_SLOT_24  = 24,
    TERA_GFUN_MBOX_SLOT_25  = 25,
    TERA_GFUN_MBOX_SLOT_26  = 26,
    TERA_GFUN_MBOX_SLOT_27  = 27,
    TERA_GFUN_MBOX_SLOT_28  = 28,
    TERA_GFUN_MBOX_SLOT_29  = 29,
    TERA_GFUN_MBOX_SLOT_30  = 30,
    TERA_GFUN_MBOX_SLOT_31  = 31,
    TERA_GFUN_MBXO_SLOT_MAX
} eTERA_GFUN_MBOX_SLOT;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_DEVICE_TYPE
///     PCoIP Device types
//
//  Description:
///     This enumeration defines all Teradici host card device types
///
/// Note:
///     The values of these macros must not change because the values are
///     used to calculate register offsets. If the values change the macros
///     that use them must be updated.
///
typedef enum
{
    TERA_PCOIP_DEV_TERA1    = 0,    ///< TERA1
    TERA_PCOIP_DEV_TERA2    = 1,    ///< TERA2
} eTERA_PCOIP_DEVICE_TYPE;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

extern INT32 tera_gfun_init(
    struct  pci_dev *dev,
    char            *driver_name
    );

extern void tera_gfun_exit(
    struct  pci_dev *dev
    );

extern INT32 tera_gfun_start(
    struct  pci_dev *dev,
    UINT32           supported_func_cap
    );

extern void tera_gfun_stop(
    struct  pci_dev *dev
    );

extern INT32 tera_gfun_started(
    struct  pci_dev *dev
    );

extern UINT32 tera_gfun_cap_get(
    struct  pci_dev *dev
    );

extern INT32 tera_gfun_mbox_read(
    struct  pci_dev     *dev,
    eTERA_GFUN_MBOX_SLOT mbox_slot,
    UINT32              *val
    );

extern INT32 tera_gfun_mbox_write(
    struct  pci_dev     *dev,
    eTERA_GFUN_MBOX_SLOT mbox_slot,
    UINT32               val
    );

extern INT32 tera_gfun_tx_ready(
    struct  pci_dev *dev
    );

#ifdef GFUN_DBG
extern INT32 tera_gfun_reg_dump(
    struct  pci_dev *dev
    );
#endif

extern INT32 tera_gfun_tx(
    struct  pci_dev     *dev,
    sTERA_MEM_BUF       *buf,
    UINT32               buf_len,
    sTERA_GFUN_PKT_CTRL *pkt_ctrl
    );

extern INT32 tera_gfun_rx(
    struct  pci_dev     *dev,
    sTERA_MEM_BUF      **buf,
    UINT32              *buf_len,
    sTERA_GFUN_PKT_CTRL *pkt_ctrl
    );

extern INT32 tera_gfun_dev_type_read(
    struct  pci_dev *dev,
    UINT32          *type
    );

