////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     HCC Module HMI User Implementation
///
///     This file implements the HCC module HMI user. All HMI user related
///     APDUs and ioctls are dealt with in this file.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"

#include "tera_pcoip_hcc_ioctl.h"
#include "tera_hcc.h"
#include "tera_hcc_internal.h"
#include "tera_hcc_hmi.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

//
// Macros to get the pointer to the HMI APDU data from an APDU pointer.
//
//      -----------------------
//      |   HMI APDU header   |
//      |---------------------|
//      |    HMI APDU data    |
//      |                     |
//
#define mHMI_APDU_DATA_PTR_GET(apdu) \
    (void *)((UINT8 *)(apdu) + sizeof(sHCC_HMI_APDU_HDR))

// ---- Local Variables -----------------------------------------------------

// HMI commands string table
#ifdef INFO
static const char *hmi_cmd_str_tbl[TERA_PCOIP_HCC_APDU_HMI_CMD_MAX] = {
    "HMI_CMD_AVAIL_CMD",
    "HMI_CMD_VAR_GET",
    "HMI_CMD_VAR_SET"
};

// HMI variables string table
static const char *hmi_var_str_tbl[TERA_PCOIP_HCC_APDU_HMI_VAR_MAX] = {
    "HMI_VAR_AVAIL_VAR",
    "HMI_VAR_LINK_INFO",
    "HMI_VAR_IPV4_CFG",
    "HMI_VAR_SESS_LINK",
    "HMI_VAR_SESS_TAG",
    "HMI_VAR_CONN_PORT",
    "HMI_VAR_SESS_STAT",
    "HMI_VAR_DEPRECATED",
    "HMI_VAR_WOL_CFG",
    "HMI_VAR_SESS_DISC",
    "HMI_VAR_SESS_TAG2",
    "HMI_VAR_CACHED_TOPO",
    "HMI_VAR_UNICODE",
    "HMI_VAR_IMG_USER_EXT_CFG",
};
#endif

// ---- Local Function Prototypes -------------------------------------------

static void hmi_avail_cmd_var_reset(
    sHCC_CBLK *cblk
    );

static INT32 hmi_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static INT32 hmi_ioctl_avail_cmd_apdu_rep(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static INT32 hmi_ioctl_var_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static INT32 hmi_ioctl_local_var_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static UINT32 hmi_ioctl_var_get_avail_var_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    );

static INT32 hmi_ioctl_var_get_sess_link_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    );

static INT32 hmi_ioctl_var_get_fwd_to_blade(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    );

static INT32 hmi_ioctl_var_set_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static INT32 hmi_ioctl_local_var_set_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    );

static INT32 hmi_ioctl_var_set_avail_var_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    );

static INT32 hmi_ioctl_var_set_fwd_to_blade(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    );

static void hmi_avail_cmd_apdu_send(
    sHCC_CBLK *cblk
    );

static void hmi_avail_cmd_apdu_proc(
    sHCC_CBLK *cblk,
    void      *data,
    UINT32     data_len
    );

static void hmi_var_get_avail_var_apdu_send(
    sHCC_CBLK *cblk
    );

static void hmi_var_set_avail_var_apdu_send(
    sHCC_CBLK *cblk
    );

static void hmi_var_get_apdu_proc(
    sHCC_CBLK *cblk,
    void      *data,
    UINT32     data_len
    );

static void hmi_apdu_var_get_avail_var_proc(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_MAP *avail_var
    );

static void hmi_var_get_sess_link_apdu_send(
    sHCC_CBLK *cblk
    );

static void hmi_apdu_var_get_sess_link_proc(
    sHCC_CBLK                              *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link
    );

static void hmi_apdu_var_get_fwd_to_user(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep
    );

static void hmi_var_set_apdu_proc(
    sHCC_CBLK *cblk,
    void      *data,
    UINT32     data_len
    );

static void hmi_var_set_avail_var_apdu_proc(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_MAP *avail_var
    );

static void hmi_var_set_fwd_to_user_apdu(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep
    );

// --------------------------------------------------------------------------
//  hmi_apdu_send
//      Send a HMI APDU
//
//  Description:
//      This function sends a HMI APDU to the blade. The contents of the APDU
//      are copied into an allocated packet and enqueued onto the egress
//      packet queue. The calling function can safely discard the APDU after
//      this function call.
//
static INT32 hmi_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    // Send the APDU
    ret = tera_scp_data_pkt_send(cblk->dev, cblk->scp_handle,
                                 TERA_PCOIP_HCC_APDU_USER_HMI,
                                 &cblk->egress_pkt_q, apdu, apdu_len);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
        ret = -ERESTARTSYS;
    }
    return ret;
} // hmi_apdu_send()

// --------------------------------------------------------------------------
//  hmi_ioctl_avail_cmd_apdu_rep
//      Reply to a HMI_AVAIL_CMD ioctl
//
//  Description:
//      This function replies to a HMI_AVAIL_CMD ioctl from userspace with the
//      list of available HMI commands.
//
static INT32 hmi_ioctl_avail_cmd_apdu_rep(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                        ret = TERA_SUCCESS;
    sTERA_PCOIP_HCC_APDU_HMI_MAP *buf;

    ASSERT(cblk != NULL);
    buf = (sTERA_PCOIP_HCC_APDU_HMI_MAP *)arg;

    if(!down_interruptible(&cblk->data_lock))
    {
        // Copy the feature map to the user buffer
        if(copy_to_user((char *)arg,
                        &cblk->hmi.avail_cmd.map[0],
                        sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HMI_AVAIL_CMD");
    }
    else
    {
        mTERA_LOG_DBG("avail_cmd.map[]=0x%08X 0x%08X 0x%08X 0x%08X 0x%08X 0x%08X 0x%08X 0x%08X",
                      buf->map[0], buf->map[1],
                      buf->map[2], buf->map[3],
                      buf->map[4], buf->map[5],
                      buf->map[6], buf->map[7]);
    }

    return ret;
} // hmi_ioctl_avail_cmd_apdu_rep()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_get_proc
//      Process a HMI_VAR_GET ioctl
//
//  Description:
//      This function processes a HMI_VAR_GET ioctl from userspace. This
//      function will determine which HMI variable is being requested and carry
//      out the correct processing. If the ioctl is not supported, it is
//      completed with STATUS_NOT_SUPPORTED.
//
static INT32 hmi_ioctl_var_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    )
{
    UINT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf;
    sTERA_PCOIP_HCC_APDU_HMI_VAR  apdu;

    // assign a pointer to the input buffer
    buf = (sTERA_PCOIP_HCC_APDU_HMI_VAR *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        switch (apdu.var)
        {
            case TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR:
            {
                // Reply with the available variables for getting
                ret = hmi_ioctl_var_get_avail_var_apdu_rep(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO:
            {
                mTERA_LOG_DBG("HMI_VAR_LINK_INFO");

                // Forward the request to the blade
                ret = hmi_ioctl_var_get_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG:
            {
                mTERA_LOG_DBG("HMI_VAR_IPV4_CFG");

                // Forward the request to the blade
                ret = hmi_ioctl_var_get_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK:
            {
                // Reply with the session link information
                ret = hmi_ioctl_var_get_sess_link_apdu_rep(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT:
            {
                mTERA_LOG_DBG("HMI_VAR_CONN_PORT");

                // Forward the request to the blade
                ret = hmi_ioctl_var_get_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT:
            {
                mTERA_LOG_DBG("HMI_VAR_SESS_STAT");

                // Forward the request to the blade
                ret = hmi_ioctl_var_get_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG:
            {
                mTERA_LOG_DBG("HMI_VAR_WOL_CFG");

                // Forward the request to the blade
                ret = hmi_ioctl_var_get_fwd_to_blade(cblk, buf);
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported HMI variable=%u", apdu.var);
                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }
    return ret;
} // hmi_ioctl_var_get_proc()

// --------------------------------------------------------------------------
//  hmi_ioctl_local_var_get_proc
//      Process a HMI_LOCAL_VAR_GET ioctl
//
//  Description:
//      This function processes a HMI_LOCAL_VAR_GET ioctl from userspace. This
//      function will determine which local HMI variable is being requested and
//      carry out the correct processing. If the ioctl is not supported, it is
//      completed with TERA_ERR_NOT_SUPPORTED.
//
static INT32 hmi_ioctl_local_var_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                      ret;
    sTERA_PCOIP_HMI_LOCAL_VAR *buf;
    sTERA_PCOIP_HMI_LOCAL_VAR  apdu;

    // assign a pointer to the input buffer
    buf = (sTERA_PCOIP_HMI_LOCAL_VAR *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        switch (apdu.var)
        {
            case TERA_PCOIP_HMI_LOCAL_VAR_HOST_SESS_STATE:
            {
                apdu.data.sess_state.host_sess_state = cblk->host_sess_state;
                mTERA_LOG_DBG("HMI_LOCAL_VAR_SESS_STATE host state=%d",
                              apdu.data.sess_state.host_sess_state);
                ret = TERA_SUCCESS;
                break;
            }
            case TERA_PCOIP_HMI_LOCAL_VAR_LOCAL_INFO:
            {
                // Fill in the fields in the response
                ret = tera_gfun_dev_type_read(cblk->dev, &(apdu.data.info.dev_type));
                if (ret == TERA_SUCCESS)
                {
                    // Dump the data in the response
                    mTERA_LOG_DBG("TERA processor device type=%u",
                                  apdu.data.info.dev_type);
                }
                else
                {
                    mTERA_LOG_ERR("Failed to get local information (%d)!!!", ret);
                }
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported local HMI variable=%u", apdu.var);
                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }

    // copy the response to the user buffer
    if (ret == TERA_SUCCESS)
    {
        if (copy_to_user(buf, &apdu, sizeof(apdu)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            ret = TERA_SUCCESS;
        }
    }

    return ret;
} // hmi_ioctl_local_var_get_proc()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_get_avail_var_apdu_rep
//      Reply to a HMI_VAR_GET ioctl with HMI_AVAIL_VAR variable
//
//  Description:
//      This function replies to a HMI_VAR_GET ioctl from userspace with the
//      list of available HMI variables for getting.
//
static UINT32 hmi_ioctl_var_get_avail_var_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    )
{
    UINT32 ret = TERA_SUCCESS;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    if(!down_interruptible(&cblk->data_lock))
    {
        // Copy the list of HMI variables available for getting to the buffer
        if(copy_to_user(&buf->data.avail_var,
                        &cblk->hmi.avail_var_get,
                        sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HMI_VAR_GET HMI_AVAIL_VAR");
    }
    else
    {
        mTERA_LOG_DBG("get avail_var: 0x%08X 0x%08X 0x%08X 0x%08X",
                          cblk->hmi.avail_var_get.map[0],
                          cblk->hmi.avail_var_get.map[1],
                          cblk->hmi.avail_var_get.map[2],
                          cblk->hmi.avail_var_get.map[3]);
        mTERA_LOG_DBG("               0x%08X 0x%08X 0x%08X 0x%08X",
                          cblk->hmi.avail_var_get.map[4],
                          cblk->hmi.avail_var_get.map[5],
                          cblk->hmi.avail_var_get.map[6],
                          cblk->hmi.avail_var_get.map[7]);
    }

    return ret;
} // hmi_ioctl_var_get_avail_var_apdu_rep()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_get_sess_link_apdu_rep
//      Reply to a HMI_VAR_GET ioctl with HMI_VAR_SESS_INFO variable
//
//  Description:
//      This function replies to a HMI_VAR_GET ioctl from userspace with the
//      session link information.
//
static INT32 hmi_ioctl_var_get_sess_link_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    )
{
    UINT32 ret = TERA_SUCCESS;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    if(!down_interruptible(&cblk->data_lock))
    {
        // Copy the session link information from the control block to the buffer
        if(copy_to_user(&buf->data.sess_link,
                        &cblk->hmi.sess_link,
                        sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HMI_VAR_GET");
    }
    else
    {
        mTERA_LOG_DBG("sess_link: flags='0x%02x', ip=%s, mac=%s",
                      cblk->hmi.sess_link.flags,
                      cblk->hmi.sess_link.ip,
                      cblk->hmi.sess_link.mac);
    }

    return ret;
} // hmi_ioctl_var_get_sess_link_apdu_rep()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_get_fwd_to_blade
//      Forward a HMI_VAR_GET ioctl to the blade
//
//  Description:
//      This function forwards a HMI_VAR_GET ioctl to the blade. The input
//      buffer of the ioctl is a HMI_VAR_GET APDU. The buffer is copied
//      to a packet and sent down to the blade. The ioctl is then put into the
//      pending ioctl queue (to be completed when a response comes back from
//      the blade).
//
static INT32 hmi_ioctl_var_get_fwd_to_blade(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    )
{
    INT32             ret;
    sHCC_HMI_APDU_VAR apdu;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // Fill out the APDU header
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET;

    // Fill out the APDU data
    if(copy_from_user(&apdu.data, buf, sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU
            ret = hmi_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                up(&cblk->pending_ioctl_lock);
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
//
// TBD - possibly enhance the code to store the ioctl in this flag variable,
// similar to what's been done in the file tera_hcc_hmi.c. It's a little more
// complex here because the same ioctl value is used for multiple responses.
// The code might need to use the buf->var field, which would be unique.
//
                // clear flag that is inspected when the workqueue is inspected
                cblk->pending_ioctl_flag = 0;

                //
                // Wait until the response is received on the ingress message Q and
                // processed by the dpc_entry. The timeout is doubled because the
                // WOL CFG data is retrieved from the client device and this can
                // take more than 1 second.
                //
                if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                     (cblk->pending_ioctl_flag == 1),
                                                     (2*msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS))) <= 0)
                {
                    mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                    up(&cblk->pending_ioctl_lock);
                    cblk->pending_ioctl_flag = 1;
                    ret = -ERESTARTSYS;
                }
                else
                {
                    // fill in the IOCTL response
                    if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
                    {
                        mTERA_LOG_ERR("copy_to_user failed");
                        ret = -ERESTARTSYS;
                    }
                    else
                    {
                        ret = TERA_SUCCESS;
                    }
                    up(&cblk->pending_ioctl_lock);
                }
            }
        }
    }
    return ret;
} // hmi_ioctl_var_get_fwd_to_blade()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_set_proc
//      Process a HMI_VAR_SET ioctl
//
//  Description:
//      This function processes a HMI_VAR_SET ioctl from userspace. This
//      function will determine which HMI variable is being requested and carry
//      out the correct processing. If the ioctl is not supported, it is
//      completed with STATUS_NOT_SUPPORTED.
//
static INT32 hmi_ioctl_var_set_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    )
{
    INT32                         ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf;
    sTERA_PCOIP_HCC_APDU_HMI_VAR  apdu;

    // assign a pointer to the input buffer
    buf = (sTERA_PCOIP_HCC_APDU_HMI_VAR *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        switch (apdu.var)
        {
            case TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR:
            {
                mTERA_LOG_DBG("HMI_VAR_AVAIL_VAR");
                ret = hmi_ioctl_var_set_avail_var_apdu_rep(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK:
            {
                mTERA_LOG_DBG("set sess_link: flags=0x%02x", apdu.data.sess_link.flags);

                // Forward the request to the blade
                ret = hmi_ioctl_var_set_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT:
            {
                mTERA_LOG_DBG("set sess_stat");

                // Forward the request to the blade
                ret = hmi_ioctl_var_set_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG:
            {
                mTERA_LOG_DBG("set WOL cfg: flags=0x%02x, mac=%s ip=%s",
                              apdu.data.wol_cfg.flags,
                              apdu.data.wol_cfg.mac,
                              apdu.data.wol_cfg.ip);

                // Forward the request to the blade
                ret = hmi_ioctl_var_set_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC:
            {
                mTERA_LOG_DBG("set sess_disc: reason=0x%08x",
                              apdu.data.sess_disc.reason_cause);

                // Forward the request to the blade
                ret = hmi_ioctl_var_set_fwd_to_blade(cblk, buf);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE:
            {
                mTERA_LOG_DBG("set unicode: enable=%d",
                              apdu.data.unicode.enabled);

                // Forward the request to the blade
                ret = hmi_ioctl_var_set_fwd_to_blade(cblk, buf);
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported HMI variable=%u", apdu.var);
                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }

    return ret;
} // hmi_ioctl_var_set_proc()

// --------------------------------------------------------------------------
//  hmi_ioctl_local_var_set_proc
//      Process a HMI_LOCAL_VAR_SET ioctl
//
//  Description:
//      This function processes a HMI_LOCAL_VAR_SET ioctl from userspace. This
//      function will determine which local HMI variable is being set and carry
//      out the correct processing. If the ioctl is not supported, it is
//      completed with TERA_ERR_NOT_SUPPORTED.
//
static INT32 hmi_ioctl_local_var_set_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    )
{
    INT32                      ret;
    sTERA_PCOIP_HMI_LOCAL_VAR *buf;
    sTERA_PCOIP_HMI_LOCAL_VAR  apdu;

    // assign a pointer to the input buffer
    buf = (sTERA_PCOIP_HMI_LOCAL_VAR *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        switch (apdu.var)
        {
            case TERA_PCOIP_HMI_LOCAL_VAR_HOST_SESS_STATE:
            {
                cblk->host_sess_state = apdu.data.sess_state.host_sess_state;
                mTERA_LOG_DBG("HMI_LOCAL_VAR_SESS_STATE host state=%d",
                              cblk->host_sess_state);
                ret = TERA_SUCCESS;
                break;
            }
            default:
            {
                mTERA_LOG_DBG("unsupported local HMI variable=%u", apdu.var);
                ret = TERA_ERR_NOT_SUPPORTED;
                break;
            }
        }
    }

    return ret;
} // hmi_ioctl_local_var_set_proc()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_set_avail_var_apdu_rep
//      Reply to a HMI_VAR_SET ioctl with HMI_AVAIL_VAR variable
//
//  Description:
//      This function replies to a HMI_VAR_SET ioctl from userspace with the
//      list of available HMI variables for setting.
//
static INT32 hmi_ioctl_var_set_avail_var_apdu_rep(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    )
{
    UINT32 ret = TERA_SUCCESS;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    if(!down_interruptible(&cblk->data_lock))
    {
        // Copy the list of HMI variables available for setting to the buffer
        if(copy_to_user(&buf->data.avail_var,
                        &cblk->hmi.avail_var_set,
                        sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HMI_VAR_SET HMI_AVAIL_VAR");
    }
    else
    {
        mTERA_LOG_DBG("set avail_var: 0x%08X 0x%08X 0x%08X 0x%08X",
                          cblk->hmi.avail_var_set.map[0],
                          cblk->hmi.avail_var_set.map[1],
                          cblk->hmi.avail_var_set.map[2],
                          cblk->hmi.avail_var_set.map[3]);
        mTERA_LOG_DBG("               0x%08X 0x%08X 0x%08X 0x%08X",
                          cblk->hmi.avail_var_set.map[4],
                          cblk->hmi.avail_var_set.map[5],
                          cblk->hmi.avail_var_set.map[6],
                          cblk->hmi.avail_var_set.map[7]);
    }

    return ret;
} // hmi_ioctl_var_set_avail_var_apdu_rep()

// --------------------------------------------------------------------------
//  hmi_ioctl_var_set_fwd_to_blade
//      Forward a HMI_VAR_SET ioctl to the blade
//
//  Description:
//      This function forwards a HMI_VAR_SET ioctl to the blade. The input
//      buffer of the ioctl is a HMI_VAR_SET APDU. The buffer is copied
//      to a packet and sent down to the blade. The ioctl is then put into the
//      pending ioctl queue (to be completed when a response comes back from
//      the blade).
//
static INT32 hmi_ioctl_var_set_fwd_to_blade(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *buf
    )
{
    INT32             ret;
    sHCC_HMI_APDU_VAR apdu;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // Fill out the APDU header
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET;

    // Fill out the APDU data
    if (copy_from_user(&apdu.data, buf, sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU
            ret = hmi_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                // clear flag that is inspected when the workqueue is inspected
                cblk->pending_ioctl_flag = 0;

                // wait until the response is received on the ingress message queue and
                // processed by the dpc_entry
                if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                     (cblk->pending_ioctl_flag == 1),
                                                     msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
                {
                    mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                    up(&cblk->pending_ioctl_lock);
                    cblk->pending_ioctl_flag = 1;
                    ret = -ERESTARTSYS;
                }
                else
                {
                    // fill in the IOCTL response
                    if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
                    {
                        mTERA_LOG_ERR("copy_to_user failed");
                        ret = -ERESTARTSYS;
                    }
                    else
                    {
                        ret = TERA_SUCCESS;
                    }
                    up(&cblk->pending_ioctl_lock);
                }
            }
        }
    }
    return ret;
} // hmi_ioctl_var_set_fwd_to_blade()

// --------------------------------------------------------------------------
//  hmi_avail_cmd_apdu_send
//      Send a HMI_AVAIL_CMD APDU
//
//  Description:
//      This function sends a HMI_AVAIL_CMD APDU to the blade to query the
//      available HMI commands.
//
static void hmi_avail_cmd_apdu_send(
    sHCC_CBLK *cblk
    )
{
    sHCC_HMI_APDU_AVAIL_CMD apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU header
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_AVAIL_CMD;

    // Send the APDU
    hmi_apdu_send(cblk, &apdu, sizeof(apdu));
} // hmi_avail_cmd_apdu_send()

// --------------------------------------------------------------------------
//  hmi_avail_cmd_apdu_proc
//      Process a HMI_AVAIL_CMD APDU
//
//  Description:
//      This function processes a HMI_AVAIL_CMD APDU from the blade. The APDU
//      contains the supported HMI commands. If the HMI_VAR_GET and/or
//      HMI_VAR_SET commands are available, then this function queries the
//      blade for available HMI variables for getting/setting.
//
static void hmi_avail_cmd_apdu_proc(
    sHCC_CBLK *cblk,
    void      *data,
    UINT32     data_len
    )
{
    sTERA_PCOIP_HCC_APDU_HMI_MAP *avail_cmd = data;
    UINT32                        i;

    if ((data == NULL) ||
        (data_len != sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP)))
    {
        mTERA_LOG_ERR("incorrect APDU data length=%u (expected=%u)",
                      data_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));
    }
    else
    {
        if (avail_cmd->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET))
        {
            // Get available HMI variables that we can get
            hmi_var_get_avail_var_apdu_send(cblk);
        }

        if (avail_cmd->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET))
        {
            // Get available HMI variables that we can set
            hmi_var_set_avail_var_apdu_send(cblk);
        }

        down(&cblk->hmi.data_lock);

        // Store the new feature map
        memcpy(&cblk->hmi.avail_cmd, avail_cmd,
               sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));

        up(&cblk->hmi.data_lock);

        mTERA_LOG_INFO("avail_cmd[0] '0x%08x'", avail_cmd->map[0]);
        mTERA_LOG_INFO("avail_cmd[0] '0x%08x'", cblk->hmi.avail_cmd.map[0]);
        mTERA_LOG_INFO("***************************************");
        for (i = 0; i < TERA_PCOIP_HCC_APDU_HMI_CMD_MAX; i++)
        {
            if (avail_cmd->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
                mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
            {
                mTERA_LOG_INFO("%s", hmi_cmd_str_tbl[i]);
            }
        }
        mTERA_LOG_INFO("***************************************");
    } // end if
} // hmi_avail_cmd_apdu_proc()

// --------------------------------------------------------------------------
//  hmi_var_get_avail_var_apdu_send
//      Send a HMI_VAR_GET APDU with HMI_VAR_AVAIL variable
//
//  Description:
//      This function sends a HMI_VAR_GET APDU to the blade to query which HMI
//      variables are available for getting.
//
static void hmi_var_get_avail_var_apdu_send(
    sHCC_CBLK *cblk
    )
{
    sHCC_HMI_APDU_VAR apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET;
    apdu.data.var = TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR;

    mTERA_LOG_DBG("send hmi var get avail var");

    // Send the APDU
    hmi_apdu_send(cblk, &apdu, sizeof(apdu));
} // hmi_var_get_avail_var_apdu_send()

// --------------------------------------------------------------------------
//  hmi_var_set_avail_var_apdu_send
//      Send a HMI_VAR_SET APDU with HMI_VAR_AVAIL variable
//
//  Description:
//      This function sends a HMI_VAR_SET APDU to the blade to query which HMI
//      variables are available for setting.
//
static void hmi_var_set_avail_var_apdu_send(
    sHCC_CBLK *cblk
    )
{
    sHCC_HMI_APDU_VAR apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(sHCC_HMI_APDU_VAR));

    // Fill out the APDU
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET;
    apdu.data.var = TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR;

    mTERA_LOG_DBG("send hmi var set avail var");

    // Send the APDU
    hmi_apdu_send(cblk, &apdu, sizeof(apdu));
} // hmi_var_set_avail_var_apdu_send()

// --------------------------------------------------------------------------
//  hmi_var_get_apdu_proc
//      Process a HMI_VAR_GET APDU
//
//  Description:
//      This function processes a HMI_VAR_GET APDU from the blade. This
//      function determines which HMI variable the APDU contains and carries
//      out the correct processing.
//
static void hmi_var_get_apdu_proc(
    sHCC_CBLK *cblk,
    void      *data,
    UINT32     data_len
    )
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep = data;

    ASSERT(cblk != NULL);

    if ((data == NULL) || (data_len != sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
    {
        mTERA_LOG_ERR("incorrect APDU data length=%u (expected=%u)",
                      data_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR));
    }
    else
    {
        switch (rep->var)
        {
            case TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR:
            {
                mTERA_LOG_DBG("rcvd HMI_VAR_AVAIL_VAR");

                hmi_apdu_var_get_avail_var_proc(cblk, &rep->data.avail_var);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO:
            {
                mTERA_LOG_DBG("rcvd link info: state=%u, rate=%u, flags=0x%02x, mac=%s",
                              rep->data.link_info.state,
                              rep->data.link_info.rate,
                              rep->data.link_info.flags,
                              rep->data.link_info.mac);

                // Forward the reply to userspace
                hmi_apdu_var_get_fwd_to_user(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG:
            {
                mTERA_LOG_DBG("rcvd HMI_VAR_IPV4_CFG");

                mTERA_LOG_DBG("ip:      '%s'", rep->data.ipv4_cfg.ip);
                mTERA_LOG_DBG("mask:    '%s'", rep->data.ipv4_cfg.mask);
                mTERA_LOG_DBG("gw:      '%s'", rep->data.ipv4_cfg.gw);
                mTERA_LOG_DBG("pri_dns: '%s'", rep->data.ipv4_cfg.pri_dns);
                mTERA_LOG_DBG("sec_dns: '%s'", rep->data.ipv4_cfg.sec_dns);

                // Forward the reply to userspace
                hmi_apdu_var_get_fwd_to_user(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK:
            {
                mTERA_LOG_DBG("rcvd sess_link: flags=0x%02x, ip=%s, mac=%s",
                              rep->data.sess_link.flags,
                              rep->data.sess_link.ip,
                              rep->data.sess_link.mac);

                // Process the session link information
                hmi_apdu_var_get_sess_link_proc(cblk, &rep->data.sess_link);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT:
            {
                mTERA_LOG_DBG("rcvd conn_port: port=%u", rep->data.conn_port.port);

                // Forward the reply to userspace
                hmi_apdu_var_get_fwd_to_user(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT:
            {
                mTERA_LOG_DBG("rcvd HMI_VAR_SESS_STAT");

                // Forward the reply to userspace
                hmi_apdu_var_get_fwd_to_user(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG:
            {
                mTERA_LOG_DBG("rcvd WOL cfg: flags=0x%02x, mac=%s ip=%s",
                              rep->data.wol_cfg.flags,
                              rep->data.wol_cfg.mac,
                              rep->data.wol_cfg.ip);

                // Forward the reply to userspace
                hmi_apdu_var_get_fwd_to_user(cblk, rep);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unknown HMI variable=%u", rep->var);
                break;
            }
        }
    } // end if
} // hmi_var_get_apdu_proc()

// --------------------------------------------------------------------------
//  hmi_apdu_var_get_avail_var_proc
//      Proccess a HMI_VAR_GET APDU with HMI_AVAIL_VAR variable
//
//  Description:
//      This function processes a HMI_VAR_GET APDU from the blade with the
//      available HMI variables for getting.
//
static void hmi_apdu_var_get_avail_var_proc(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_MAP *avail_var
    )
{
    UINT32 i;

    ASSERT(cblk != NULL);

    // Get session link information if variable is available for the first time
    if (!(cblk->hmi.avail_var_get.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)] &
          mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)) &&
        (avail_var->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)] &
         mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)))
    {
        //
        // This should be the only place that sends a session link information
        // request APDU because we want the session link information replies to
        // be controlled.
        //
        hmi_var_get_sess_link_apdu_send(cblk);
    }

    down(&cblk->hmi.data_lock);

    // Store the available HMI variables
    memcpy(&cblk->hmi.avail_var_get, avail_var,
           sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));

    up(&cblk->hmi.data_lock);

    mTERA_LOG_INFO("***************************************");
    for (i = 0; i < TERA_PCOIP_HCC_APDU_HMI_VAR_MAX; i++)
    {
        if (avail_var->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
        {
            mTERA_LOG_INFO("%s", hmi_var_str_tbl[i]);
        }
    }
    mTERA_LOG_INFO("***************************************");
} // hmi_apdu_var_get_avail_var_proc()

// --------------------------------------------------------------------------
//  hmi_var_get_sess_link_apdu_send
//      Send a HMI_VAR_GET APDU with HMI_SESS_LINK variable
//
//  Description:
//      This function sends a HMI_VAR_GET APDU to the blade to get the session
//      link information variable.
//
static void hmi_var_get_sess_link_apdu_send(
    sHCC_CBLK *cblk
    )
{
    sHCC_HMI_APDU_VAR apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    apdu.hdr.cmd = TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET;
    apdu.data.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK;

    mTERA_LOG_DBG("send hmi get sess_link");

    // Send the APDU
    hmi_apdu_send(cblk, &apdu, sizeof(apdu));
} // hmi_var_get_sess_link_apdu_send()

// --------------------------------------------------------------------------
//  hmi_apdu_var_get_sess_link_proc
//      Process a HMI_VAR_GET APDU with HMI_SESS_LINK variable
//
//  Description:
//      This function processes a HMI_VAR_GET APDU from the blade with the
//      session link information. The session link information is cached in the
//      control block. All HMI_SESS_LINK_NOTIFY ioctls in the pending ioctl
//      queue will be completed to notify users in userspace of a session link
//      information change.
//
static void hmi_apdu_var_get_sess_link_proc(
    sHCC_CBLK                              *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link
    )
{
    UINT8      transition_cnt;

    down(&cblk->hmi.data_lock);

    //
    // Save the current transition count - we'll write it back into the
    // control block afterwards because the transition count is a driver-owned
    // variable
    //
    transition_cnt = cblk->hmi.sess_link.transition_cnt;

    // Look for session link transitions
    if (!(cblk->hmi.sess_link.flags &
          TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) &&
        (sess_link->flags &
         TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
    {
        // Detected a disconnected->connected transistion
        transition_cnt++;
    }
    else if ((cblk->hmi.sess_link.flags &
              TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) &&
             !(sess_link->flags &
               TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
    {
        // Detected a connected->disconnected transistion
        transition_cnt++;
    }

    // Store the new session info in the control block
    memcpy(&cblk->hmi.sess_link, sess_link,
           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK));

    // Restore the transition count
    cblk->hmi.sess_link.transition_cnt = transition_cnt;

    up(&cblk->hmi.data_lock);
} // hmi_apdu_var_get_sess_link_proc()

// --------------------------------------------------------------------------
//  hmi_apdu_var_get_fwd_to_user
//      Forward a HMI_VAR_GET APDU to userspace
//
//  Description:
//      This function forwards a HMI_VAR_GET APDU to userspace. The APDU is
//      copied to the output buffer of an ioctl in the pending ioctl queue. The
//      ioctl is then completed to send the APDU up to userspace.
//
static void hmi_apdu_var_get_fwd_to_user(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep
    )
{
    down(&cblk->hmi.data_lock);

    // Store the response in the control block
    memcpy(&cblk->ioctl_rsp.data.get_rep, rep,
           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR));

    up(&cblk->hmi.data_lock);

    // set the flag that will wake up the waiting ioctl thread
    cblk->pending_ioctl_flag = 1;

    // wake up the waiting thread
    wake_up_interruptible(&cblk->pending_ioctl_q);
} // hmi_apdu_var_get_fwd_to_user()

// --------------------------------------------------------------------------
//  hmi_var_set_apdu_proc
//      Process a HMI_VAR_SET APDU
//
//  Description:
//      This function processes a HMI_VAR_SET APDU from the blade. This
//      function determines which HMI variable the APDU contains and carries
//      out the correct processing.
//
static void hmi_var_set_apdu_proc(
    sHCC_CBLK *cblk,       // Control block
    void      *data,       // APDU data
    UINT32     data_len    // APDU data length
    )
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep = data;

    ASSERT(cblk != NULL);

    if ((data == NULL) || (data_len != sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR)))
    {
        mTERA_LOG_ERR("incorrect APDU data length=%u (expected=%u)",
                      data_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR));
    }
    else
    {
        switch (rep->var)
        {
            case TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR:
            {
                hmi_var_set_avail_var_apdu_proc(cblk, &rep->data.avail_var);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK:
            {
                mTERA_LOG_DBG("rcvd sess_link: flags='0x%02x', ip=%s, mac=%s",
                              rep->data.sess_link.flags,
                              rep->data.sess_link.ip,
                              rep->data.sess_link.mac);

                // Forward the reply to userspace
                hmi_var_set_fwd_to_user_apdu(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT:
            {
                mTERA_LOG_DBG("rcvd sess_stat");

                // Forward the reply to userspace
                hmi_var_set_fwd_to_user_apdu(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG:
            {
                mTERA_LOG_DBG("rcvd WOL cfg: flags=0x%02x, mac=%s ip=%s",
                              rep->data.wol_cfg.flags,
                              rep->data.wol_cfg.mac,
                              rep->data.wol_cfg.ip);

                // Forward the reply to userspace
                hmi_var_set_fwd_to_user_apdu(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC:
            {
                mTERA_LOG_DBG("rcvd sess disc: reason=0x%x",
                              rep->data.sess_disc.reason_cause);

                // Forward the reply to userspace
                hmi_var_set_fwd_to_user_apdu(cblk, rep);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE:
            {
                mTERA_LOG_DBG("rcvd unicode: enabled=%d",
                              rep->data.unicode.enabled);

                // Forward the reply to userspace
                hmi_var_set_fwd_to_user_apdu(cblk, rep);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unsupported HMI variable=%u", rep->var);
                break;
            }
        }
    } // end if
} // hmi_var_set_apdu_proc()

// --------------------------------------------------------------------------
//  hmi_var_set_avail_var_apdu_proc
//      Proccess a HMI_VAR_SET APDU with HMI_AVAIL_VAR variable
//
//  Description:
//      This function processes a HMI_VAR_SET APDU from the blade with the
//      available HMI variables for setting.
//
static void hmi_var_set_avail_var_apdu_proc(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_MAP *avail_var
    )
{
    UINT32 i;

    ASSERT(cblk != NULL);

    down(&cblk->hmi.data_lock);

    // Store the available HMI variables
    memcpy(&cblk->hmi.avail_var_set, avail_var,
           sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));

    up(&cblk->hmi.data_lock);

    mTERA_LOG_INFO("HMI_VAR_AVAIL_VAR");
    mTERA_LOG_INFO("***************************************");
    for (i = 0; i < TERA_PCOIP_HCC_APDU_HMI_VAR_MAX; i++)
    {
        if (avail_var->map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
        {
            mTERA_LOG_INFO("%s", hmi_var_str_tbl[i]);
        }
    }
    mTERA_LOG_INFO("***************************************");
} // hmi_var_set_avail_var_apdu_proc()

// --------------------------------------------------------------------------
//  hmi_var_set_fwd_to_user_apdu
//      Forward a HMI_VAR_SET APDU to userspace
//
//  Description:
//      This function fowards a HMI_VAR_SET APDU to userspace. A HMI_VAR_SET
//      ioctl, with the same HMI variable as in the APDU, in the pending ioctl
//      queue is completed by this function. This signifies a successful HMI
//      variable set.
//
static void hmi_var_set_fwd_to_user_apdu(
    sHCC_CBLK                    *cblk,
    sTERA_PCOIP_HCC_APDU_HMI_VAR *rep
    )
{
    down(&cblk->hmi.data_lock);

    // Store the response in the control block
    memcpy(&cblk->ioctl_rsp.data.set_req, rep,
           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR));

    up(&cblk->hmi.data_lock);

    // set the flag that will wake up the waiting ioctl thread
    cblk->pending_ioctl_flag = 1;

    // wake up the waiting thread
    wake_up_interruptible(&cblk->pending_ioctl_q);
} // hmi_var_set_fwd_to_user_apdu()

// --------------------------------------------------------------------------
//  hcc_hmi_init
//      Initialize the HMI user
//
//  Description:
//      This function initializes the HMI user. This function is responsible
//      for initializing any and all user-specific objects.
//
INT32 hcc_hmi_init(
    sHCC_CBLK *cblk    // HCC control block
    )
{
    INT32   ret = TERA_SUCCESS;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // Create a spinlock to protect local data
    init_MUTEX(&cblk->hmi.data_lock);

    mTERA_LOG_DBG("end");

    return ret;
} // hcc_hmi_init()

// --------------------------------------------------------------------------
//  hcc_hmi_start
//      Start the HCC HMI user portion
//
//  Description:
//      This function starts the HCC HMI user portion.
//
INT32 hcc_hmi_start(
    sHCC_CBLK *cblk
    )
{

    // Reset the lists of HMI available commands and variables
    hmi_avail_cmd_var_reset(cblk);

    return TERA_SUCCESS;
} // hcc_hmi_start()

// --------------------------------------------------------------------------
//  hcc_hmi_stop
//      Stop the KMP user
//
//  Description:
//      This function stops the HMI user. This function is responsible
//      for stopping any and all user-specific objects, events, and threads.
//
void hcc_hmi_stop(
    sHCC_CBLK *cblk
    )
{
    UINT8 transition_cnt;

    ASSERT(cblk != NULL);

    //
    // If we are still in the connected state, then we will assume that the
    // session will disconnect while we are in the stopped state.
    //
    if (cblk->hmi.sess_link.flags &
        TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED)
    {
        // Save the transition count
        transition_cnt = cblk->hmi.sess_link.transition_cnt;

        // We do a manual transition here to disconnected state
        memset(&cblk->hmi.sess_link, 0, sizeof(cblk->hmi.sess_link));

        // Going from connected->disconnected
        transition_cnt++;

        // Record the transition count back into the control block copy
        cblk->hmi.sess_link.transition_cnt = transition_cnt;
    }
} // hcc_hmi_stop()

// --------------------------------------------------------------------------
//  hcc_hmi_ioctl_proc
//      Process a HMI ioctl request
//
//  Description:
//      This function processes a HMI ioctl request. If the ioctl is not
//      supported, then it is completed with STATUS_NOT_SUPPORTED.
//
INT32 hcc_hmi_ioctl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32           ret;

    ASSERT(cblk != NULL);

    switch (ioctl)
    {
        case TERA_PCOIP_HCC_IOCTL_HMI_AVAIL_CMD:
        {
            // Send the user the list of available commands
            ret = hmi_ioctl_avail_cmd_apdu_rep(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET:
        {
            ret = hmi_ioctl_var_get_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET:
        {
            ret = hmi_ioctl_var_set_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_HMI_LOCAL_VAR_GET:
        {
            ret = hmi_ioctl_local_var_get_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_HMI_LOCAL_VAR_SET:
        {
            ret = hmi_ioctl_local_var_set_proc(cblk, arg);
            break;
        }
        default:
        {
            mTERA_LOG_DBG("unsupported ioctl=0x%08x", ioctl);
            ret = TERA_ERR_NOT_SUPPORTED;
            break;
        }
    }
    return ret;
} // hcc_hmi_ioctl_proc()

// --------------------------------------------------------------------------
//  hcc_hmi_apdu_proc
//      Process an HMI APDU
//
//  Description:
//      This function processes an HMI APDU from the ingress descriptor
//      queue. Only known APDUs are serviced.
//
//      NOTE: The APDU is freed by the calling function after this function
//      returns, so it should NOT be referenced for future use.
//
void hcc_hmi_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sHCC_HMI_APDU_HDR *hdr = apdu;
    void              *data = mHMI_APDU_DATA_PTR_GET(apdu);
    UINT32             data_len;

    ASSERT(cblk != NULL);
    ASSERT(apdu != NULL);

    if (apdu_len < sizeof(sHCC_HMI_APDU_HDR))
    {
        mTERA_LOG_ERR("invalid APDU length");
    }
    else if (cblk->initialized)
    {
        // Calculate the data length
        data_len = apdu_len - sizeof(sHCC_HMI_APDU_HDR);

        switch (hdr->cmd)
        {
            case TERA_PCOIP_HCC_APDU_HMI_CMD_AVAIL_CMD:
            {
                hmi_avail_cmd_apdu_proc(cblk, data, data_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET:
            {
                hmi_var_get_apdu_proc(cblk, data, data_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET:
            {
                hmi_var_set_apdu_proc(cblk, data, data_len);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unsupported HMI command=%u", hdr->cmd);
                break;
            }
        }
    } // end if
} // hcc_hmi_apdu_proc()

// --------------------------------------------------------------------------
//  hmi_avail_cmd_var_reset
//      Reset the list of available HMI commands and variables
//
//  Description:
//      This function resets the list of available HMI commands and variables.
//      Only the commands available request and reply APDUs will be
//      advertised.
//
static void hmi_avail_cmd_var_reset(
    sHCC_CBLK *cblk
    )
{
    ASSERT(cblk != NULL);

    // Reset the list of available HMI commands
    memset(&cblk->hmi.avail_cmd, 0, sizeof(cblk->hmi.avail_cmd));

    // Reset the list of available HMI variables for getting/setting
    memset(&cblk->hmi.avail_var_get, 0, sizeof(cblk->hmi.avail_var_get));
    memset(&cblk->hmi.avail_var_set, 0, sizeof(cblk->hmi.avail_var_set));
} // hmi_avail_cmd_var_reset()

// --------------------------------------------------------------------------
//  hcc_hmi_scp_chan_open_proc
//      Process a SCP channel open event
//
//  Description:
//      This function processes a SCP channel open event. This function sends
//      out a request to get the available HMI commands from the firmware.
//
void hcc_hmi_scp_chan_open_proc(
    sHCC_CBLK *cblk
    )
{
    ASSERT(cblk != NULL);

    mTERA_LOG_DBG("send hmi avail cmd apdu");

    // Get the available HMI commands
    hmi_avail_cmd_apdu_send(cblk);

} // hcc_hmi_scp_chan_open_proc()

