////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     HCC Module IMG User Implementation
///
///     This file implements the HCC module IMG user. All IMG user related
///     APDUs and ioctls are dealt with in this file. All byte-swapping that
///     needs to be done for inbound and outbound APDUs are all handled here.
///     Applications in userspace do not need to worry about it.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"

#include "tera_hcc.h"
#include "tera_hcc_internal.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_bstream.h"
#include "tera_tlv.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

static INT32 img_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static INT32 img_ioctl_host_topology_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

static INT32 img_ioctl_host_topology_apdu_send_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    );

static INT32 img_ioctl_async_host_topology_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    );

static INT32 img_ioctl_cache_topology_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

static INT32 img_ioctl_cache_topology_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    );

static INT32 img_ioctl_user_cfg_req_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

static INT32 img_ioctl_user_cfg_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    );

static INT32 img_ioctl_user_cfg_cache_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    );

static void img_host_topology_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void img_cache_topology_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void img_host_user_cfg_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void img_host_user_cfg_req_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void img_user_settings_encode(
    UINT32  image_qual_pref,
    UINT8   user_cfg_tlv_bstream[MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES],
    UINT32 *num_bytes_written
    );

static void img_user_settings_decode(
    sTERA_PCOIP_HCC_IMG_MODE  *current_mode,
    UINT8                     *user_cfg_tlv_bstream,
    UINT32                     user_cfg_tlv_bstream_length,
    sTERA_PCOIP_HCC_IMG_MODE  *new_mode
    );

// --------------------------------------------------------------------------
//  img_apdu_send
//      Send a IMG APDU
//
//  Description:
//      This function sends a IMG APDU to the blade. The contents of the APDU
//      are copied into an allocated packet and enqueued onto the egress
//      packet queue. The calling function can safely discard the APDU after
//      this function call.
//
static INT32 img_apdu_send(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    // Send the APDU
    ret = tera_scp_data_pkt_send(cblk->dev, cblk->scp_handle,
                                 TERA_PCOIP_HCC_APDU_USER_IMG,
                                 &cblk->egress_pkt_q, apdu, apdu_len);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
        ret = -ERESTARTSYS;
    }

    return ret;
} // img_apdu_send()

// --------------------------------------------------------------------------
//  img_ioctl_host_topology_apdu_get_proc
//      Process a IMG_DISP_TOPO_GET ioctl
//
//  Description:
//      This function processes a IMG_DISP_TOPO_GET ioctl from userspace. This
//      function creates and sends an APDU to the firmware to get display
//      topology that it has. The ioctl is then put into the pending ioctl
//      queue (to be completed when a response comes back from the blade).
//
static INT32 img_ioctl_host_topology_apdu_get_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ *buf;
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ  apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HCC_APDU_IMG_TYPE_DISP_TOPO_REQ);

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = img_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                up(&cblk->pending_ioctl_lock);
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                buf = (sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ *)arg;
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(apdu)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // img_ioctl_host_topology_apdu_get_proc()

// --------------------------------------------------------------------------
//  img_ioctl_host_topology_apdu_send_proc
//      Process a IMG_DISP_TOPO_SEND ioctl
//
//  Description:
//      This function processes a IMG_DISP_TOPO_SEND ioctl from userspace. This
//      function copies the display topology from the ioctl's input buffer into
//      an APDU and sends it down to the firmware. The ioctl is completed in
//      this function.
//
static INT32 img_ioctl_host_topology_apdu_send_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg
    )
{
    INT32                               ret;
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO *buf;
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO  apdu;
    UINT32                              i;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO *)arg;
    if (copy_from_user(&apdu, buf, sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.pri_num = mTERA_UINT32_BYTE_SWAP(apdu.pri_num);

        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%d, height=%d, rotation=%d",
                          i, apdu.disp[i].pos_x, apdu.disp[i].pos_y,
                          apdu.disp[i].width, apdu.disp[i].height,
                          apdu.disp[i].rotation);

            apdu.disp[i].pos_x = mTERA_UINT32_BYTE_SWAP(apdu.disp[i].pos_x);
            apdu.disp[i].pos_y = mTERA_UINT32_BYTE_SWAP(apdu.disp[i].pos_y);
            apdu.disp[i].width = mTERA_UINT16_BYTE_SWAP(apdu.disp[i].width);
            apdu.disp[i].height = mTERA_UINT16_BYTE_SWAP(apdu.disp[i].height);
            apdu.disp[i].rotation = mTERA_UINT32_BYTE_SWAP(apdu.disp[i].rotation);
        }

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = img_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // img_ioctl_host_topology_apdu_send_proc()

// --------------------------------------------------------------------------
//  img_ioctl_async_host_topology_apdu_get_proc
//      Process a IMG_ASYNC_DISP_TOPO_GET ioctl
//
//  Description:
//      This function processes a IMG_ASYNC_DISP_TOPO_GET ioctl from userspace.
//      This returns a copy of the last received asynchronous display topology
//      message from the FW. It also returns an ID which changes each time a
//      new asynchronous display topology message is received.
//
static INT32 img_ioctl_async_host_topology_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                                    ret = TERA_SUCCESS;
    sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO *buf;
    INT                                       i;

    ASSERT(cblk != NULL);

    buf = (sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO *)arg;

    if(!down_interruptible(&cblk->data_lock))
    {
        // Copy the information from the control block to the buffer
        buf->msg_id = cblk->id_disp_topo;
        if(copy_to_user(&buf->apdu,
                        &cblk->async_disp_topo,
                        sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HCC_IOCTL_IMG_ASYNC_DISP_TOPO_GET");
    }
    else
    {
        // Dump the data inside the APDU
        mTERA_LOG_DBG("id=%u, apdu=0x%08X, pri=%d",
                      cblk->id_disp_topo,
                      cblk->async_disp_topo.apdu_type,
                      cblk->async_disp_topo.pri_num);
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rotation=%u",
                          i,
                          cblk->async_disp_topo.disp[i].pos_x,
                          cblk->async_disp_topo.disp[i].pos_y,
                          cblk->async_disp_topo.disp[i].width,
                          cblk->async_disp_topo.disp[i].height,
                          cblk->async_disp_topo.disp[i].rotation);
        }
    }

    return ret;
} // img_ioctl_async_host_topology_apdu_get_proc()

// --------------------------------------------------------------------------
//  img_ioctl_cache_topology_apdu_get_proc
//      Process a IMG_CACHE_DISP_TOPO_GET ioctl
//
//  Description:
//      This function processes a IMG_CACHE_DISP_TOPO_GET ioctl from
//      userspace. This function creates and sends an APDU to the firmware to
//      read the client cached display topology settings. The ioctl is then
//      put into the pending ioctl queue (to be completed when a response comes
//      back from the blade).
//
static INT32 img_ioctl_cache_topology_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32                                     ret;
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO *buf;
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO  apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO *)arg;
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_GET);
    apdu.pri_num = 0;

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = img_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                up(&cblk->pending_ioctl_lock);
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(apdu)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // img_ioctl_cache_topology_apdu_get_proc()

// --------------------------------------------------------------------------
//  img_ioctl_cache_topology_apdu_set_proc
//      Process a IMG_CACHE_DISP_TOPO_SET ioctl
//
//  Description:
//      This function processes a IMG_CACHE_DISP_TOPO_SET ioctl from userspace.
//      This function copies the display topology from the ioctl's input buffer
//      into an APDU and sends it down to the firmware. The APDU is sent to the
//      client device and the settings are stored in the client display
//      topology settings. This command should only be sent to zero clients
//      which support it.The ioctl is completed in this function.
//
//  Notes:
//      This APDU is useful when trying to configure topology settings not
//      supported by the client OSD, such as fine grained adjustment of the
//      origin settings of each display. Users have more flexibility/control
//      using the Windows settings. The Windows interface also allows users to
//      select the primary display, which is the display at located at origin
//      (0,0).
//
static INT32 img_ioctl_cache_topology_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                     ret;
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO *buf;
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO  apdu;
    UINT32                                    i;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO *)arg;
    if (copy_from_user(&apdu, buf, sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);

        mTERA_LOG_DBG("APDU type=0x%08X, pri=%d, primary=%d",
                      apdu.apdu_type,
                      apdu.pri_num,
                      apdu.topo.primary);
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rot=%u, rate=%u",
                          i, apdu.topo.topology[i].pos_x,
                          apdu.topo.topology[i].pos_y,
                          apdu.topo.topology[i].width,
                          apdu.topo.topology[i].height,
                          apdu.topo.topology[i].rotation,
                          apdu.topo.refresh_rate[i]);
            mTERA_LOG_DBG("        name=%s, serial=%s, vid=%s, pid=%s",
                          apdu.iden[i].name,
                          apdu.iden[i].serial,
                          apdu.iden[i].vid,
                          apdu.iden[i].pid);
        }

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = img_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // img_ioctl_cache_topology_apdu_set_proc()

// --------------------------------------------------------------------------
//  img_ioctl_user_cfg_req_apdu_get_proc
//      Process a IMG_USER_CFG_REQ_GET ioctl
//
//  Description:
//      This function processes a IMG_USER_CFG_REQ_GET ioctl from
//      userspace. This function creates and sends an APDU to the firmware to
//      read the imaging user config settings. The ioctl is then put into the
//      pending ioctl queue (to be completed when a response comes back from
//      the blade).
//
static INT32 img_ioctl_user_cfg_req_apdu_get_proc(
    sHCC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32                                            ret;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA      *buf;
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA *)arg;
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_REQ);
    apdu.pri_num = 0;

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = img_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                up(&cblk->pending_ioctl_lock);
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    mTERA_LOG_DBG("qual=%d, b2l=%d",
                                  cblk->ioctl_rsp.data.usr_cfg.user_cfg_settings.image_qual_pref,
                                  cblk->ioctl_rsp.data.usr_cfg.user_cfg_settings.build_to_lossless_enable);
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // img_ioctl_user_cfg_req_apdu_get_proc()

// --------------------------------------------------------------------------
//  img_ioctl_user_cfg_apdu_set_proc
//      Process a IMG_USER_CFG_SET ioctl
//
//  Description:
//      This function processes a IMG_USER_CFG_SET ioctl from userspace.
//      This function copies the imaging user config settings from the
//      ioctl's input buffer into an APDU and sends it to the firmware.
//      The ioctl is completed in this function.
//
static INT32 img_ioctl_user_cfg_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                                   ret;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA             *buf;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA              buf_data;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_TX_PARAMS   apdu;
    UINT32                                                  user_cfg_tlv_bstream_length;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA *)arg;
    if (copy_from_user(&buf_data, buf, sizeof(buf_data)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        apdu.host_user_cfg_req.apdu_type = mTERA_UINT32_BYTE_SWAP(buf_data.host_user_cfg_req.apdu_type);
        apdu.host_user_cfg_req.pri_num = buf_data.host_user_cfg_req.pri_num;
        img_user_settings_encode(buf_data.user_cfg_settings.image_qual_pref,
                                 &apdu.user_cfg_tlv_bstream[0],
                                 &user_cfg_tlv_bstream_length);
        user_cfg_tlv_bstream_length += sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_RX_PARAMS);

        mTERA_LOG_DBG("APDU type=0x%08X, pri=%d, len=%d",
                      apdu.host_user_cfg_req.apdu_type,
                      apdu.host_user_cfg_req.pri_num,
                      user_cfg_tlv_bstream_length);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = img_apdu_send(cblk, &apdu, user_cfg_tlv_bstream_length);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // img_ioctl_user_cfg_apdu_set_proc()

// --------------------------------------------------------------------------
//  img_ioctl_user_cfg_cache_apdu_set_proc
//      Process a IMG_USER_CFG_CACHE_SET ioctl
//
//  Description:
//      This function processes a IMG_USER_CFG_CACHE_SET ioctl from userspace.
//      This function copies the imaging user config cache settings from the
//      ioctl's input buffer into an APDU and sends it to the firmware.
//      The APDU is sent to the client device and the current image settings
//      that are in use are stored in permanently in the client image user
//      settings. The ioctl is completed in this function.
//
static INT32 img_ioctl_user_cfg_cache_apdu_set_proc(
    sHCC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                                   ret;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA       *buf;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA        buf_data;
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_TX_PARAMS   apdu;
    UINT32                                                  user_cfg_tlv_bstream_length;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA *)arg;
    if (copy_from_user(&buf_data, buf, sizeof(buf_data)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        apdu.host_user_cfg_req.apdu_type = mTERA_UINT32_BYTE_SWAP(buf_data.host_user_cfg_req.apdu_type);
        apdu.host_user_cfg_req.pri_num = buf_data.host_user_cfg_req.pri_num;
        img_user_settings_encode(buf_data.image_qual_pref,
                                 &apdu.user_cfg_tlv_bstream[0],
                                 &user_cfg_tlv_bstream_length);
        user_cfg_tlv_bstream_length += sizeof(sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_RX_PARAMS);

        mTERA_LOG_DBG("APDU type=0x%08X, pri=%d, len=%d",
                      apdu.host_user_cfg_req.apdu_type,
                      apdu.host_user_cfg_req.pri_num,
                      user_cfg_tlv_bstream_length);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = img_apdu_send(cblk, &apdu, user_cfg_tlv_bstream_length);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // img_ioctl_user_cfg_cache_apdu_set_proc()

// --------------------------------------------------------------------------
//  img_host_topology_apdu_proc
//      Process a display topology APDU
//
//  Description:
//      This function processes a display topology APDU from the firmware. The
//      APDU is either a response to a display topology request sent down to
//      the firmware, or if the display topology has changed. This function
//      looks for a pending IMG_DISP_TOPO_GET ioctl and copies the display
//      topology to the output buffer of the ioctl and completes it.
//
static void img_host_topology_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO *disp_topo = apdu;
    UINT32                              i;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO));
    }
    // if the correct IOCTL is not pending
    else if (cblk->pending_ioctl_flag != TERA_PCOIP_HCC_IOCTL_IMG_DISP_TOPO_GET)
    {
        // Dump the data inside the APDU
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rotation=%u",
                          i, disp_topo->disp[i].pos_x,
                          disp_topo->disp[i].pos_y,
                          disp_topo->disp[i].width,
                          disp_topo->disp[i].height,
                          disp_topo->disp[i].rotation);
        }

        // store a copy of the message and timestamp in the control block
        down(&cblk->hmi.data_lock);
        memcpy(&cblk->async_disp_topo, disp_topo, sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO));
        cblk->id_disp_topo = cblk->id_disp_topo + 1;
        up(&cblk->hmi.data_lock);

        // Ioctl could have been cancelled
        mTERA_LOG_INFO("no pending HCC_IOCTL_IMG_DISP_TOPO_GET");
    }
    else
    {
        down(&cblk->data_lock);

        // Store the response in the control block
        memcpy(&cblk->ioctl_rsp.data.img_disp_topo, disp_topo,
               sizeof(sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO));

        up(&cblk->data_lock);

        // Dump the data inside the APDU
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rotation=%u",
                          i, disp_topo->disp[i].pos_x,
                          disp_topo->disp[i].pos_y,
                          disp_topo->disp[i].width,
                          disp_topo->disp[i].height,
                          disp_topo->disp[i].rotation);
        }

        // set the flag that will wake up the waiting ioctl thread
        cblk->pending_ioctl_flag = 1;

        // wake up the waiting thread
        wake_up_interruptible(&cblk->pending_ioctl_q);
    } // end if
} // img_host_topology_apdu_proc()

// --------------------------------------------------------------------------
//  img_cache_topology_apdu_proc
//      Process a cache display topology APDU
//
//  Description:
//      This function processes a cache display topology get APDU from the
//      firmware. The APDU is a response to a get cache display topology
//      request sent down to the firmware. This function looks for a pending
//      IMG_CACHE_DISP_TOPO_GET ioctl and copies the display topology to the
//      output buffer of the ioctl and completes it.
//
static void img_cache_topology_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO *disp_topo = apdu;
    UINT32                                    i;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO));
    }
    // if the correct IOCTL is not pending
    else if (cblk->pending_ioctl_flag != TERA_PCOIP_HCC_IOCTL_IMG_CACHE_DISP_TOPO_GET)
    {
        // Dump the data inside the APDU
        mTERA_LOG_DBG("APDU type=0x%08X, pri num=%d, primary=%d",
                      disp_topo->apdu_type,
                      disp_topo->pri_num,
                      disp_topo->topo.primary);
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rot=%u, rate=%u",
                          i, disp_topo->topo.topology[i].pos_x,
                          disp_topo->topo.topology[i].pos_y,
                          disp_topo->topo.topology[i].width,
                          disp_topo->topo.topology[i].height,
                          disp_topo->topo.topology[i].rotation,
                          disp_topo->topo.refresh_rate[i]);
            mTERA_LOG_DBG("       name='%s', serial='%s', vid='%s', pid='%s'",
                          disp_topo->iden[i].name,
                          disp_topo->iden[i].serial,
                          disp_topo->iden[i].vid,
                          disp_topo->iden[i].pid);
        }

        // Ioctl could have been cancelled
        mTERA_LOG_INFO("no pending HCC_IOCTL_IMG_CACHE_DISP_TOPO_GET");
    }
    else
    {
        down(&cblk->data_lock);

        // Store the response in the control block
        memcpy(&cblk->ioctl_rsp.data.img_cache_topo_req, disp_topo,
                      sizeof(sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO));

        up(&cblk->data_lock);

        // Dump the data inside the APDU
        mTERA_LOG_DBG("APDU type=0x%08X, pri num=%d, primary=%d",
                      disp_topo->apdu_type,
                      disp_topo->pri_num,
                      disp_topo->topo.primary);
        for (i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("disp%u: pos_x=%d, pos_y=%d, width=%u, height=%u, rot=%u, rate=%u",
                          i, disp_topo->topo.topology[i].pos_x,
                          disp_topo->topo.topology[i].pos_y,
                          disp_topo->topo.topology[i].width,
                          disp_topo->topo.topology[i].height,
                          disp_topo->topo.topology[i].rotation,
                          disp_topo->topo.refresh_rate[i]);
            mTERA_LOG_DBG("       name='%s', serial='%s', vid='%s', pid='%s'",
                          disp_topo->iden[i].name,
                          disp_topo->iden[i].serial,
                          disp_topo->iden[i].vid,
                          disp_topo->iden[i].pid);
        }

        // set the flag that will wake up the waiting ioctl thread
        cblk->pending_ioctl_flag = 1;

        // wake up the waiting thread
        wake_up_interruptible(&cblk->pending_ioctl_q);
    } // end if
} // img_cache_topology_apdu_proc()

// --------------------------------------------------------------------------
//  img_user_settings_encode
///     Encode user settings data
//
//  Description:
///     Encodes a byte stream with T/L/V's representing the user settings
///     associated with a user configuration.
//
/// \return
///     None
//
void img_user_settings_encode(
    UINT32  image_qual_pref,                                           ///< Input:     User mode to encode
    UINT8   user_cfg_tlv_bstream[MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES], ///< Output:    Encoded T/L/V byte stream
    UINT32 *num_bytes_written                                          ///< Output:    Number of bytes written
    )
{
    sTERA_BSTREAM_CBLK  user_cfg_cb;

    // Initialize our byte stream for encoding
    tera_bstream_init( &user_cfg_cb,
                       &user_cfg_tlv_bstream[0],
                       MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES );

    // Encode frame rate vs quality factor
    tera_tlv_set_tlv32( &user_cfg_cb,
                        MGMT_IMG_APDU_FOURCC_PREFERENCE_FRAME_RATE_VS_QUALITY,
                        image_qual_pref );

    // Any subsequent user settings we wish to add would get added here.

    // Return the number of bytes written to the byte stream
    *num_bytes_written = tera_bstream_get_curpos( &user_cfg_cb );
} // img_user_settings_encode

// --------------------------------------------------------------------------
//  img_user_settings_decode
///     Parse user settings data
//
//  Description:
///     Parses a given set of user configuration settings and returns
///     the settings in the same format as stored internally in the system.
///
///     Note that this is the function that gets bigger
///     as we add more user configuration settings
//
/// \return
///     None
//
void img_user_settings_decode(
    sTERA_PCOIP_HCC_IMG_MODE  *current_mode,                  ///< Input:     Current user config settings
    UINT8                     *user_cfg_tlv_bstream,          ///< Input:     Encoded user settings T/L/V byte stream
    UINT32                     user_cfg_tlv_bstream_length,   ///< Input:     Length of incoming byte stream
    sTERA_PCOIP_HCC_IMG_MODE  *new_mode                       ///< Output:    New user cfg settings returned to the caller
    )
{
    sTERA_BSTREAM_CBLK  user_cfg_cb;
    UINT32              type;
    UINT32              length;
    UINT32              save_tlv_offset;
    UINT32              next_tlv_offset;
    UINT32              net_type;
    char                *type_str;

    // Start off by assuming our new mode is identical to our current mode
    *new_mode = *current_mode;

    // Only read data if we have a non-zero size byte stream
    if ( user_cfg_tlv_bstream_length != 0 )
    {
        // Initialize the byte stream for decoding
        tera_bstream_init( &user_cfg_cb,
                           &user_cfg_tlv_bstream[0],
                           user_cfg_tlv_bstream_length );

        // Process user settings parameters until we reach the end of
        // the byte stream
        while ( tera_bstream_get_num_left( &user_cfg_cb ) != 0 )
        {
            save_tlv_offset = tera_bstream_get_curpos( &user_cfg_cb );

            tera_tlv_get_tl( &user_cfg_cb,
                             &type,
                             &length,
                             &next_tlv_offset );

            // Convert type from host byte order to network byte order for printing
            net_type = mTERA_UINT32_BYTE_SWAP(type);

            type_str = (char *) &net_type;

            mTERA_LOG_DBG("Received user setting %c%c%c%c",
                          type_str[0],
                          type_str[1],
                          type_str[2],
                          type_str[3]);

            if ( type == MGMT_IMG_APDU_FOURCC_PREFERENCE_FRAME_RATE_VS_QUALITY )
            {
                // consistency check: length field
                if ( length != sizeof( UINT32 ) )
                {
                    // Invalid T/L/V given to us, break out of our loop
                    mTERA_LOG_ERR("Invalid image preference length: %d != %d",
                                  length,
                                  (UINT32)sizeof(UINT32) );
                    break;
                }

                // Retrieve frame rate vs quality factor
                tera_tlv_get_v32( &user_cfg_cb,
                                  &new_mode->image_qual_pref );
            }
            else if ( type == MGMT_IMG_APDU_FOURCC_BUILD_TO_LOSSLESS_ENABLE )
            {
                // consistency check: length field
                if ( length != sizeof( UINT32 ) )
                {
                    // Invalid T/L/V given to us, break out of our loop
                    mTERA_LOG_ERR("Invalid build to lossless enable length: %d != %d",
                                  length,
                                  (UINT32)sizeof(UINT32) );
                    break;
                }

                // Retrieve build to lossless enable
                tera_tlv_get_v32( &user_cfg_cb,
                                  &new_mode->build_to_lossless_enable );
            }
            else
            {
                // Unexpected T/L/V type. Make sure we go forward and
                // don't overrun our byte stream.
                if (   ( next_tlv_offset >  save_tlv_offset )
                    && ( next_tlv_offset <= user_cfg_tlv_bstream_length ) )
                {
                    mTERA_LOG_DBG("Unknown user setting %c%c%c%c",
                                  type_str[0],
                                  type_str[1],
                                  type_str[2],
                                  type_str[3] );

                    tera_bstream_set_curpos( &user_cfg_cb,
                                             next_tlv_offset );
                }
                else
                {
                    // Invalid T/L/V given to us, break out of our loop
                    mTERA_LOG_ERR("Invalid user setting %c%c%c%c Length %d",
                                  type_str[0],
                                  type_str[1],
                                  type_str[2],
                                  type_str[3],
                                  length );
                    break;
                }
            }
        }
    }
} // img_user_settings_decode()

// --------------------------------------------------------------------------
//  img_host_user_cfg_apdu_proc
//      Process a user configure APDU
//
//  Description:
//      This function processes a user configure APDU from the firmware. The
//      APDU sent asynchronously by the FW at the start of a session or when
//      the image quality settings are modified by the FW.
//
//      This function stores a copy of the received user config settings in
//      the control block.
//
static void img_host_user_cfg_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_TX_PARAMS *user_cfg_apdu = apdu;
    sTERA_PCOIP_HCC_IMG_MODE                          new_setting;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len <= sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected>=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS));
    }
    else
    {
        img_user_settings_decode(&cblk->user_cfg_setting,
                                 user_cfg_apdu->user_cfg_tlv_bstream,
                                 apdu_len - sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS),
                                 &new_setting);

        mTERA_LOG_DBG("img_qual_pref=%d, b2l=%d",
                      new_setting.image_qual_pref,
                      new_setting.build_to_lossless_enable);

        //
        // Store a copy of the message in the control block. Currently the
        // host SW doesn't do anything else with the message. Having a copy
        // in the control block is useful because the driver initializes
        // a structure containing this data when decoding newly received
        // messages that might not contain all of the parameters in the
        // structure.
        //
        // Also set the flag indicating the host user config APDU was received.
        // The host SW uses this to determine whether certain APDUs can be sent
        // to the FW. FW releases prior to 3.5.0 have a problem that can cause
        // the FW to crash when processing APDU messages above 20 (0x14), which
        // did not exist when the FW was originally built.
        //
        down(&cblk->hmi.data_lock);
        memcpy(&cblk->user_cfg_setting, &new_setting, sizeof(sTERA_PCOIP_HCC_IMG_MODE));
        up(&cblk->hmi.data_lock);
    }
} // img_host_user_cfg_apdu_proc()

// --------------------------------------------------------------------------
//  img_host_user_cfg_req_apdu_proc
//      Process a user configure request APDU
//
//  Description:
//      This function processes a user configure request APDU from the
//      firmware. The APDU is a response to a user config request sent to
//      the firmware.
//
//      This function looks for a pending IMG_USER_CFG_REQ_GET ioctl and
//      copies the user config settings to the output buffer of the ioctl and
//      completes it. If an IOCTL is not pending the response is discarded.
//
static void img_host_user_cfg_req_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_TX_PARAMS *user_cfg_apdu = apdu;
    sTERA_PCOIP_HCC_IMG_MODE                              new_setting;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len <= sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_RX_PARAMS)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected>=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_RX_PARAMS));
    }
    // if the correct IOCTL is not pending
    else if (cblk->pending_ioctl_flag != TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_REQ_GET)
    {
        // Ioctl could have been cancelled
        mTERA_LOG_INFO("no pending HCC_IOCTL_IMG_USER_CFG_REQ_GET");
    }
    else
    {
        img_user_settings_decode(&cblk->user_cfg_setting,
                                 user_cfg_apdu->user_cfg_tlv_bstream,
                                 apdu_len - sizeof(sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS),
                                 &new_setting);

        mTERA_LOG_DBG("img_qual_pref=%d, b2l=%d",
                      new_setting.image_qual_pref,
                      new_setting.build_to_lossless_enable);

        down(&cblk->data_lock);

        // Store the response in the control block
        memcpy(&cblk->ioctl_rsp.data.usr_cfg.user_cfg_settings,
               &new_setting,
               sizeof(sTERA_PCOIP_HCC_IMG_MODE));

        up(&cblk->data_lock);

        // set the flag that will wake up the waiting ioctl thread
        cblk->pending_ioctl_flag = 1;

        // wake up the waiting thread
        wake_up_interruptible(&cblk->pending_ioctl_q);
    }
} // img_host_user_cfg_req_apdu_proc()

// --------------------------------------------------------------------------
//  hcc_img_init
//      Initialize the IMG user
//
//  Description:
//      This function initializes the IMG user. This function is responsible
//      for initializing any and all user-specific objects.
//
INT32 hcc_img_init(
    sHCC_CBLK *cblk    // HCC control block
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // Placeholder for now
    ret = TERA_SUCCESS;

    mTERA_LOG_DBG("end");

    return ret;
} // hcc_img_init()

// --------------------------------------------------------------------------
//  hcc_img_ioctl_proc
//      Process a IMG ioctl request
//
//  Description:
//      This function processes a IMG ioctl request. If the ioctl is not
//      supported, then it is completed with STATUS_NOT_SUPPORTED.
//
INT32 hcc_img_ioctl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32           ret;

    ASSERT(cblk != NULL);

    switch (ioctl)
    {
        case TERA_PCOIP_HCC_IOCTL_IMG_DISP_TOPO_GET:
        {
            ret = img_ioctl_host_topology_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_DISP_TOPO_SEND:
        {
            ret = img_ioctl_host_topology_apdu_send_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_ASYNC_DISP_TOPO_GET:
        {
            ret = img_ioctl_async_host_topology_apdu_get_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_CACHE_DISP_TOPO_GET:
        {
            ret = img_ioctl_cache_topology_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_CACHE_DISP_TOPO_SET:
        {
            ret = img_ioctl_cache_topology_apdu_set_proc(cblk, arg);
            break;
        }
        //
        // TBD - this IOCTL might not be working, it won't be used in the beta
        //       so I decided to postpone looking into the potential problem
        //
        case TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_SET:
        {
            ret = img_ioctl_user_cfg_apdu_set_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_REQ_GET:
        {
            ret = img_ioctl_user_cfg_req_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HCC_IOCTL_IMG_USER_CFG_CACHE_SET:
        {
            ret = img_ioctl_user_cfg_cache_apdu_set_proc(cblk, arg);
            break;
        }
        default:
        {
            mTERA_LOG_DBG("unsupported ioctl=0x%08x", ioctl);
            ret = TERA_ERR_NOT_SUPPORTED;
            break;
        }
    }
    return ret;
} // hcc_img_ioctl_proc()

// --------------------------------------------------------------------------
//  hcc_img_apdu_proc
//      Process an IMG APDU
//
//  Description:
//      This function processes an IMG APDU from the ingress descriptor
//      queue. Only known APDUs are serviced.
//
//      NOTE: The APDU is freed by the calling function after this function
//      returns, so it should NOT be referenced for future use.
//
void hcc_img_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE *apdu_type = apdu;

    ASSERT(cblk != NULL);
    ASSERT(apdu != NULL);

    if (apdu_len < sizeof(eTERA_PCOIP_HCC_APDU_IMG_TYPE))
    {
        mTERA_LOG_ERR("invalid APDU length");
    }
    else if (cblk->initialized)
    {
        // Need to byte swap the APDU type value
        *apdu_type = mTERA_UINT32_BYTE_SWAP(*apdu_type);

        switch (*apdu_type)
        {
            case TERA_PCOIP_HCC_APDU_IMG_TYPE_DISP_TOPO:
            {
                mTERA_LOG_DBG("HCC_APDU_IMG_TYPE_DISP_TOPO len=%d", apdu_len);

                img_host_topology_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_GET:
            {
                mTERA_LOG_DBG("HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_GET len=%d", apdu_len);

                img_cache_topology_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG:
            {
                mTERA_LOG_DBG("HCC_APDU_IMG_TYPE_HOST_USER_CFG len=%d", apdu_len);

                img_host_user_cfg_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_REQ:
            {
                mTERA_LOG_DBG("HCC_APDU_IMG_TYPE_HOST_USER_CFG_REQ len=%d", apdu_len);

                img_host_user_cfg_req_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_CACHE_SET:
            {
                mTERA_LOG_ERR("rcvd unexpected HCC_APDU_IMG_TYPE_HOST_USER_CFG_CACHE_SET len=%d",
                              apdu_len);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unsupported IMG APDU type=0x%x len=%d",
                              *apdu_type, apdu_len);
                break;
            }
        }
    }
} // hcc_img_apdu_proc()

// --------------------------------------------------------------------------
//  hcc_img_scp_chan_open_proc
//      Process a SCP channel open event
//
//  Description:
//      This function processes a SCP channel open event. This function sends
//      out a request to get the available IMG commands from the firmware.
//
void hcc_img_scp_chan_open_proc(
    sHCC_CBLK *cblk
    )
{
    ASSERT(cblk != NULL);

    mTERA_LOG_DBG("start");

    mTERA_LOG_DBG("end");
} // hcc_img_scp_chan_open_proc()

