////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Host Communications Channel Module Internal
///
///     This file defines the internal HDC module function prototypes and
///     definitions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_pcoip_hdc_apdu.h"
#include "tera_mem.h"
#include "tera_gfun.h"
#include "tera_msg.h"
#include "tera_pkt.h"
#include "tera_scp.h"

// ---- Constants and Types -------------------------------------------------

typedef enum
{
    HCC_DEVIF_STATE_INIT    = 0,     ///< Initial state - not yet created
    HCC_DEVIF_STATE_OPENED  = 1,     ///< Opened state - ready for ioctls
    HCC_DEVIF_STATE_CLOSING = 2,
    HCC_DEVIF_STATE_CLOSED  = 3,
    HCC_DEVIF_STATE_ERROR   = 4,     ///< Error state
} eHCC_DEVIF_STATE;

// This constant determines how many of the last client cursor reports are
// saved by the driver. The driver searches through these reports when it
// checks if the host cursor was moved by something other than the client.
// LAN testing showed this constant needs to be around 4. Doubling the value
// to be safe.
#define MAX_CLIENT_CUR_POINTS       (8)

// --------------------------------------------------------------------------
//  sHDC_KMP_CBLK
///     KMP user sub-control block
//
//  Description:
///     This structure defines the KMP user sub-control block. This control
///     block contains variables specific to the KMP user and should only be
///     accessed by the KMP user functions.
///
typedef struct sHDC_KMP_CBLK_STRUCT
{
    struct semaphore                    data_lock;          ///< Local data semaphore
//    ULONGLONG                           cur_systime_last_evt; ///< last cursor event system time
    UINT32                              cur_tstamp_last_evt;  ///< last cursor event timestamp
    UINT32                              kbd_last_key_press;   ///< Last local keyboard key press
    UINT32                              kbd_last_key_break_pend; ///< Flag indicating the last HID key
                                                                 /// press break needs to be injected
                                                                 /// (this is done when the client
                                                                 /// handles repeat processing)
    UINT32                              last_consumer_key;    ///< value of the last consumer key reported,
                                                              ///  this field is cleared when the key is released
    UINT32                              kbd_consumer_key_break_pend; ///< Flag indicating the last consumer key
                                                                     /// press break needs to be injected
                                                                     /// (this is done when the client
                                                                     /// handles repeat processing)
    UINT32                              last_gen_desk_key;    ///< value of the last generic desktop key reported,
                                                              ///  this field is cleared when the key is released
    UINT32                              kbd_gen_desk_key_break_pend; ///< Flag indicating the last generic key
                                                                     /// press break needs to be injected
                                                                     /// (this is done when the client
                                                                     /// handles repeat processing)
    UINT16                                     lang;          ///< Language of the keyboard
    UINT8                                      modifier;      ///< modifier state (valid for non-USB)
    sTERA_MSG_Q                                cur_evt_msg_q; ///< Local cursor event message queue
    sTERA_MSG_Q                                kbd_evt_msg_q; ///< Local keyboard event message queue
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT        kbd_repeat;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK          kbd_lock;      ///< Keyboard lock states
    sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT         feat_stat;     ///< Copy last feat status msg
    sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE disp_tbl;      ///< Copy of active display topology
    sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL        driver_cfg;    ///< Driver config settings

    // state variables
    UINT32    lkbd_last_key_id;         ///< Last local keyboard key id
    UINT32    lkbd_last_key_page;       ///< Last local keyboard key page
    UINT32    lkbd_last_key_state;      ///< Last local keyboard key state
    UINT32    lcur_last_btn_state;      ///< Last local cursor button state
    UINT32    lcur_last_cursor_disp;    ///< Display the last local cursor was on
    UINT32    idx_last_client_pt;       ///< Index of the last stored client cursor point
    UINT16    lcur_last_abs_x[MAX_CLIENT_CUR_POINTS]; ///< Last absolute X positions
    UINT16    lcur_last_abs_y[MAX_CLIENT_CUR_POINTS]; ///< Last absolute Y positions
    unsigned long max_age_cur_rep[MAX_CLIENT_CUR_POINTS]; ///< max age to consider cursor report valid

    // OS defined structs
    struct input_dev *lkbd;    ///< Keyboard input device
    struct input_dev *lcur_abs;///< Absolute cursor (mouse) input device
    struct input_dev *lcur_rel;///< Relative cursor (mouse) input device
} sHDC_KMP_CBLK;

// --------------------------------------------------------------------------
//  sHDC_CBLK
///     HDC module control block
//
//  Description:
///     This structure defines the Host Communications Channel (HDC) module
///     control block. The control block is per device instance, and not per
///     driver instance. The control block is attached to the context area of
///     each device instance.
///
typedef struct sHDC_CBLK_STRUCT
{
    struct pci_dev         *dev;                ///< Device object
    INT                     initialized;        ///< Initialized flag
    INT                     started;            ///< Started flag
    TERA_SCP_HANDLE         scp_handle;         ///< SCP handle

    wait_queue_head_t       pending_ioctl_q;    ///< Pending ioctl queue
    struct semaphore        pending_ioctl_lock; ///< Pending ioctl semaphore
    INT32                   pending_ioctl_flag; ///< flag used to signal the FW req completed
    sTERA_HDC_MSG_IOCTL_RSP ioctl_rsp;          ///< copy of data from FW that belongs in the ioctl response
    struct semaphore        data_lock;          ///< Local data semaphore
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct delayed_work     work;               ///< dpc_entry work queue task
#else
    struct work_struct      work;               ///< dpc_entry work queue task
#endif

    sTERA_PKT_Q             egress_pkt_q;       ///< Egress packet queue
    sHDC_KMP_CBLK           kmp;                ///< KMP user control block
} sHDC_CBLK;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

//
// KMP user function prototypes
//

extern INT32 hdc_kmp_init(
    sHDC_CBLK *cblk
    );

extern void hdc_kmp_cleanup(
    sHDC_CBLK *cblk
    );

extern INT32 hdc_kmp_start(
    sHDC_CBLK *cblk
    );

extern void hdc_kmp_stop(
    sHDC_CBLK *cblk
    );

extern void hdc_kmp_scp_chan_open_proc(
    sHDC_CBLK *cblk
    );

extern INT32 hdc_kmp_ioctl_proc(
    sHDC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

extern void hdc_kmp_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

extern INT hdc_kmp_thread_proc(
    sHDC_CBLK *cblk
    );

