////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     HDC Module KMP User Implementation
///
///     This file implements the HDC module KMP user. All KMP user related
///     packets/messages and ioctls are dealt with in this file.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"

#include "tera_pcoip_hdc_ioctl.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_hdc.h"
#include "tera_hdc_internal.h"

#include <linux/input.h>

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// driver local cursor report processing algorith type table
#ifdef DBG
static const char *lcur_evt_alg_str_tbl[TERA_PCOIP_HDC_LCUR_MAX_ALG] = {
    "Absolute",
    "Relative"
};
#endif

// This constant determines the maximum age a cursor report is considered
// valid. The application can send a host cursor report to the driver and the
// driver must determine if the cursor was moved by the client or some other
// entity. The driver maintains an array of the last X cursor reports and each
// entry in the array includes a timestamp that is equal to the time the
// report was received plus this constant. This is considered the maximum age
// the cursor report is considered to be valid.
//
// The code must look at the timestamp to deal with the situation where a host
// mouse moves the cursor back along the same path the client mouse moved the
// cursor. For example, if the client mouse moved the cursor from left to right
// and later the host mouse moved the cursor from right to left at the same
// speed the driver would find old client reports that are the same as the
// new host cursor reports and not move the cursor. This problem is avoided
// by aging the reports.
#define  MAX_AGE_CURSOR_REPORT  msecs_to_jiffies(400)

// The following variable is used to notify the keyboard driver that the state
// of the keyboard LEDs has changed.
static  unsigned long kbd_event_arrived = 0;

//
// The first 232 (0xE8) entries of this array maps USB keyboard (page 7) ID
// values received from the PCoIP local keyboard to keycode values processed by
// the event handler.
//
// This table is also used during initialization. The driver registers with the
// event handler notifying the event handler what keys it is capable of
// generating. If the driver is modified to add support for another key the key's
// keycode must be added to this table and the function
// kmp_thread_update_keyboard_apdu_proc() must also be modified to support the
// new key.
//
// To verify new keys are working use the Linux showkey function. This function
// runs from an xterm by the root. It displays the key's keycode. First
// run a test with the keyboard connected through a hub or disable the
// local cursor feature. During this step make note of the keycode generated
// by each key press. Next, enable the local cursor feature or connect the
// keyboard directly to the client. Repeat the test using showkey to verify
// each keypress works.
//
#define NUM_ENTRIES_KEYCODE_TABLE           256
static const unsigned char kbd_keycode[NUM_ENTRIES_KEYCODE_TABLE] = {
    0,     0,     0,     0,                                      // HID page 7, ID 0x0
    KEY_A, KEY_B, KEY_C, KEY_D,                                  // HID page 7, ID 0x4
    KEY_E, KEY_F, KEY_G, KEY_H,                                  // HID page 7, ID 0x8
    KEY_I, KEY_J, KEY_K, KEY_L,                                  // HID page 7, ID 0xC
    KEY_M, KEY_N, KEY_O, KEY_P,                                  // HID page 7, ID 0x10
    KEY_Q, KEY_R, KEY_S, KEY_T,                                  // HID page 7, ID 0x14
    KEY_U, KEY_V, KEY_W, KEY_X,                                  // HID page 7, ID 0x18
    KEY_Y, KEY_Z, KEY_1, KEY_2,                                  // HID page 7, ID 0x1C
    KEY_3, KEY_4, KEY_5, KEY_6,                                  // HID page 7, ID 0x20
    KEY_7, KEY_8, KEY_9, KEY_0,                                  // HID page 7, ID 0x24
    KEY_ENTER, KEY_ESC, KEY_BACKSPACE, KEY_TAB,                  // HID page 7, ID 0x28
    KEY_SPACE, KEY_MINUS, KEY_EQUAL, KEY_LEFTBRACE,              // HID page 7, ID 0x2C
    KEY_RIGHTBRACE, KEY_BACKSLASH, KEY_BACKSLASH, KEY_SEMICOLON, // HID page 7, ID 0x30
    KEY_APOSTROPHE, KEY_GRAVE, KEY_COMMA, KEY_DOT,               // HID page 7, ID 0x34
    KEY_SLASH, KEY_CAPSLOCK, KEY_F1, KEY_F2,                     // HID page 7, ID 0x38
    KEY_F3, KEY_F4, KEY_F5, KEY_F6,                              // HID page 7, ID 0x3C
    KEY_F7, KEY_F8, KEY_F9, KEY_F10,                             // HID page 7, ID 0x40
    KEY_F11, KEY_F12, KEY_SYSRQ, KEY_SCROLLLOCK,                 // HID page 7, ID 0x44
    KEY_PAUSE, KEY_INSERT, KEY_HOME, KEY_PAGEUP,                 // HID page 7, ID 0x48
    KEY_DELETE, KEY_END, KEY_PAGEDOWN, KEY_RIGHT,                // HID page 7, ID 0x4C
    KEY_LEFT, KEY_DOWN, KEY_UP, KEY_NUMLOCK,                     // HID page 7, ID 0x50
    KEY_KPSLASH, KEY_KPASTERISK, KEY_KPMINUS, KEY_KPPLUS,        // HID page 7, ID 0x54
    KEY_KPENTER, KEY_KP1, KEY_KP2, KEY_KP3,                      // HID page 7, ID 0x58
    KEY_KP4, KEY_KP5, KEY_KP6, KEY_KP7,                          // HID page 7, ID 0x5C
    KEY_KP8, KEY_KP9, KEY_KP0, KEY_KPDOT,                        // HID page 7, ID 0x60
    KEY_102ND, KEY_COMPOSE, KEY_POWER, KEY_KPEQUAL,              // HID page 7, ID 0x64
    KEY_F13, KEY_F14, KEY_F15, KEY_F16,                          // HID page 7, ID 0x68
    KEY_F17, KEY_F18, KEY_F19, KEY_F20,                          // HID page 7, ID 0x6C
    KEY_F21, KEY_F22, KEY_F23, KEY_F24,                          // HID page 7, ID 0x70
    KEY_OPEN, KEY_HELP, KEY_PROPS, KEY_FRONT,                    // HID page 7, ID 0x74
    KEY_STOP, KEY_AGAIN, KEY_UNDO, KEY_CUT,                      // HID page 7, ID 0x78
    KEY_COPY, KEY_PASTE, KEY_FIND, KEY_MUTE,                     // HID page 7, ID 0x7C
    KEY_VOLUMEUP, KEY_VOLUMEDOWN,  0,  0,                        // HID page 7, ID 0x80
    0, KEY_KPCOMMA,  0, KEY_RO,                                  // HID page 7, ID 0x84
    KEY_KATAKANAHIRAGANA, KEY_YEN, KEY_HENKAN, KEY_MUHENKAN,     // HID page 7, ID 0x88
    KEY_KPJPCOMMA,  0,  0,  0,                                   // HID page 7, ID 0x8C
    KEY_HANGEUL, KEY_HANJA, KEY_KATAKANA, KEY_HIRAGANA,          // HID page 7, ID 0x90
    KEY_ZENKAKUHANKAKU,  0,  0,  0,                              // HID page 7, ID 0x94
    0,  0,  0,  0,                                               // HID page 7, ID 0x98
    0,  0,  0,  0,                                               // HID page 7, ID 0x9C
    0,  0,  0,  0,                                               // HID page 7, ID 0xA0
    0,  0,  0,  0,                                               // HID page 7, ID 0xA4
    0,  0,  0,  0,                                               // HID page 7, ID 0xA8
    0,  0,  0,  0,                                               // HID page 7, ID 0xAC
    0,  0,  0,  0,                                               // HID page 7, ID 0xB0
    0,  0,  0,  0,                                               // HID page 7, ID 0xB4
    0,  0,  0,  0,                                               // HID page 7, ID 0xB8
    0,  0,  0,  0,                                               // HID page 7, ID 0xBC
    0,  0,  0,  0,                                               // HID page 7, ID 0xC0
    0,  0,  0,  0,                                               // HID page 7, ID 0xC4
    0,  0,  0,  0,                                               // HID page 7, ID 0xC8
    0,  0,  0,  0,                                               // HID page 7, ID 0xCC
    0,  0,  0,  0,                                               // HID page 7, ID 0xD0
    0,  0,  0,  0,                                               // HID page 7, ID 0xD4
    0,  0,  0,  0,                                               // HID page 7, ID 0xD8
    0,  0,  0,  0,                                               // HID page 7, ID 0xDC
    KEY_LEFTCTRL, KEY_LEFTSHIFT, KEY_LEFTALT, KEY_LEFTMETA,      // HID page 7, ID 0xE0
    KEY_RIGHTCTRL, KEY_RIGHTSHIFT, KEY_RIGHTALT, KEY_RIGHTMETA,  // HID page 7, ID 0xE4
    KEY_NEXTSONG, KEY_PREVIOUSSONG, KEY_STOPCD, KEY_PLAYPAUSE,   // Consumer page C, ID 0xE8
    KEY_MUTE, KEY_BASSBOOST, KEY_VOLUMEUP, KEY_VOLUMEDOWN,       // Consumer page C, ID 0xEC
    KEY_MEDIA, KEY_MAIL, KEY_CALC, KEY_COMPUTER,                 // Consumer page C, ID 0xF0
    KEY_SEARCH, KEY_HOMEPAGE, KEY_BACK, KEY_FORWARD,             // Consumer page C, ID 0xF4
    KEY_STOP, KEY_REFRESH, KEY_BOOKMARKS, KEY_CALC,              // Consumer page C, ID 0xF8
    KEY_POWER, KEY_SLEEP, KEY_WAKEUP, 0                          // Generic desktop page 1, ID 0xFC
};

//
// The following constant determines the index of the last USB scancode from
// the HID page (7) that the driver supports mapping. The client should not
// send keycodes from page 7 that exceed this value.
//
#define MAX_IDX_USB_KEYBOARD_KEYCODE    0xE7

// The following constants determine the range of the X & Y coordinates
// reported to the event manager.
#define TERA_LCUR_MAX_WIDTH    0xFFFF
#define TERA_LCUR_MAX_HEIGHT   0xFFFF

// The following constants are used to determine when to inject keyboard and
// cursor data to the upper level drivers
#define TICKS_PER_MS_INTERRUPT_TIME     10000
#define TICKS_PER_SEC_INTERRUPT_TIME    (1000 * TICKS_PER_MS_INTERRUPT_TIME)

// The following constants are the USB scancodes for the keys that affect the
// value of the modifier field.
#define LEFT_CTRL_KEY_USB_CODE      0x000700E0
#define LEFT_CTRL_KEY_USB_CODE      0x000700E0
#define LEFT_SHFT_KEY_USB_CODE      0x000700E1
#define LEFT_ALT_KEY_USB_CODE       0x000700E2
#define LEFT_WIN_KEY_USB_CODE       0x000700E3
#define RIGHT_CTRL_KEY_USB_CODE     0x000700E4
#define RIGHT_SHFT_KEY_USB_CODE     0x000700E5
#define RIGHT_ALT_KEY_USB_CODE      0x000700E6
#define RIGHT_WIN_KEY_USB_CODE      0x000700E7
#define FIRST_MODIFIER_USB_CODE     (LEFT_CTRL_KEY_USB_CODE & 0x0000FFFF)

// The following constants are the bit masks for the each of the keys that
// affect the value of the modifier field.
#define LEFT_CTRL_KEY_MOD_BIT_MSK   0x01
#define LEFT_SHFT_KEY_MOD_BIT_MSK   0x02
#define LEFT_ALT_KEY_MOD_BIT_MSK    0x04
#define LEFT_WIN_KEY_MOD_BIT_MSK    0x08
#define RIGHT_CTRL_KEY_MOD_BIT_MSK  0x10
#define RIGHT_SHFT_KEY_MOD_BIT_MSK  0x20
#define RIGHT_ALT_KEY_MOD_BIT_MSK   0x40
#define RIGHT_WIN_KEY_MOD_BIT_MSK   0x80

// The following constants were taken from the FW file tera_mgmt_rdp_fsm.c.
// The constants correspond to different values found in the Language
// field of the Set Keyboard Locale APDU sent from the Client to the Host at
// the start of a session. One of the PS/2 keys is dependent on the type of
// keyboard connected to the client. The function that maps PS/2 keys to USB
// keys uses the language field to determine how to map the 002B/00AB PS/2 key.
// In the current implementation, the following languages map this key to the
// USB 07/31 key and all other language values map the key to the 07/32 key.
#define KBD_LANG_USA        0x0409  // USA ISO-8859-1              - en-us: English (United States)
#define KBD_LANG_JAPANESE   0x0411  // Japanese 106                - ja: Japanese
#define KBD_LANG_KOREAN     0x0412  // Korean Dubeolsik ISO-8859-1 - ko-kr: Korean/Hangul (South Korea)

//
// Define the following constant when you want to view additional log messages
// that show when cursor data reports are received from the client and
// processed by the driver. Multiple log messages are generated for each event,
// which can be somewhat large.
//
#undef CUR_REPORT_DBG
//#define CUR_REPORT_DBG  1

//
// Define the following constant when you want to view additional log messages
// that show when keyboard data reports are received from the client and
// processed by the driver. Multiple log messages are generated for each event,
// which can be somewhat large.
//
// NOTE: this should only be defined for internal builds because it displays
// actual key press data.
//
//#undef KBD_REPORT_DBG
#define KBD_REPORT_DBG  1

//
// The following tables are used to convert PS/2 keyboard scancodes into
// USB HID scancodes. The tables return the value 0x00 for unassigned
// and DO NOT USE scancodes.
//
// The scancode values were obtained from the "USB HID to PS/2 Scan Code
// Translation Table" document from Microsoft at the following web site:
// http://download.microsoft.com/download/1/6/1/161ba512-40e2-4cc9-843a-923143f3456c/translate.pdf
//
#define PS2_LKUP_UNUSED         0x00
#define UPPER_24_MASK           0xFFFFFF00
#define LOWER_8_MASK            0x000000FF

//
// PS/2 to USB scancode lookup table #1. This table handles the PS/2
// scancodes with the upper 24 bits equal to 0. It returns
// only the code, the caller must append the page, which equals 7 for
// all of the PS/2 scancodes handled by this table.
//
// Note: This table maps the 002B/00AB key to the value 07/32. This may need to
//       be changed based on the language of the keyboard. The code that uses
//       this table must check to see if the mapped key should be overridden.
//
static const UINT8 ps2_lkup_table1[256] = {
//00-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x00, 0x29, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23,
    0x24, 0x25, 0x26, 0x27, 0x2D, 0x2E, 0x2A, 0x2B,
//10-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x14, 0x1A, 0x08, 0x15, 0x17, 0x1C, 0x18, 0x0C,
    0x12, 0x13, 0x2F, 0x30, 0x28, 0xE0, 0x04, 0x16,
//20-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x07, 0x09, 0x0A, 0x0B, 0x0D, 0x0E, 0x0F, 0x33,
    0x34, 0x35, 0xE1, 0x32, 0x1D, 0x1B, 0x06, 0x19,
//30-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x05, 0x11, 0x10, 0x36, 0x37, 0x38, 0xE5, 0x55,
    0xE2, 0x2C, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E,
//40-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x3F, 0x40, 0x41, 0x42, 0x43, 0x53, 0x47, 0x5F,
    0x60, 0x61, 0x56, 0x5C, 0x5D, 0x5E, 0x57, 0x59,
//50-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x5A, 0x5B, 0x62, 0x63, 0x00, 0x00, 0x64, 0x44,
    0x45, 0x67, 0x00, 0x00, 0x8C, 0x00, 0x00, 0x00,
//60-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x00, 0x00, 0x00, 0x00, 0x68, 0x69, 0x6A, 0x6B,
    0x6C, 0x6D, 0x6E, 0x6F, 0x70, 0x71, 0x72, 0x00,
//70-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x88, 0x91, 0x90, 0x87, 0x00, 0x00, 0x73, 0x93,
    0x92, 0x8A, 0x00, 0x8B, 0x00, 0x89, 0x85, 0x00,
//80-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x00, 0x29, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23,
    0x24, 0x25, 0x26, 0x27, 0x2D, 0x2E, 0x2A, 0x2B,
//90-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x14, 0x1A, 0x08, 0x15, 0x17, 0x1C, 0x18, 0x0C,
    0x12, 0x13, 0x2F, 0x30, 0x28, 0xE0, 0x04, 0x16,
//A0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x07, 0x09, 0x0A, 0x0B, 0x0D, 0x0E, 0x0F, 0x33,
    0x34, 0x35, 0xE1, 0x32, 0x1D, 0x1B, 0x06, 0x19,
//B0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x05, 0x11, 0x10, 0x36, 0x37, 0x38, 0xE5, 0x55,
    0xE2, 0x2C, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E,
//C0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x3F, 0x40, 0x41, 0x42, 0x43, 0x53, 0x47, 0x5F,
    0x60, 0x61, 0x56, 0x5C, 0x5D, 0x5E, 0x57, 0x59,
//D0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x5A, 0x5B, 0x62, 0x63, 0x00, 0x00, 0x64, 0x44,
    0x45, 0x67, 0x00, 0x00, 0x8C, 0x00, 0x00, 0x00,
//E0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x00, 0x00, 0x00, 0x00, 0x68, 0x69, 0x6A, 0x6B,
    0x6C, 0x6D, 0x6E, 0x6F, 0x70, 0x71, 0x72, 0x00,
//F0-0,8   1,9   2,A   3,B   4,C   5,D   6,E   7,F
    0x88, 0x91, 0x90, 0x87, 0x00, 0x00, 0x94, 0x93,
    0x92, 0x8A, 0x00, 0x8B, 0x02, 0x89, 0x85, 0x01,
};

//
// PS/2 to USB scancode lookup table #2. This table handles the PS/2
// scancodes with the upper 24 bits equal to 0x0000E0 or 0x00E0F0. This
// lookup table returns both the page and code combined.
//
// Note: according to the tables in the Microsoft document the entries in this
//       table should be used when the upper 24-bit of the PS/2 scancode equals
//       0x0000E0 0r 00E0F), but testing with the soft-client shows the upper
//       24 bits of the scancode always equals 0x000001 for the keys in this
//       table.
//
static const UINT32 ps2_lkup_table2[256] = {
//00-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x070048, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
//10-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00B6, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x0C00B5, 0x000000, 0x000000, 0x070058, 0x0700E4, 0x000000, 0x000000,
//20-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00E2, 0x0C0192, 0x0C00CD, 0x000000, 0x0C00B7, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x0C00EA, 0x000000,
//30-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00E9, 0x000000, 0x0C0223, 0x000000, 0x000000, 0x070054, 0x000000, 0x070046,
    0x0700E6, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
//40-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x070048, 0x07004A,
    0x070052, 0x07004B, 0x000000, 0x070050, 0x000000, 0x07004F, 0x000000, 0x07004D,
//50-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x070051, 0x07004E, 0x070049, 0x07004C, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x0700E3, 0x0700E7, 0x070065, 0x070066, 0x010082,
//60-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x010083, 0x000000, 0x0C0221, 0x0C022A, 0x0C0227,
    0x0C0226, 0x0C0225, 0x0C0224, 0x0C0194, 0x0C018A, 0x0C0183, 0x000000, 0x000000,
//70-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x070091, 0x070090, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
//80-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
//90-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00B6, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x0C00B5, 0x000000, 0x000000, 0x070058, 0x0700E4, 0x000000, 0x000000,
//A0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00E2, 0x0C0192, 0x0C00CD, 0x000000, 0x0C00B7, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x0C00EA, 0x000000,
//B0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x0C00E9, 0x000000, 0x0C0223, 0x000000, 0x000000, 0x070054, 0x000000, 0x070046,
    0x0700E6, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
//C0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x070048, 0x07004A,
    0x070052, 0x07004B, 0x000000, 0x070050, 0x000000, 0x07004F, 0x000000, 0x07004D,
//D0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x070051, 0x07004E, 0x070049, 0x07004C, 0x000000, 0x000000, 0x010081, 0x010082,
    0x000000, 0x000000, 0x000000, 0x0700E3, 0x0700E7, 0x070065, 0x070066, 0x000000,
//E0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x010083, 0x000000, 0x0C0221, 0x0C022A, 0x0C0227,
    0x0C0226, 0x0C0225, 0x0C0224, 0x0C0194, 0x0C018A, 0x0C0183, 0x000000, 0x000000,
//F0-0,8      1,9       2,A       3,B       4,C       5,D       6,E       7,F
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
    0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000, 0x000000,
};

//
// Define the following constant when you want to view additional log messages
// that show host cursor position reports compared to client cursor data.
//
//#undef HOST_CUR_DBG
#define HOST_CUR_DBG  1

#define TERA_USB_PAGE_GEN_DESK      0x0001
#define TERA_USB_PAGE_KEYBOARD      0x0007
#define TERA_USB_PAGE_CONSUMER      0x000C

// Macros return non-zero value if the key scancode is from the Consumer Page
// (0x0C), Keyboard Page (0x07) or Generic Desktop (0x01)
#define mTERA_CONSUMER_PAGE_KEYCODE(key) \
    (((key) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_PAGE) == (TERA_USB_PAGE_CONSUMER << 16))
#define mTERA_KEYBOARD_PAGE_KEYCODE(key) \
    (((key) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_PAGE) == (TERA_USB_PAGE_KEYBOARD << 16))
#define mTERA_GEN_DESK_PAGE_KEYCODE(key) \
    (((key) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_PAGE) == (TERA_USB_PAGE_GEN_DESK << 16))

// Macro returns non-zero value if the key state indicates the key is unicode
#define mTERA_UNICODE_KEYCODE(state) \
    ((state) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_UNICODE)

// Macro returns non-zero value if the key state indicates the key is a key
// press event
#define mTERA_KEY_PRESS_EVENT(state) \
    ((state) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_MAKE)

// These macros define the generic desktop keys the host SW driver supports.
// These values are sent from the client to the driver. The driver maps these
// keys to specific bits in the HID report.
#define GEN_DESK_SYS_POWERDOWN_KEYCODE          0x0081
#define GEN_DESK_SYS_SLEEP_KEYCODE              0x0082
#define GEN_DESK_SYS_WAKEUP_KEYCODE             0x0083

// These macros define the bit-fields of the generic desktop keys the host SW
// driver supports. These values are included in the HID report sent from the
// driver to the OS.
#define GEN_DESK_SYS_POWERDOWN_HID_REPORT_BIT   0x01
#define GEN_DESK_SYS_SLEEP_HID_REPORT_BIT       0x02
#define GEN_DESK_SYS_WAKEUP_HID_REPORT_BIT      0x04

// Macro returns non-zero value if the key scancode is a generic desktop key
// the driver supports. Currently the driver only supports the following 3 keys
// - System Power Down
// - System Sleep
// - System Wake Up
// The driver only supports these keys because we've only found keyboards that
// support these keys. If we need to support additional keys the host SW will
// need to be updated.
#define mTERA_GEN_DESK_PAGE_VALID_KEYCODE(key) \
    ((((key) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_CODE) >= GEN_DESK_SYS_POWERDOWN_KEYCODE) && \
     (((key) & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_CODE) <= GEN_DESK_SYS_WAKEUP_KEYCODE))

// The following constants are stored in the USB reports sent to the upper
// layer driver. These values must match the Report ID fields in the
// hid_report_descr structure in the file tera_hid.c. Ideally they should be
// put in a common include file and referenced by both this file and the file
// tera_hid.c. Right now neither of these files include any common header files
// other than the files in the common/inc directory, so I decided not to
// introduce a new dependency. If the value of these constants change the
// corresponding values in the hid_report_descr structure must also be updated.
#define USB_MOUSE_REPORT_ID         1
#define USB_KEYBOARD_REPORT_ID      2
#define USB_CONSUMER_REPORT_ID      3
#define USB_GEN_DESK_REPORT_ID      4
#define TERA_CONSUMER_REPORT_LEN    3
#define TERA_GEN_DESK_REPORT_LEN    2

static INT is_client_kbd_repeat_en(
    sHDC_CBLK *cblk
    );

static void kmp_key_lcur_press_reset(
    sHDC_CBLK *cblk
    );

static INT32 kmp_apdu_send(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static int kmp_send_kbd_output_data(
    sHDC_CBLK *cblk
    );

static INT32 kmp_ioctl_hdc_open_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    );

static INT32 kmp_ioctl_hdc_close_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_feature_activation_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    );

static INT32 kmp_ioctl_last_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_keyboard_repeat_rate_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_update_pointer_shape_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_change_pointer_shape_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_set_pointer_location_if_new_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_set_pointer_location_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_set_pointer_visibility_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_fade_out_control_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_mouse_speed_query_apdu_get_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    );

static INT32 kmp_ioctl_mouse_speed_update_apdu_set_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_keyboard_repeat_rate_query_apdu_get_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    );

static INT32 kmp_ioctl_disp_topo_set(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_ioctl_driver_cfg_set(
    sHDC_CBLK *cblk,
    unsigned long arg
    );

static INT32 kmp_apdu_kbd_conv_ps2_to_usb(
    UINT32     code,
    UINT16     lang,
    UINT32    *conv_code
    );

static void kmp_apdu_kbd_update_ps2_mod_state(
    UINT32     usb_code,
    UINT8      state,
    UINT8     *mod
    );

static void kmp_update_keyboard_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void kmp_update_mouse_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void kmp_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void kmp_set_keyboard_locale_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void kmp_set_keyboard_repeat_and_lock_apdu_send(
    sHDC_CBLK *cblk
    );

static void kmp_mouse_speed_update_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static void kmp_set_keyboard_repeat_rate_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

static INT kmp_thread_update_keyboard_apdu_proc(
    sHDC_CBLK *cblk
    );

static INT kmp_thread_update_mouse_apdu_proc(
    sHDC_CBLK *cblk
    );

static int kmp_output_kbd_event_proc(
    struct input_dev *dev,
    unsigned int type,
    unsigned int code,
    int value
    );

static INT32 kmp_lkbd_dev_init(
    sHDC_KMP_CBLK *cblk
    );

static INT32 kmp_lcur_dev_init(
    sHDC_KMP_CBLK *cblk
    );

static UINT32 kmp_get_disp_cursor_is_on(
    sHDC_KMP_CBLK *cblk_kmp,
    UINT16         abs_x,
    UINT16         abs_y
    );

// --------------------------------------------------------------------------
//  is_client_kbd_repeat_en
//      Indicates whether client keyboard repeat processing is enabled
//
//  Description:
//      This function returns 0 if client keyboard repeat processing is
//      disabled and 1 when client keyboard repeat processing is enabled.
//
static INT is_client_kbd_repeat_en(
    sHDC_CBLK *cblk
    )
{
    if ((cblk->kmp.kbd_repeat.apdu_type == (eTERA_PCOIP_HDC_APDU_KMP_TYPE)(mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REPEAT))) &&
        (cblk->kmp.kbd_repeat.delay != mTERA_UINT16_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY)) &&
        (cblk->kmp.kbd_repeat.interval != mTERA_UINT16_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE)))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // is_client_kbd_repeat_en()

// --------------------------------------------------------------------------
//  kmp_key_lcur_press_reset
//      Reset any key presses
//
//  Description:
//      This function resets any key presses that are pressed. To do this the
//      function sends an IOCTL_HID_READ_REPORT with no key presses, and no
//      modifier keys pressed.
//
static void kmp_key_lcur_press_reset(
    sHDC_CBLK *cblk
    )
{
    //
    // Report a key up event if the last key press indication was a key down
    // event
    //
    if(cblk->kmp.lkbd_last_key_state)
    {
        cblk->kmp.lkbd_last_key_state = 0;
        input_event(cblk->kmp.lkbd, EV_MSC, MSC_SCAN,
                    (cblk->kmp.lkbd_last_key_page << 16) + cblk->kmp.lkbd_last_key_id);
        input_report_key(cblk->kmp.lkbd,
                         kbd_keycode[cblk->kmp.lkbd_last_key_id],
                         0);
        input_sync(cblk->kmp.lkbd);
    }

    if (cblk->kmp.lcur_last_btn_state)
    {
        mTERA_LOG_DBG("last button state = 0x%02X, clearing",
                      cblk->kmp.lcur_last_btn_state);
        cblk->kmp.lcur_last_btn_state = 0;

        // report coordinates via the event interface
        input_report_key(cblk->kmp.lcur_abs, BTN_LEFT,   0);
        input_report_key(cblk->kmp.lcur_abs, BTN_MIDDLE, 0);
        input_report_key(cblk->kmp.lcur_abs, BTN_RIGHT,  0);
        input_report_key(cblk->kmp.lcur_abs, BTN_SIDE,   0);
        input_report_key(cblk->kmp.lcur_abs, BTN_EXTRA,  0);
        input_sync(cblk->kmp.lcur_abs);
    }
} // kmp_key_lcur_press_reset()

// --------------------------------------------------------------------------
//  kmp_apdu_send
//      Send a KMP APDU
//
//  Description:
//      This function sends a KMP APDU to the blade. The contents of the APDU
//      are copied into an allocated packet and enqueued onto the egress
//      packet queue. The calling function can safely discard the APDU after
//      this function call.
//
static INT32 kmp_apdu_send(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    // Send the APDU
    ret = tera_scp_data_pkt_send(cblk->dev, cblk->scp_handle,
                                 TERA_PCOIP_HDC_APDU_USER_KMP,
                                 &cblk->egress_pkt_q, apdu, apdu_len);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
        ret = -ERESTARTSYS;
    }

    return ret;
} // kmp_apdu_send()

// --------------------------------------------------------------------------
//  kmp_send_kbd_output_data
///     Send a keyboard output data message to the firmware
//
//  Description:
///     The keyboard output data message contains fields that control the state
///     of the keyboard LEDs. When the state of any of the keyboard LEDs change
///     this function is invoked. It creates a message and sends it to the
///     egress queue for transmission.
//
static int kmp_send_kbd_output_data(
    sHDC_CBLK *cblk
    )
{
    UINT32  ret = TERA_SUCCESS;

    ASSERT(cblk != NULL);

    if(!down_interruptible(&cblk->kmp.data_lock))
    {
        cblk->kmp.kbd_lock.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_LOCK);
        cblk->kmp.kbd_lock.tstamp = 0;
        cblk->kmp.kbd_lock.state = ((cblk->kmp.lkbd->led[0] & (1 << LED_NUML))    ? TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_NUM_LOCK : 0) +
                                   ((cblk->kmp.lkbd->led[0] & (1 << LED_CAPSL))   ? TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_CAPS_LOCK : 0) +
                                   ((cblk->kmp.lkbd->led[0] & (1 << LED_SCROLLL)) ? TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_SCROLL_LOCK : 0) +
                                   ((cblk->kmp.lkbd->led[0] & (1 << LED_COMPOSE)) ? TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_COMPOSE_LOCK : 0) +
                                   ((cblk->kmp.lkbd->led[0] & (1 << LED_KANA))    ? TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_KANA_LOCK : 0);
        mTERA_LOG_DBG("Num Lock - %s",    ((cblk->kmp.kbd_lock.state & TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_NUM_LOCK) ? "ON" : "OFF"));
        mTERA_LOG_DBG("Caps Lock - %s",   ((cblk->kmp.kbd_lock.state & TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_CAPS_LOCK) ? "ON" : "OFF"));
        mTERA_LOG_DBG("Scroll Lock - %s", ((cblk->kmp.kbd_lock.state & TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_SCROLL_LOCK) ? "ON" : "OFF"));
        mTERA_LOG_DBG("Compose - %s",     ((cblk->kmp.kbd_lock.state & TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_COMPOSE_LOCK) ? "ON" : "OFF"));
        mTERA_LOG_DBG("Kana - %s",        ((cblk->kmp.kbd_lock.state & TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_KANA_LOCK) ? "ON" : "OFF"));

        up(&cblk->kmp.data_lock);

        // Send the repeat rate and lock state APDUs
        kmp_set_keyboard_repeat_and_lock_apdu_send(cblk);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
        mTERA_LOG_ERR("error processing send HDC_APDU_KMP_KBD_LOCK");
    }

    return ret;
} // kmp_send_kbd_output_data()

// --------------------------------------------------------------------------
//  kmp_ioctl_hdc_open_proc
//      Process a HDC open ioctl
//
//  Description:
//      This function processes a HDC open ioctl. If the SCP channel is ready,
//      then the ioctl is completed immediately with success. Otherwise, it is
//      forwarded to the pending ioctl queue where it will be completed when
//      the SCP channel becomes ready. As such, this ioctl call is potentially
//      blocking.
//
static INT32 kmp_ioctl_hdc_open_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    // Check if the SCP channel is ready
    if (tera_scp_chan_is_ready(cblk->scp_handle,
                               TERA_PCOIP_HDC_APDU_USER_KMP))
    {
        mTERA_LOG_DBG("SCP chan ready, completing HDC_IOCTL_KMP_HDC_OPEN ioctl");

        // SCP channel is up so we complete the ioctl right away
        ret = TERA_SUCCESS;
    }
    else
    {
        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            mTERA_LOG_DBG("SCP chan not ready, waiting to complete HDC_IOCTL_KMP_HDC_OPEN ioctl");
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 (5*msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS))) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                ret = TERA_SUCCESS;
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // kmp_ioctl_hdc_open_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_hdc_close_proc
//      Process a HDC close ioctl
//
//  Description:
//      This function processes a HDC close ioctl. This function resets ALL of
//      the SCP channels, then completes the ioctl.
//
static INT32 kmp_ioctl_hdc_close_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    ASSERT(cblk != NULL);

    // Reset the SCP channels
    tera_scp_reset(cblk->scp_handle);

    // make sure there are no active keys pressed or cursor events when the
    // session ends
    kmp_key_lcur_press_reset(cblk);

    return TERA_SUCCESS;
} // kmp_ioctl_hdc_close_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_feature_activation_apdu_proc
//      Process a KMP feature enable ioctl
//
//  Description:
//      This function processes a KMP feature enable ioctl.
//
static INT32 kmp_ioctl_feature_activation_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                             ret;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN *buf;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN  apdu;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.en = mTERA_UINT32_BYTE_SWAP(apdu.en);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("Sent Feature Activation APDU: APDU=0x%08X, en=%d",
                              mTERA_UINT32_BYTE_SWAP(apdu.apdu_type),
                              mTERA_UINT32_BYTE_SWAP(apdu.en));


                if (apdu.en == 0)
                {
                    // Clear any keyboard or cursor presses
                    kmp_key_lcur_press_reset(cblk);
                }
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // kmp_ioctl_feature_activation_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_feature_status_apdu_proc
//      Process a KMP feature status ioctl
//
//  Description:
//      This function processes a KMP feature status ioctl, which is used to
//      get the status of the KMP features. The status reports the availablility
//      of the local cursor and keyboard, and errors, if any. This function
//      creates a KMP feature status APDU and sends it to the firmware. The
//      ioctl is then placed onto the pending ioctl queue where it will be
//      completed when the reply from the firmware is received.
//
static INT32 kmp_ioctl_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT     *buf;
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_REQ  apdu;

    ASSERT(cblk != NULL);

    // Fill out the APDU
    buf = (sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT *)arg;
    memset(&apdu, 0, sizeof(apdu));
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_STAT_REQ);

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                up(&cblk->pending_ioctl_lock);
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    mTERA_LOG_DBG("Feature status error code=0x%08X", buf->err);
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // kmp_ioctl_feature_status_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_last_feature_status_apdu_proc
//      Process a KMP last feature status ioctl
//
//  Description:
//      This function processes a KMP last feature status ioctl, which returns
//      a copy of the last status of the KMP features. The status reports the
//      availablility of the local cursor and keyboard, and errors, if any.
//
static INT32 kmp_ioctl_last_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                              ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT *buf;

    buf = (sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT *)arg;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    if(!down_interruptible(&cblk->kmp.data_lock))
    {
        // Copy the last feature status msg to the output buffer
        if(copy_to_user(buf,
                        &cblk->kmp.feat_stat,
                        sizeof(sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT)))
        {
            mTERA_LOG_ERR("copy_to_user failed");
            ret = TERA_ERR_FAILURE;
        }
        up(&cblk->kmp.data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HDC_IOCTL_KMP_LAST_FEAT_STAT");
    }

    return ret;
} // kmp_ioctl_last_feature_status_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_keyboard_repeat_rate_apdu_proc
//      Process a keyboard repeat rate set ioctl
//
//  Description:
//      This function processes a keyboard repeat rate set ioctl
//
static INT32 kmp_ioctl_keyboard_repeat_rate_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                               ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT *buf;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT  apdu;

    buf = (sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT *)arg;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // Make a copy of the user's buffer and do byte swapping
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        if(!down_interruptible(&cblk->kmp.data_lock))
        {
            cblk->kmp.kbd_repeat.interval = mTERA_UINT16_BYTE_SWAP(apdu.interval);
            cblk->kmp.kbd_repeat.delay = mTERA_UINT16_BYTE_SWAP(apdu.delay);
            cblk->kmp.kbd_repeat.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
            cblk->kmp.kbd_repeat.tstamp = mTERA_UINT32_BYTE_SWAP(apdu.tstamp);

            up(&cblk->kmp.data_lock);

            // Send the repeat rate and lock state APDUs
            kmp_set_keyboard_repeat_and_lock_apdu_send(cblk);
        }
        else
        {
            ret = TERA_ERR_FAILURE;
        }

        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("error processing HDC_IOCTL_KMP_KBD_REPEAT");
        }
    }
    return ret;
} // kmp_ioctl_keyboard_repeat_rate_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_update_pointer_shape_apdu_proc
//      Process a local cursor shape bitmap ioctl
//
//  Description:
//      This function processes a local cursor shape bitmap ioctl. The shape is
//      forwarded to the firmware in an APDU.
//
//  Notes:
//      The APDU exchanged between the host SW and the FW contains all of the
//      fields in the sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP structure plus
//      the bitmap. There is one exception to this statement, the last field
//      in the structure (apdu_len) is excluded from the message sent to the
//      FW. The length field was added to the Linux code because the Linux
//      ioctl implementation does not allow the driver to determine the length
//      of the buffer supplied by the caller (Windows provides this). This
//      function reads the structure portion of the caller's buffer so it
//      can extract the apdu_len field. Once it knows the length it allocates
//      a buffer and copies the structure and the bitmap into the buffer before
//      sending the buffer to the FW. When the copying is performed this
//      function removes the length field.
//
static INT32 kmp_ioctl_update_pointer_shape_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP  apdu;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP *apdu_and_bmp;
    UINT8                                  *apdu_bmp_addr;
    UINT8                                  *user_bmp_addr;
    UINT32                                  bmp_len;

    //
    // The size of the input buffer varies depending on the size of the cursor
    // bitmap. First read APDU structure, which contains the length of the
    // entire structure (APDU header + bitmap).
    //
    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
#if 0
        mTERA_LOG_DBG("apdu=x%X time=%d type=%d cache=%d x=%d y=%d, w=%d, h=%d, ad=%d, xd=%d, len=%d\n",
                      apdu.apdu_type,    ///< APDU type
                      apdu.tstamp,       ///< Timestamp
                      apdu.bmp_type,     ///< Bitmap type
                      apdu.cache,        ///< Caching {0-6: index; 7: enable}
                      apdu.hot_x,        ///< X hotspot
                      apdu.hot_y,        ///< Y hotspot
                      apdu.width,        ///< Bitmap width
                      apdu.height,       ///< Bitmap height
                      apdu.and_depth,    ///< AND mask depth
                      apdu.xor_depth,    ///< XOR mask depth
                      apdu.apdu_len);    ///< Length of this structure +
#endif

        // Allocate memory to store the APDU + bitmap
        // TBD - if possible replace this with a function in tera_mem.c
        apdu_and_bmp = kzalloc(apdu.apdu_len, GFP_KERNEL);
        if (apdu_and_bmp == NULL)
        {
            mTERA_LOG_ERR("failed to allocate buffer");

            ret = TERA_ERR_NO_MEMORY;
        }
        else
        {
            // copy the APDU to the buffer (this does not copy the bitmap)
            memcpy(apdu_and_bmp, &apdu, sizeof(apdu));

            //
            // copy the bitmap from the user buffer (the bitmap immediately
            // follows the APDU)
            //
            apdu_bmp_addr = (UINT8 *)(&apdu_and_bmp->apdu_len);
                         // (UINT8 *)apdu_and_bmp + sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP) - sizeof(apdu.apdu_len);
            user_bmp_addr = (UINT8 *)buf + sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP);
            bmp_len = apdu.apdu_len - sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP);
            if (copy_from_user(apdu_bmp_addr, user_bmp_addr, bmp_len))
            {
                mTERA_LOG_ERR("copy_from_user failed");
                ret = -ERESTARTSYS;
            }
            else
            {
                // Do HTON byte swappping on the APDU
                apdu_and_bmp->apdu_type = mTERA_UINT32_BYTE_SWAP(apdu_and_bmp->apdu_type);
                apdu_and_bmp->tstamp = mTERA_UINT32_BYTE_SWAP(apdu_and_bmp->tstamp);

                // Send the APDU
                ret = kmp_apdu_send(cblk, apdu_and_bmp, apdu.apdu_len - sizeof(apdu.apdu_len));
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("failed to send APDU");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    mTERA_LOG_DBG("sent update pointer shape: tstamp=%d, bmp type=0x%02X, cache=0x%02X",
                                  apdu.tstamp,
                                  apdu.bmp_type,
                                  apdu.cache);
                    mTERA_LOG_DBG("hot spot(%d, %d), width=%d, height=%d, and_depth=0x%02X, xor_depth=0x%02X",
                                  apdu.hot_x, apdu.hot_y,
                                  apdu.width, apdu.height,
                                  apdu.and_depth, apdu.xor_depth);
                }
            } // end if
            kfree(apdu_and_bmp);
        }
    } // end if
    return ret;
} // kmp_ioctl_update_pointer_shape_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_change_pointer_shape_apdu_proc
//      Process a local cursor shape ioctl
//
//  Description:
//      This function processes a local cursor shape ioctl, which is used to
//      to activate a cached cursor shape stored on the client. This function
//      makes a copy of the APDU in the ioctl input buffer, does HTON byte
//      swapping and then sends it to the firmware. The ioctl is completed in
//      this function.
//
static INT32 kmp_ioctl_change_pointer_shape_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                               ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE  apdu;

    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE *)arg;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // Make a copy of the user's buffer and do byte swapping
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Do HTON byte swappping on the APDU
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.tstamp = mTERA_UINT32_BYTE_SWAP(apdu.tstamp);

        if(!down_interruptible(&cblk->pending_ioctl_lock))
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            up(&cblk->pending_ioctl_lock);

            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("sent change pointer shape: %d",
                              apdu.index);
            }
        }
        else
        {
            ret = TERA_ERR_FAILURE;
        }

        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("error processing HDC_IOCTL_KMP_KBD_REPEAT");
        }
    }
    return ret;
} // kmp_ioctl_change_pointer_shape_apdu_proc()

// --------------------------------------------------------------------------
//  is_cur_data_new
//      Returns non-zero value if host cursor location is not found in the
//      array of the last X client cursor reports.
//
//  Description:
//      This function receives a host cursor data point. It compares this point
//      to the last X received client cursor reports. If the host cursor
//      location is not in the array of the last X received client cursor
//      reports the function returns a non-zero value.
//
//  Return:
//      Zero if the location of the host was found in the last X client reports
//      Non-zero if the host cursor report was not in the array of client
//        reports (meaning the cursor was moved by something other than the
//        client)
//
static INT32 is_cur_data_new(
    sHDC_CBLK *cblk,
    UINT16     abs_x,
    UINT16     abs_y
    )
{
    INT i;

    if (cblk->kmp.disp_tbl.num_active_displays)
    {
        for (i=0; i<MAX_CLIENT_CUR_POINTS; i++)
        {
            // Is this point within the allowed range of this point and the
            // client cursor report is not old
            if (((INT16)abs_x > (INT16)(cblk->kmp.lcur_last_abs_x[i] - cblk->kmp.disp_tbl.desktop_x_tolerance)) &&
                ((INT16)abs_x < (INT16)(cblk->kmp.lcur_last_abs_x[i] + cblk->kmp.disp_tbl.desktop_x_tolerance)) &&
                ((INT16)abs_y > (INT16)(cblk->kmp.lcur_last_abs_y[i] - cblk->kmp.disp_tbl.desktop_y_tolerance)) &&
                ((INT16)abs_y < (INT16)(cblk->kmp.lcur_last_abs_y[i] + cblk->kmp.disp_tbl.desktop_y_tolerance)) &&
                (time_before(jiffies, cblk->kmp.max_age_cur_rep[i])))
            {
                // cursor is not new
#ifdef CUR_REPORT_DBG
                mTERA_LOG_DBG("host cursor data (%d,%d) found at entry %d (%d, %d)",
                              abs_x,
                              abs_y,
                              i,
                              cblk->kmp.lcur_last_abs_x[i],
                              cblk->kmp.lcur_last_abs_y[i]);
#endif
                return 0;
                break;
            }
        }
    }

#ifdef CUR_REPORT_DBG
    // dump the last set of points
    mTERA_LOG_DBG("host cursor data (%d,%d) not found", abs_x, abs_y);
    for (i=0; i<MAX_CLIENT_CUR_POINTS; i+=4)
    {
        mTERA_LOG_DBG("entry %d (%d,%d) (%d,%d) (%d,%d) (%d,%d) ",
                      i,
                      cblk->kmp.lcur_last_abs_x[i], cblk->kmp.lcur_last_abs_y[i],
                      cblk->kmp.lcur_last_abs_x[i+1], cblk->kmp.lcur_last_abs_y[i+1],
                      cblk->kmp.lcur_last_abs_x[i+2], cblk->kmp.lcur_last_abs_y[i+2],
                      cblk->kmp.lcur_last_abs_x[i+3], cblk->kmp.lcur_last_abs_y[i+3]);
    }
#endif

    return 1;
} // is_cur_data_new()

// --------------------------------------------------------------------------
//  kmp_ioctl_set_pointer_location_if_new_proc
//      Process a local cursor position if new ioctl
//
//  Description:
//      This function processes a local cursor position if new IOCTL. This
//      IOCTL updates the location of the client local cursor if the host
//      cursor point is not found in the array of client cursor position data.
//      If the host point is not found it is sent to the client using
//      the TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_POS APDU. The ioctl is completed
//      in this function.
//
static INT32 kmp_ioctl_set_pointer_location_if_new_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                    ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW  cmd;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS         apdu;

    ASSERT(cblk != NULL);

    // Make a copy of the user's buffer and do byte swapping
    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW *)arg;
    if (copy_from_user(&cmd, buf, sizeof(cmd)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // If the cursor point is new send it to the client
        if (is_cur_data_new(cblk, cmd.abs_x, cmd.abs_y))
        {
            // fill in the fields in the APDU
            apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_POS);
            apdu.tstamp = 0;
            apdu.x = mTERA_UINT16_BYTE_SWAP(cmd.abs_x);
            apdu.y = mTERA_UINT16_BYTE_SWAP(cmd.abs_y);

            // acquire the semaphore preventing simultaneous FW requests
            if(down_interruptible(&cblk->pending_ioctl_lock))
            {
                mTERA_LOG_ERR("down_interruptible failed");
                ret = -ERESTARTSYS;
            }
            else
            {
                // Send the APDU and return (there is no response)
                ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                    ret = -ERESTARTSYS;
                }
                else
                {
                    mTERA_LOG_DBG("sent set pointer location: x=%d, y=%d",
                                  cmd.abs_x, cmd.abs_y);
                }
                up(&cblk->pending_ioctl_lock);
            }

            // Add the new position to the array of client reports. This ensures
            // the driver knows the current location of the client cursor. This is
            // necessary because the client does not send a report after moving the
            // cursor to the new position.
            if(!down_interruptible(&cblk->kmp.data_lock))
            {
                cblk->kmp.idx_last_client_pt = (cblk->kmp.idx_last_client_pt + 1) % MAX_CLIENT_CUR_POINTS;
                cblk->kmp.lcur_last_abs_x[cblk->kmp.idx_last_client_pt] = cmd.abs_x;
                cblk->kmp.lcur_last_abs_y[cblk->kmp.idx_last_client_pt] = cmd.abs_y;
                cblk->kmp.max_age_cur_rep[cblk->kmp.idx_last_client_pt] = jiffies + MAX_AGE_CURSOR_REPORT;

                up(&cblk->kmp.data_lock);
            }
            else
            {
                mTERA_LOG_ERR("error adding cursor position to array of client cursor reports");
            }
        }
    }
    return ret;
} // kmp_ioctl_set_pointer_location_if_new_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_set_pointer_location_apdu_proc
//      Process a local cursor position ioctl
//
//  Description:
//      This function processes a local cursor position ioctl, which is used
//      to set the location of the local cursor. This function makes a copy of
//      the APDU in the ioctl input buffer, does HTON byte swapping, then sends
//      it off to the firmware. The ioctl is completed in this function.
//
static INT32 kmp_ioctl_set_pointer_location_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                             ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_POS  apdu;
    UINT32                            tstamp;
    UINT16                            x;
    UINT16                            y;

    ASSERT(cblk != NULL);

    // Make a copy of the user's buffer and do byte swapping
    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_POS *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // save a copy of the unswapped fields
        tstamp = apdu.tstamp;
        x = apdu.x;
        y = apdu.y;

        // Do HTON byte swappping on the APDU
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.tstamp = mTERA_UINT32_BYTE_SWAP(apdu.tstamp);
        apdu.x = mTERA_UINT16_BYTE_SWAP(apdu.x);
        apdu.y = mTERA_UINT16_BYTE_SWAP(apdu.y);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("sent set pointer location: x=%d, y=%d, tstamp=%d",
                              x, y, tstamp);
            }
            up(&cblk->pending_ioctl_lock);
        }

        // Add the new position to the array of client reports. This ensures
        // the driver knows the current location of the client cursor. This is
        // necessary because the client does not send a report after moving the
        // cursor to the new position.
        if(!down_interruptible(&cblk->kmp.data_lock))
        {
            cblk->kmp.idx_last_client_pt = (cblk->kmp.idx_last_client_pt + 1) % MAX_CLIENT_CUR_POINTS;
            cblk->kmp.lcur_last_abs_x[cblk->kmp.idx_last_client_pt] = x;
            cblk->kmp.lcur_last_abs_y[cblk->kmp.idx_last_client_pt] = y;
            cblk->kmp.max_age_cur_rep[cblk->kmp.idx_last_client_pt] = jiffies + MAX_AGE_CURSOR_REPORT;

            up(&cblk->kmp.data_lock);
        }
        else
        {
            mTERA_LOG_ERR("error adding cursor position to array of client cursor reports");
        }
    }
    return ret;
} // kmp_ioctl_set_pointer_location_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_set_pointer_visibility_apdu_proc
//      Process a local cursor visibility ioctl
//
//  Description:
//      This function processes a local cursor visibility ioctl, which is used
//      to show/hide the local cursor overlay. This function makes a copy of
//      the APDU in the ioctl input buffer, does HTON byte swapping, then sends
//      it off to the firmware. The ioctl is completed in this function.
//
static INT32 kmp_ioctl_set_pointer_visibility_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                             ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS  apdu;
    UINT32                            tstamp;

    ASSERT(cblk != NULL);

    // Make a copy of the user's buffer and do byte swapping
    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // save a copy of the unswapped fields
        tstamp = apdu.tstamp;

        // Do HTON byte swappping on the APDU
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.tstamp = mTERA_UINT32_BYTE_SWAP(apdu.tstamp);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("sent set pointer visibility: visibility=%d, tstamp=%d",
                              apdu.visible, tstamp);
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // kmp_ioctl_set_pointer_visibility_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_fade_out_control_apdu_proc
//      Process a local cursor idle hide ioctl
//
//  Description:
//      This function processes a local cursor idle hide ioctl, which is used
//      to set the amount of idle time (in seconds) before the local cursor
//      overlay is hidden. This function makes a copy of the APDU in the ioctl
//      input buffer, does HTON byte swapping, then sends it off to the
//      firmware. The ioctl is completed in this function.
//
static INT32 kmp_ioctl_fade_out_control_apdu_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                   ret;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE *buf;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE  apdu;
    UINT32                                  time;

    ASSERT(cblk != NULL);

    // Make a copy of the user's buffer and do byte swapping
    buf = (sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // save a copy of the unswapped fields
        time = apdu.time;

        // Do HTON byte swappping on the APDU
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.time = mTERA_UINT32_BYTE_SWAP(apdu.time);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("sent fade out control: interval=%d", time);
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // kmp_ioctl_fade_out_control_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_mouse_speed_query_apdu_get_proc
//      Process a local cursor speed get ioctl
//
//  Description:
//      This function processes a local cursor speed get ioctl, which is used
//      to get the local cursor speed from the client. This function creates a
//      local cursor speed get APDU and sends it to the firmware. The ioctl is
//      then placed onto the pending ioctl queue where it will be completed
//      when the reply from the firmware is received.
//
static INT32 kmp_ioctl_mouse_speed_query_apdu_get_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    )
{
    INT32                                     ret;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD    *buf;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    buf = (sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD *)arg;
    memset(&apdu, 0, sizeof(apdu));
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ);

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                up(&cblk->pending_ioctl_lock);
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    mTERA_LOG_DBG("local cursor speed=%d", cblk->ioctl_rsp.data.cur_speed.spd);
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // kmp_ioctl_mouse_speed_query_apdu_get_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_mouse_speed_update_apdu_set_proc
//      Process a local cursor speed set ioctl
//
//  Description:
//      This function processes a local cursor speed set ioctl. This function
//      creates a local cursor speed get APDU and sends it to the firmware to
//      set the speed. The ioctl is completed in this function.
//
static INT32 kmp_ioctl_mouse_speed_update_apdu_set_proc(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    INT32                                  ret;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD *buf;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD  apdu;
    UINT32                                 speed;

    ASSERT(cblk != NULL);

    // Make a copy of the user's buffer and do byte swapping
    buf = (sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD *)arg;
    if (copy_from_user(&apdu, buf, sizeof(apdu)))
    {
        mTERA_LOG_ERR("copy_from_user failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // save a copy of the unswapped fields
        speed = apdu.spd;

        // Do HTON byte swappping on the APDU
        apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(apdu.apdu_type);
        apdu.spd = mTERA_UINT32_BYTE_SWAP(apdu.spd);

        // acquire the semaphore preventing simultaneous FW requests
        if(down_interruptible(&cblk->pending_ioctl_lock))
        {
            mTERA_LOG_ERR("down_interruptible failed");
            ret = -ERESTARTSYS;
        }
        else
        {
            // Send the APDU and return (there is no response)
            ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
                ret = -ERESTARTSYS;
            }
            else
            {
                mTERA_LOG_DBG("sent mouse speed update: speed=%d", speed);
            }
            up(&cblk->pending_ioctl_lock);
        }
    }
    return ret;
} // kmp_ioctl_mouse_speed_update_apdu_set_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_keyboard_repeat_rate_query_apdu_get_proc
//      Process a local keyboard repeat rate get ioctl
//
//  Description:
//      This function processes a local keyboard repeat rate get ioctl, which
//      reads the local keyboard repeat rate settings from the client.
//      Only the zero client supports the keyboard repeat rate get APDU. The
//      soft client ignores this APDU.
//
//      This function creates a local keyboard repeat rate get APDU and sends
//      it to the firmware. The ioctl is then placed on the pending ioctl
//      queue where it will be completed when the reply from the firmware is
//      received.
//
static INT32 kmp_ioctl_keyboard_repeat_rate_query_apdu_get_proc(
    sHDC_CBLK *cblk,
    unsigned long arg,
    unsigned int ioctl
    )
{
    INT32                                     ret;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT      *buf;
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ apdu;

    ASSERT(cblk != NULL);

    memset(&apdu, 0, sizeof(apdu));

    // Fill out the APDU
    buf = (sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT *)arg;
    memset(&apdu, 0, sizeof(apdu));
    apdu.apdu_type = mTERA_UINT32_BYTE_SWAP(TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ);

    // acquire the semaphore preventing simultaneous FW requests
    if(down_interruptible(&cblk->pending_ioctl_lock))
    {
        mTERA_LOG_ERR("down_interruptible failed");
        ret = -ERESTARTSYS;
    }
    else
    {
        // Send the APDU
        ret = kmp_apdu_send(cblk, &apdu, sizeof(apdu));
        if (ret != TERA_SUCCESS)
        {
            up(&cblk->pending_ioctl_lock);
            mTERA_LOG_ERR("failed to send APDU (ret=%d)", ret);
            ret = -ERESTARTSYS;
        }
        else
        {
            //
            // Set flag equal to the outstanding ioctl, the flag is inspected
            // when the workqueue is inspected.
            //
            cblk->pending_ioctl_flag = ioctl;

            // wait until the response is received on the ingress message queue and
            // processed by the dpc_entry
            if (wait_event_interruptible_timeout(cblk->pending_ioctl_q,
                                                 (cblk->pending_ioctl_flag == 1),
                                                 msecs_to_jiffies(TERA_IOCTL_TIMEOUT_MS)) <= 0)
            {
                up(&cblk->pending_ioctl_lock);
                mTERA_LOG_ERR("wait_event_interruptible_timeout failed");
                cblk->pending_ioctl_flag = 1;
                ret = -ERESTARTSYS;
            }
            else
            {
                // fill in the IOCTL response
                if(copy_to_user(buf, &cblk->ioctl_rsp, sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT)))
                {
                    mTERA_LOG_ERR("copy_to_user failed");
                    ret = -ERESTARTSYS;
                }
                else
                {
                    ret = TERA_SUCCESS;
                }
                up(&cblk->pending_ioctl_lock);
            }
        }
    }
    return ret;
} // kmp_ioctl_keyboard_repeat_rate_query_apdu_get_proc()

// --------------------------------------------------------------------------
//  kmp_ioctl_disp_topo_set
//      Store a copy of the active display topology
//
//  Description:
//      This function stores a copy of the active display topology in the
//      HDC control block. This information is used to detect when the
//      local cursor moves from one screen to another. It is also used to
//      convert the client cursor coordinates to coordinates understood by
//      the host.
//
static INT32 kmp_ioctl_disp_topo_set(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                                      ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE *buf;
    INT                                         i;

    buf = (sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE *)arg;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // lock the control block
    if(!down_interruptible(&cblk->kmp.data_lock))
    {
        // Copy the topology data to the control block
        if (copy_from_user(&cblk->kmp.disp_tbl, buf, sizeof(sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE)))
        {
            mTERA_LOG_ERR("copy_from_user failed");
            ret = -ERESTARTSYS;
        }
        up(&cblk->kmp.data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (TERA_SUCCESS == ret)
    {
        mTERA_LOG_DBG("active %d, min (%d, %d), max (%d, %d), size desk %d X %d primary %d X %d",
                      cblk->kmp.disp_tbl.num_active_displays,
                      cblk->kmp.disp_tbl.min_x,
                      cblk->kmp.disp_tbl.min_y,
                      cblk->kmp.disp_tbl.max_x,
                      cblk->kmp.disp_tbl.max_y,
                      cblk->kmp.disp_tbl.desktop_width,
                      cblk->kmp.disp_tbl.desktop_height,
                      cblk->kmp.disp_tbl.primary_width,
                      cblk->kmp.disp_tbl.primary_height);

        for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
        {
            mTERA_LOG_DBG("d%d: act %d, pix min (%4d, %4d) max (%4d, %4d), abs min (%5d, %5d) max (%5d, %5d), size %4d X %4d, rot %d",
                          i,
                          cblk->kmp.disp_tbl.displays[i].active,
                          cblk->kmp.disp_tbl.displays[i].min_pix_x,
                          cblk->kmp.disp_tbl.displays[i].min_pix_y,
                          cblk->kmp.disp_tbl.displays[i].max_pix_x,
                          cblk->kmp.disp_tbl.displays[i].max_pix_y,
                          cblk->kmp.disp_tbl.displays[i].min_abs_x,
                          cblk->kmp.disp_tbl.displays[i].min_abs_y,
                          cblk->kmp.disp_tbl.displays[i].max_abs_x,
                          cblk->kmp.disp_tbl.displays[i].max_abs_y,
                          cblk->kmp.disp_tbl.displays[i].width,
                          cblk->kmp.disp_tbl.displays[i].height,
                          cblk->kmp.disp_tbl.displays[i].rotation);
        }
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HDC_IOCTL_KMP_DISP_TOPO_SET");
    }

    return ret;
} // kmp_ioctl_disp_topo_set()

// --------------------------------------------------------------------------
//  kmp_ioctl_driver_cfg_set
//      Update the driver configuration settings
//
//  Description:
//      This function updates the driver configurations settings.
//
static INT32 kmp_ioctl_driver_cfg_set(
    sHDC_CBLK *cblk,
    unsigned long arg
    )
{
    UINT32                               ret = TERA_SUCCESS;
    sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL *buf;

    buf = (sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL *)arg;

    ASSERT(cblk != NULL);
    ASSERT(buf != NULL);

    // lock the control block
    if(!down_interruptible(&cblk->kmp.data_lock))
    {
        // Copy the topology data to the control block
        if (copy_from_user(&cblk->kmp.driver_cfg, buf, sizeof(sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL)))
        {
            mTERA_LOG_ERR("copy_from_user failed");
            ret = -ERESTARTSYS;
        }
        up(&cblk->kmp.data_lock);
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    if (TERA_SUCCESS == ret)
    {
        mTERA_LOG_DBG("cursor event processing mode: %s",
                      lcur_evt_alg_str_tbl[cblk->kmp.driver_cfg.lcur_mode]);
    }

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("error processing HDC_IOCTL_KMP_DRIVER_CFG_SET");
    }

    return ret;
} // kmp_ioctl_driver_cfg_set()

// --------------------------------------------------------------------------
//  kmp_apdu_kbd_update_ps2_mod_state
//      Update the state of the USB modifier field
//
//  Description:
//      This function updates the state of the modifier mask. USB kbd devices
//      report to the upper layer driver keycodes and modifier state data.
//      There are 8 modifiers that correspond to the state of 8 special keys
//      on the keyboard (left-[alt, ctrl, shift and win] and right-[alt, ctrl,
//      shift and win]). This function looks at each key reported to the
//      upper level driver. If it's one of the special keys it then
//      updates the modifier bitmask based on the state of the key.
//
static void kmp_apdu_kbd_update_ps2_mod_state(
    UINT32     usb_code,
    UINT8      state,
    UINT8     *mod
    )
{
    UINT8   mask;

    // check to see if the key is one of the special keys
    switch (usb_code)
    {
        case LEFT_CTRL_KEY_USB_CODE:
            mask = LEFT_CTRL_KEY_MOD_BIT_MSK;
            break;
        case LEFT_SHFT_KEY_USB_CODE:
            mask = LEFT_SHFT_KEY_MOD_BIT_MSK;
            break;
        case LEFT_ALT_KEY_USB_CODE:
            mask = LEFT_ALT_KEY_MOD_BIT_MSK;
            break;
        case LEFT_WIN_KEY_USB_CODE:
            mask = LEFT_WIN_KEY_MOD_BIT_MSK;
            break;
        case RIGHT_CTRL_KEY_USB_CODE:
            mask = RIGHT_CTRL_KEY_MOD_BIT_MSK;
            break;
        case RIGHT_SHFT_KEY_USB_CODE:
            mask = RIGHT_SHFT_KEY_MOD_BIT_MSK;
            break;
        case RIGHT_ALT_KEY_USB_CODE:
            mask = RIGHT_ALT_KEY_MOD_BIT_MSK;
            break;
        case RIGHT_WIN_KEY_USB_CODE:
            mask = RIGHT_WIN_KEY_MOD_BIT_MSK;
            break;
        default:
            mask = 0;
            break;
    }

    // if the key is a modifier key
    if(mask != 0)
    {
        // set the mask bit if the key is pressed
        if(state)
        {
            *mod = *mod | mask;
        }
        // clear the mask bit if the key was released
        else
        {
            *mod = *mod & ~mask;
        }
    }
} // kmp_apdu_kbd_update_ps2_mod_state()

// --------------------------------------------------------------------------
//  kmp_apdu_kbd_conv_ps2_to_usb
//      Convert a PS/2 keyboard scancode to a USB scancode
//
//  Description:
//      This function converts a PS/2 scancode to a USB scancode. The code
//      assumes the scancode is a PS/2 Set 1 code.
//
//      Notes about special keys:
//          - PS/2 keys 2B and AB map to 07-31 or 07-32 based on the language
//            field received in the Set Keyboard Locale APDU
//          - PS/2 keys E05E and E0DE map to 07-66, the translation table
//            includes an entry saying this key should map to 01-81
//          - PS/2 keys 76 and F6 map to 07-94, the translation table includes
//            an entry saying these keys should map to 07-73
//          - The soft client sends the value 00000100 for the Pause key
//            instead of 00E11D45 or 00E19DC5. Currently this key is converted
//            to the USB code 07-48 but the agent also reports other keys as
//            the pause key because the soft client reports the same value
//            (00000100) for other keys (mute, volume up/down, ...). At some
//            point the soft client needs to be modified so that it only reports
//            the value 00000100 for the pause key.
//
//  Returns:
//      TERA_SUCCESS if the scancode was successfully converted
//      TERA_ERR_FAILURE if the function failed to convert the scancode
//
static INT32 kmp_apdu_kbd_conv_ps2_to_usb(
    UINT32     ps2_code,
    UINT16     lang,
    UINT32    *usb_code
    )
{
    INT32 ret = TERA_SUCCESS;
    INT32 code;

    // Handle the codes where the upper 24 bits equal 0x000000
    if((ps2_code & UPPER_24_MASK) == 0)
    {
        // lookup the USB code
        code = (UINT32)ps2_lkup_table1[ps2_code & 0x000000FF];

        // append the USB page if a valid code was returned
        if(code != PS2_LKUP_UNUSED)
        {
            *usb_code = (code & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_CODE) | 0x00070000;
        }
        else
        {
            ret = TERA_ERR_FAILURE;
        }

        // handle the special mapping of the 002B/00AB key
        if((ps2_code == 0x2B) || (ps2_code == 0xAB))
        {
            if((lang == KBD_LANG_USA) ||
               (lang == KBD_LANG_JAPANESE) ||
               (lang == KBD_LANG_KOREAN))
            {
                *usb_code = 0x00070031;
            }
            mTERA_LOG_DBG("mapped PS/2 key 0x%04x to USB scancode 0x%08x (lang=0x%04x)",
                          ps2_code, *usb_code, lang);
        }
    }
    //
    // Handle the codes where the upper 24 bits equal 0x000001 (note: read the
    // note at the beginning of the ps2_lkup_table2 lookup table)
    //
    else if((ps2_code & UPPER_24_MASK) == 0x00000100)
    {
        // lookup the USB page and code
        *usb_code = (UINT32)ps2_lkup_table2[ps2_code & 0x000000FF];

        // check if a valid page and code were returned
        if(*usb_code == PS2_LKUP_UNUSED)
        {
            ret = TERA_ERR_FAILURE;
        }
    }
    else
    {
        ret = TERA_ERR_FAILURE;
    }

    return ret;
} // kmp_apdu_kbd_conv_ps2_to_usb()

// --------------------------------------------------------------------------
//  kmp_update_keyboard_apdu_proc
//      Process a KMP local keyboard event APDU
//
//  Description:
//      This function processes a KMP local keyboard event APDU. Multiple
//      keyboard events may be batched up in a single message. This function
//      extracts each event and puts it into the kmp event message queue,
//      where it will be service by the kmp thread.
//
//      This function also deals with translating PS/2 keyboard scancodes
//      to USB scancodes if the keyboard event indicates the key is a PS/2
//      key.
//
// TBD - the 1.4.5 linux code updated the event timestamp using the jiffies
//       variable, I'm not sure if this needs to be added in here
//
static void kmp_update_keyboard_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT *kbd_evt = apdu;
    UINT32                            i;
    UINT32                            num_evt;
    sTERA_MSG                        *msg;
    UINT32                            usb_code = 0;

    ASSERT(cblk != NULL);

    // Calculate number of keyboard event APDUs in the buffer
    num_evt = apdu_len / sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT);

    if ((apdu == NULL) || (num_evt == 0) ||
        ((apdu_len % sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT)) != 0))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=multiple of %u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT));
    }
    else
    {
        for (i = 0; i < num_evt; i++)
        {
            // Do NTOH byte swaps on the fields
            kbd_evt[i].tstamp = mTERA_UINT32_BYTE_SWAP(kbd_evt[i].tstamp);
            kbd_evt[i].scancode = mTERA_UINT32_BYTE_SWAP(kbd_evt[i].scancode);
            kbd_evt[i].lang = mTERA_UINT16_BYTE_SWAP(kbd_evt[i].lang);
            kbd_evt[i].layout = mTERA_UINT16_BYTE_SWAP(kbd_evt[i].layout);

            // for the time being discard unicode keys
            if (mTERA_UNICODE_KEYCODE(kbd_evt[i].state))
            {
                mTERA_LOG_ERR("Received unsupport unicode keyboard event (state=0x%X)",
                              kbd_evt[i].state);
                continue;
            }

            // Process PS/2 keyboard events - this happens when the key
            // is not a USB scancode and not a Unicode key
            if ((!(kbd_evt[i].flags &
                  TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_FLAGS_USB_SCANCODE)) &&
                (!mTERA_UNICODE_KEYCODE(kbd_evt[i].state)))
            {
                // Perform PS/2 to USB scancode conversion, exit if the
                // conversion fails
                if(kmp_apdu_kbd_conv_ps2_to_usb(kbd_evt[i].scancode, cblk->kmp.lang, &usb_code) != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("Failed to convert PS/2 key (discarding)");
                    mTERA_LOG_ERR("tstamp=%u, scancode=0x%08x, lang=0x%04x, layout=0x%04x",
                                  kbd_evt[i].tstamp, kbd_evt[i].scancode,
                                  kbd_evt[i].lang, kbd_evt[i].layout);
                    mTERA_LOG_ERR("state=%u, flags=%u, lock=%u, modifiers=%u",
                                  kbd_evt[i].state, kbd_evt[i].flags,
                                  kbd_evt[i].lock, kbd_evt[i].mod);
                    continue;
                }
                else
                {
#if KBD_REPORT_DBG
                    mTERA_LOG_DBG("converted PS/2 key 0x%08x to USB 0x%08x",
                                  kbd_evt[i].scancode, usb_code);
#endif

                    // update the modifier field if necessary
                    kmp_apdu_kbd_update_ps2_mod_state(usb_code,
                                                      kbd_evt[i].state,
                                                      &cblk->kmp.modifier);
                    kbd_evt[i].mod = cblk->kmp.modifier;
                }
                kbd_evt[i].scancode = usb_code;
            }

            // Allocate a message
            msg = tera_msg_alloc(cblk->dev,
                                 sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT), 1);
            if (msg == NULL)
            {
                mTERA_LOG_ERR("failed to allocate message");

                //TODO This is really bad!! We should ASSERT here
                continue;
            }

            // Copy the keyboard event into the message
            memcpy(mTERA_MSG_BUF_GET(msg), &kbd_evt[i],
                   sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT));

            // Enqueue the message to the keyboard event Q
            tera_msg_q_enqueue(&cblk->kmp.kbd_evt_msg_q, msg);
        } // end for
    } // end if

    // process keyboard output data if pending
    if(test_and_clear_bit(1, &kbd_event_arrived))
    {
        kmp_send_kbd_output_data(cblk);
    }

} // kmp_update_keyboard_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_update_mouse_apdu_proc
//      Process a KMP local cursor event APDU
//
//  Description:
//      This function processes a KMP local cursor event APDU.
//
// TBD - the 1.4.5 linux code updated the event timestamp using the jiffies
//       variable, I'm not sure if this needs to be added in here
//
static void kmp_update_mouse_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT *cur_evt = apdu;
    UINT32                            i;
    UINT32                            num_evt;
    sTERA_MSG                        *msg;

    ASSERT(cblk != NULL);

    // Calculate number of cursor event APDUs in the buffer
    num_evt = apdu_len / sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT);

    if ((apdu == NULL) || (num_evt == 0) ||
        ((apdu_len % sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT)) != 0))

    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=multiple of %u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT));
    }
    else
    {
        for (i = 0; i < num_evt; i++)
        {
            // Do NTOH byte swaps on the fields
            cur_evt[i].tstamp = mTERA_UINT32_BYTE_SWAP(cur_evt[i].tstamp);
            cur_evt[i].abs_x = mTERA_UINT16_BYTE_SWAP(cur_evt[i].abs_x);
            cur_evt[i].abs_y = mTERA_UINT16_BYTE_SWAP(cur_evt[i].abs_y);
            cur_evt[i].rel_x = mTERA_UINT16_BYTE_SWAP(cur_evt[i].rel_x);
            cur_evt[i].rel_y = mTERA_UINT16_BYTE_SWAP(cur_evt[i].rel_y);
            cur_evt[i].wheel = -1 * mTERA_UINT16_BYTE_SWAP(cur_evt[i].wheel);

            // Allocate a message
            msg = tera_msg_alloc(cblk->dev,
                                 sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT), 1);
            if (msg == NULL)
            {
                mTERA_LOG_ERR("failed to allocate message");

                //TODO This is really bad!! We should ASSERT here
                continue;
            }

            // Copy the cursor event into the message
            memcpy(mTERA_MSG_BUF_GET(msg), &cur_evt[i],
                   sizeof(sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT));

            // Enqueue the message to the message queue
            tera_msg_q_enqueue(&cblk->kmp.cur_evt_msg_q, msg);
        } // end for
    } // end if
} // kmp_update_mouse_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_feature_status_apdu_proc
//      Process a KMP feature status APDU
//
//  Description:
//      This function processes a KMP feature status APDU from the firmware.
//
static void kmp_feature_status_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT *buf;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT));
    }
    else
    {
        // if the correct IOCTL is not pending
        buf = apdu;
        if (cblk->pending_ioctl_flag != TERA_PCOIP_HDC_IOCTL_KMP_FEAT_STAT)
        {
            // Update the copy of the last feature status msg
            down(&cblk->kmp.data_lock);
            cblk->kmp.feat_stat.err = mTERA_UINT32_BYTE_SWAP(buf->err);
            cblk->kmp.feat_stat.cnt++;
            up(&cblk->kmp.data_lock);

            // Ioctl could have been cancelled
            mTERA_LOG_INFO("no pending HDC_IOCTL_KMP_FEAT_STAT: err=0x%08x, count=%d",
                           cblk->kmp.feat_stat.err,
                           cblk->kmp.feat_stat.cnt);
        }
        else
        {
            // Store the response in the control block
            down(&cblk->data_lock);
            cblk->ioctl_rsp.data.feat_stat.apdu_type = mTERA_UINT32_BYTE_SWAP(buf->apdu_type);
            cblk->ioctl_rsp.data.feat_stat.err = mTERA_UINT32_BYTE_SWAP(buf->err);
            up(&cblk->data_lock);

            // set the flag that will wake up the waiting ioctl thread
            cblk->pending_ioctl_flag = 1;

            // wake up the waiting thread
            wake_up_interruptible(&cblk->pending_ioctl_q);
        } // end if
    }
} // kmp_feature_status_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_set_keyboard_locale_apdu_proc
//      Process a KMP local keyboard locale APDU
//
//  Description:
//      This function processes a KMP local keyboard locale APDU from the
//      firmware.
//
static void kmp_set_keyboard_locale_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE *kbd_locale = apdu;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE));
    }
    else
    {
        mTERA_LOG_DBG("rcvd set kbd locale: lang=0x%04x, layout=0x%04x",
                      mTERA_UINT16_BYTE_SWAP(kbd_locale->lang),
                      mTERA_UINT16_BYTE_SWAP(kbd_locale->layout));

        //
        // Send the keyboard repeat rate and lock state to the client as part
        // of the keyboard sync event
        //
        kmp_set_keyboard_repeat_and_lock_apdu_send(cblk);

        //
        // store the language field, it is used when mapping PS/2 keys to USB
        // scancodes
        //
        cblk->kmp.lang = mTERA_UINT16_BYTE_SWAP(kbd_locale->lang);
    }
} // kmp_set_keyboard_locale_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_set_keyboard_repeat_and_lock_apdu_send
//      Send KMP keyboard repeat rate and lock state APDUs
//
//  Description:
//      This function sends the keyboard repeat rate and lock state APDUs to
//      the firmware. The APDUs are only sent if their respective keyboard
//      repeat rate/lock states APDUs are valid in the control block. This
//      function acquires the data lock, so the calling function must NOT have
//      acquired the lock, else a deadlock will occur.
//
static void kmp_set_keyboard_repeat_and_lock_apdu_send(
    sHDC_CBLK *cblk
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT kbd_repeat;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK   kbd_lock;

    ASSERT(cblk != NULL);

    //
    // Make a copy of the keyboard repeat and lock APDUs stored in the control
    // block. The data lock is acquired and held while the copy takes place.
    //
    down(&cblk->kmp.data_lock);
    memcpy(&kbd_repeat, &cblk->kmp.kbd_repeat, sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT));
    memcpy(&kbd_lock, &cblk->kmp.kbd_lock, sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK));
    up(&cblk->kmp.data_lock);

    // send the keyboard repeat rate APDU if valid in the control block
    if (kbd_repeat.apdu_type != 0)
    {
        // Send the keyboard repeat APDU
        if (kmp_apdu_send(cblk, &kbd_repeat,
                          sizeof(kbd_repeat)) != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to send kbd_repeat APDU");

            // SCP channel may not yet be up
        }
#ifdef DBG
        else
        {
            UINT16 interval = mTERA_UINT16_BYTE_SWAP(kbd_repeat.interval);
            UINT16 delay = mTERA_UINT16_BYTE_SWAP(kbd_repeat.delay);
            UINT32 tstamp = mTERA_UINT32_BYTE_SWAP(kbd_repeat.tstamp);
            mTERA_LOG_DBG("sent kbd repeat rate: tstamp=%d, interval=%d, delay=%d",
                          tstamp, interval, delay);
        }
#endif
    }
    else
    {
        mTERA_LOG_DBG("Not sending invalid kbd repeat rate apdu: 0x%08x",
                      kbd_repeat.apdu_type);
    }

    // send the keyboard lock APDU if valid in the control block
    if (kbd_lock.apdu_type != 0)
    {
        // Send the keyboard lock APDU
        if (kmp_apdu_send(cblk, &kbd_lock,
                          sizeof(kbd_lock)) != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to send lock APDU");

            // SCP channel may not yet be up
        }
#ifdef DBG
        else
        {
            UINT32 tstamp = mTERA_UINT32_BYTE_SWAP(kbd_repeat.tstamp);
            mTERA_LOG_DBG("sent kbd lock status: tstamp=%d, state=%d",
                          tstamp, kbd_lock.state);
        }
#endif
    }
    else
    {
        mTERA_LOG_DBG("Not sending invalid kbd lock status apdu: 0x%08x",
                      kbd_lock.apdu_type);
    }
} // kmp_set_keyboard_repeat_and_lock_apdu_send()

// --------------------------------------------------------------------------
//  kmp_mouse_speed_update_apdu_proc
//      Process a local cursor speed APDU
//
//  Description:
//      This function processes a KMP mouse speed update APDU from the
//      firmware. This function looks for a pending KMP_CUR_SPD_GET
//      ioctl and copies the contents of the received APDU to the
//      output buffer of the ioctl and completes it.
//
static void kmp_mouse_speed_update_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD *buf = apdu;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD));
    }
    else
    {
        // if the correct IOCTL is not pending
        if (cblk->pending_ioctl_flag != TERA_PCOIP_HDC_IOCTL_KMP_CUR_SPD_GET)
        {
            // Ioctl could have been cancelled
            mTERA_LOG_INFO("no pending HDC_IOCTL_KMP_CUR_SPD_GET");
        }
        else
        {
            down(&cblk->data_lock);

            // Store the response in the control block
            cblk->ioctl_rsp.data.cur_speed.apdu_type = mTERA_UINT32_BYTE_SWAP(buf->apdu_type);
            cblk->ioctl_rsp.data.cur_speed.spd = mTERA_UINT32_BYTE_SWAP(buf->spd);

            up(&cblk->data_lock);

            // set the flag that will wake up the waiting ioctl thread
            cblk->pending_ioctl_flag = 1;

            // wake up the waiting thread
            wake_up_interruptible(&cblk->pending_ioctl_q);
        } // end if
    }
} // kmp_mouse_speed_update_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_set_keyboard_repeat_rate_apdu_proc
//      Process a local keyboard repeat rate APDU
//
//  Description:
//      This function processes a keyboard repeat rate APDU from the firmware.
//      This function looks for a pending ioctl and copies the keyboard repeat
//      rate APDU data to the output buffer of the ioctl and completes it.
//
static void kmp_set_keyboard_repeat_rate_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT *buf = apdu;

    ASSERT(cblk != NULL);

    if ((apdu == NULL) ||
        (apdu_len != sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT)))
    {
        mTERA_LOG_ERR("incorrect APDU length=%u (expected=%u)",
                      apdu_len, (UINT32)sizeof(sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT));
    }
    else
    {
        // if the correct IOCTL is not pending
        if (cblk->pending_ioctl_flag != TERA_PCOIP_HDC_IOCTL_KMP_KBD_REPEAT_GET)
        {
            // Ioctl could have been cancelled
            mTERA_LOG_INFO("no pending HDC_IOCTL_KMP_KBD_REPEAT_GET");
        }
        else
        {
            down(&cblk->data_lock);

            // Copy the keyboard repeat rate from the APDU to the output buffer
            // and do NTOH byte swapping on the fields
            cblk->ioctl_rsp.data.kbd_repeat.apdu_type = mTERA_UINT32_BYTE_SWAP(buf->apdu_type);
            cblk->ioctl_rsp.data.kbd_repeat.tstamp = mTERA_UINT32_BYTE_SWAP(buf->tstamp);
            cblk->ioctl_rsp.data.kbd_repeat.interval = mTERA_UINT16_BYTE_SWAP(buf->interval);
            cblk->ioctl_rsp.data.kbd_repeat.delay = mTERA_UINT16_BYTE_SWAP(buf->delay);

            mTERA_LOG_DBG("rcvd kbd repeat rate: tstamp=%d, interval=%d, delay=%d",
                          cblk->ioctl_rsp.data.kbd_repeat.tstamp,
                          cblk->ioctl_rsp.data.kbd_repeat.interval,
                          cblk->ioctl_rsp.data.kbd_repeat.delay);

            up(&cblk->data_lock);

            // set the flag that will wake up the waiting ioctl thread
            cblk->pending_ioctl_flag = 1;

            // wake up the waiting thread
            wake_up_interruptible(&cblk->pending_ioctl_q);
        } // end if
    }
} // kmp_set_keyboard_repeat_rate_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_thread_update_keyboard_apdu_proc
//      Process a local keyboard event in the background thread
//
//  Description:
//      This function processes a local keyboard event in the background
//      thread. If there is a oustanding local keyboard event waiting to be
//      injected back into the kernel, this function removes it from the
//      keyboard event message queue and copies the event into a HID_READ_REPORT
//      ioctl and completes it.
//
static INT kmp_thread_update_keyboard_apdu_proc(
    sHDC_CBLK *cblk
    )
{
    INT                               ret = 0;
    sTERA_MSG                        *msg;
    sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT *kbd_evt;
    UINT32                            key_code;
    UINT32                            key_scancode;
    UINT32                            key_page;
    unsigned int                      report_key;

    ASSERT(cblk != NULL);

    // Get a local keyboard event, if any, from the message queue
    msg = tera_msg_q_dequeue(&cblk->kmp.kbd_evt_msg_q);
    if (msg != NULL)
    {
        kbd_evt = mTERA_MSG_BUF_GET(msg);
        key_scancode = kbd_evt->scancode;
        key_code =  key_scancode & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_CODE;
        key_page = (key_scancode & TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_PAGE) >> 16;

#if KBD_REPORT_DBG
        mTERA_LOG_DBG("rcvd scancode %08X state %d rpt_en %d key_break_pend %d last_key_press %X",
                      key_scancode,
                      mTERA_KEY_PRESS_EVENT(kbd_evt->state),
                      is_client_kbd_repeat_en(cblk),
                      cblk->kmp.kbd_last_key_break_pend,
                      cblk->kmp.kbd_last_key_press);
#endif
        //
        // If this is a key up event and the keycode is from the keyboard page
        // (HID page 7) the event might need to be discarded based on where
        // repeat detection is done and the value of the last key pressed.
        //
        // * If the client IS NOT performing keyboard repeat processing discard
        // all non control (shift, alt, ctrl, gui) key press up transitions
        // except for last key press.
        //
        // * If the client IS performing keyboard repeat processing discard all
        // non control (shift, alt, ctrl, gui) key press up transitions when
        // a key break is not pending.
        //
        if ((mTERA_KEYBOARD_PAGE_KEYCODE(key_scancode)) &&
            (!mTERA_KEY_PRESS_EVENT(kbd_evt->state)) &&
            (key_code < FIRST_MODIFIER_USB_CODE) &&
            ((is_client_kbd_repeat_en(cblk) && !cblk->kmp.kbd_last_key_break_pend) ||
             (!is_client_kbd_repeat_en(cblk) && (key_code != cblk->kmp.kbd_last_key_press))))
        {
            // Free the message
#if KBD_REPORT_DBG
            mTERA_LOG_DBG("Discarding keyboard key input event key_code=0x%X kbd_last_key_press=0x%X",
                          key_code, cblk->kmp.kbd_last_key_press);
#endif
            tera_msg_free(cblk->dev, msg);
        }

        //
        // If this is a key up event and the keycode is from the consumer page
        // (0x0C) the event is discarded when the client performs
        // repeat detection OR when the client is not performing repeat
        // detection and the key is not the last pressed consumer key.
        //
        else if ((mTERA_CONSUMER_PAGE_KEYCODE(key_scancode)) &&
                 (!mTERA_KEY_PRESS_EVENT(kbd_evt->state)) &&
                 ((is_client_kbd_repeat_en(cblk)) ||
                  (!is_client_kbd_repeat_en(cblk) && (cblk->kmp.last_consumer_key != key_scancode))))
        {
            // Free the message
            mTERA_LOG_DBG("Discarding consumer key input event");
            tera_msg_free(cblk->dev, msg);
        }
        //
        // If this is a key up event and the keycode is from the generic
        // desktop page (0x01) the event is discarded when the client performs
        // repeat detection OR when the client is not performing repeat
        // detection and the key is not the last pressed generic desktop key.
        //
        else if ((mTERA_GEN_DESK_PAGE_KEYCODE(key_scancode)) &&
                 (!mTERA_KEY_PRESS_EVENT(kbd_evt->state)) &&
                 ((is_client_kbd_repeat_en(cblk)) ||
                  (!is_client_kbd_repeat_en(cblk) && (cblk->kmp.last_gen_desk_key != key_scancode))))
        {
            // Free the message
            mTERA_LOG_DBG("Discarding generic desktop key input event");
            tera_msg_free(cblk->dev, msg);
        }
        //
        // Discard all keys that are not from the keyboard, consumer or
        // generic desktop pages
        //
        else if (!(mTERA_CONSUMER_PAGE_KEYCODE(key_scancode)) &&
                 !(mTERA_KEYBOARD_PAGE_KEYCODE(key_scancode)) &&
                 !(mTERA_GEN_DESK_PAGE_KEYCODE(key_scancode)))
        {
            mTERA_LOG_ERR("Discarding unsupported key code 0x%08X", key_scancode);

            // Free the message
            tera_msg_free(cblk->dev, msg);
        }
        // Discard all unsupported generic desktop keys
        else if (mTERA_GEN_DESK_PAGE_KEYCODE(key_scancode) &&
                 !(mTERA_GEN_DESK_PAGE_VALID_KEYCODE(key_scancode)))
        {
            mTERA_LOG_ERR("Discarding unsupported generic desktop key code 0x%08X", key_scancode);

            // Free the message
            tera_msg_free(cblk->dev, msg);
        }
        else
        {
            //
            // If client repeat processing is enabled and a key break event
            // is pending inject the key break and clear the pending event
            // to prepare for the next key press event.
            //
            if (is_client_kbd_repeat_en(cblk))
            {
                if (cblk->kmp.kbd_last_key_break_pend &&
                    mTERA_KEYBOARD_PAGE_KEYCODE(key_scancode))
                {
#if KBD_REPORT_DBG
                    mTERA_LOG_DBG("injecting key break");
#endif
                    kbd_evt->state &= !TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_MAKE;
                    cblk->kmp.kbd_last_key_break_pend = 0;
                }
                else if (cblk->kmp.kbd_consumer_key_break_pend &&
                         mTERA_CONSUMER_PAGE_KEYCODE(key_scancode))
                {
#if KBD_REPORT_DBG
                    mTERA_LOG_DBG("injecting key break for consumer key");
#endif
                    kbd_evt->state &= !TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_MAKE;
                    cblk->kmp.kbd_consumer_key_break_pend = 0;
                }
                else if (cblk->kmp.kbd_gen_desk_key_break_pend &&
                         mTERA_GEN_DESK_PAGE_KEYCODE(key_scancode))
                {
#if KBD_REPORT_DBG
                    mTERA_LOG_DBG("injecting key break for generic desktop key");
#endif
                    kbd_evt->state &= !TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_MAKE;
                    cblk->kmp.kbd_gen_desk_key_break_pend = 0;
                }
            }

#if KBD_REPORT_DBG
            mTERA_LOG_DBG("state=%u flg=%u lck=%u mod=%u t=%u lang=0x%04x lay=0x%04x code=0x%08x jiff_ms=%u",
                          kbd_evt->state, kbd_evt->flags,
                          kbd_evt->lock, kbd_evt->mod,
                          kbd_evt->tstamp, kbd_evt->lang,
                          kbd_evt->layout, key_scancode,
                          jiffies_to_msecs(jiffies));
#endif

            // if the key is from the keyboard page
            if (mTERA_KEYBOARD_PAGE_KEYCODE(key_scancode))
            {
                if ((mTERA_KEY_PRESS_EVENT(kbd_evt->state)) &&
                    (key_code < FIRST_MODIFIER_USB_CODE))
                {
                    // Record as last key press
                    cblk->kmp.kbd_last_key_press = key_code;

                    //
                    // If client repeat processing is enabled set the flag
                    // forcing the driver to generate a break event
                    //
                    if (is_client_kbd_repeat_en(cblk))
                    {
                        cblk->kmp.kbd_last_key_break_pend = 1;
                    }
                }
                // ensure key_code is valid before using it as an index
                if (key_code > MAX_IDX_USB_KEYBOARD_KEYCODE)
                {
                    mTERA_LOG_ERR("Unsupported keyboard key_code=%d (0x%X)",
                                  key_code, key_code);
                    report_key = 0;
                }
                else
                {
                    report_key = kbd_keycode[key_code];
                }
            }
            // else if the key event is a consumer key
            else if (mTERA_CONSUMER_PAGE_KEYCODE(key_scancode))
            {
                // if this is a key press event
                if (mTERA_KEY_PRESS_EVENT(kbd_evt->state))
                {
                    cblk->kmp.last_consumer_key = key_scancode;

                    //
                    // If client repeat processing is enabled set the flag
                    // forcing the driver to generate a break event
                    //
                    if (is_client_kbd_repeat_en(cblk))
                    {
                        cblk->kmp.kbd_consumer_key_break_pend = 1;
                    }
                }
                // process the key release event
                else
                {
                    // put the key in the report
                    if (cblk->kmp.last_consumer_key == key_scancode)
                    {
                        cblk->kmp.last_consumer_key = 0;
                    }
                    else
                    {
                        mTERA_LOG_ERR("Rcvd key break event for unexpected consumer key 0x%08X",
                                      key_scancode);
                    }
                }

                switch (key_code)
                {
                    case 0xB5:  report_key = KEY_NEXTSONG;     break;
                    case 0xB6:  report_key = KEY_PREVIOUSSONG; break;
                    case 0xB7:  report_key = KEY_STOPCD;       break;
                    case 0xCD:  report_key = KEY_PLAYPAUSE;    break;
                    case 0xE2:  report_key = KEY_MUTE;         break;
                    case 0xE5:  report_key = KEY_BASSBOOST;    break;
                    case 0xE9:  report_key = KEY_VOLUMEUP;     break;
                    case 0xEA:  report_key = KEY_VOLUMEDOWN;   break;
                    case 0x183: report_key = KEY_MEDIA;        break;
                    case 0x18A: report_key = KEY_MAIL;         break;
                    case 0x192: report_key = KEY_CALC;         break;
                    case 0x194: report_key = KEY_COMPUTER;     break;
                    case 0x221: report_key = KEY_SEARCH;       break;
                    case 0x223: report_key = KEY_HOMEPAGE;     break;
                    case 0x224: report_key = KEY_BACK;         break;
                    case 0x225: report_key = KEY_FORWARD;      break;
                    case 0x226: report_key = KEY_STOP;         break;
                    case 0x227: report_key = KEY_REFRESH;      break;
                    case 0x22A: report_key = KEY_BOOKMARKS;    break;

                    // input.h does not have a corresponding keycode for the
                    // following keys
                    case 0xE7:      // Loudness
                    case 0x152:     // Bass up
                    case 0x153:     // Bass down
                    case 0x154:     // Treble up
                    case 0x155:     // Treble down
                    default:
                    {
                        mTERA_LOG_ERR("Received unmapped consumer scancode 0x%X key_code 0x%X",
                                      key_scancode, key_code);
                        report_key = 0;
                        break;
                    }
                }

#if KBD_REPORT_DBG
                mTERA_LOG_DBG("consumer translated scancode: %02X last key=0x%x report_key=%d (0x%X)",
                              key_scancode, cblk->kmp.last_consumer_key, report_key, report_key);
#endif
            }
            // the key is from the generic desktop
            else
            {
                // if this is a key press event
                if (mTERA_KEY_PRESS_EVENT(kbd_evt->state))
                {
                    cblk->kmp.last_gen_desk_key = key_scancode;

                    //
                    // If client repeat processing is enabled set the flag
                    // forcing the driver to generate a break event
                    //
                    if (is_client_kbd_repeat_en(cblk))
                    {
                        cblk->kmp.kbd_gen_desk_key_break_pend = 1;
                    }
                }
                // process the key release event
                else
                {
                    // put the key in the report
                    if (cblk->kmp.last_gen_desk_key == key_scancode)
                    {
                        cblk->kmp.last_gen_desk_key = 0;
                    }
                    else
                    {
                        mTERA_LOG_ERR("Rcvd key break event for unexpected generic desktop key 0x%08X",
                                      key_scancode);
                    }
                }

                //
                // Fill in the report. The report contains information
                // on 1 generic desktop key press. If this is a key
                // release event the report is cleared. Each
                // supported keys is mapped to a specific bit in the
                // HID report.
                //
                switch (key_code)
                {
                    case GEN_DESK_SYS_POWERDOWN_KEYCODE: report_key = KEY_POWER;  break;
                    case GEN_DESK_SYS_SLEEP_KEYCODE:     report_key = KEY_SLEEP;  break;
                    case GEN_DESK_SYS_WAKEUP_KEYCODE:    report_key = KEY_WAKEUP; break;

                    default:
                    {
                        mTERA_LOG_ERR("Received unmapped consumer scancode 0x%X key_code %X",
                                      key_scancode, key_code);
                        report_key = 0;
                        break;
                    }
                }

#if KBD_REPORT_DBG
                mTERA_LOG_DBG("generic desktop translated scancode: %02X last key=0x%x report_key=%d (0x%X)",
                              key_scancode, cblk->kmp.last_gen_desk_key, report_key, report_key);
#endif
            }

            //
            // If the last reported key was a make event for non-control key
            // from the USB keyboard page and the next event is also
            // a make event inject a break event for the previous key.
            //
            // If this is not done the last key gets into a state where it must
            // be pressed twice before it works properly.
            //
            if ((cblk->kmp.lkbd_last_key_state) &&
                ((cblk->kmp.lkbd_last_key_id < FIRST_MODIFIER_USB_CODE) &&
                 (cblk->kmp.lkbd_last_key_page == TERA_USB_PAGE_KEYBOARD)) &&
                (kbd_evt->state))
            {
                input_report_key(cblk->kmp.lkbd, kbd_keycode[cblk->kmp.lkbd_last_key_id], 0);
            }

            // Keep track of the last key event data
            cblk->kmp.lkbd_last_key_id    = key_code;
            cblk->kmp.lkbd_last_key_page  = key_page;
            cblk->kmp.lkbd_last_key_state = kbd_evt->state;

            // report the key information to the kernel
            if (report_key == 0)
            {
                mTERA_LOG_ERR("Unable to translate scancode=0x%X key_code=0x%X",
                              kbd_evt->scancode,
                              key_code);
            }
            else
            {
                input_report_key(cblk->kmp.lkbd, report_key, kbd_evt->state);
                input_sync(cblk->kmp.lkbd);
#if KBD_REPORT_DBG
                mTERA_LOG_DBG("code=0x%08X report_key=%d (0x%X) key_code=0x%X state=%d jiff_ms=%u",
                              kbd_evt->scancode,
                              report_key,
                              report_key,
                              key_code,
                              kbd_evt->state,
                              jiffies_to_msecs(jiffies));
#endif
            }

            // We processed a local keyboard event
            ret = 1;

            if (is_client_kbd_repeat_en(cblk) &&
                ((mTERA_KEYBOARD_PAGE_KEYCODE(key_scancode) &&
                  cblk->kmp.kbd_last_key_break_pend) ||
                 (mTERA_CONSUMER_PAGE_KEYCODE(key_scancode) &&
                  cblk->kmp.kbd_consumer_key_break_pend) ||
                 (mTERA_GEN_DESK_PAGE_KEYCODE(key_scancode) &&
                  cblk->kmp.kbd_gen_desk_key_break_pend)))
            {
                //
                // Enqueue message back onto the Q which injects the break
                // event for the key press in the next thread cycle
                //
                tera_msg_q_head_enqueue(&cblk->kmp.kbd_evt_msg_q, msg);
            }
            else
            {
                // Free the message
                tera_msg_free(cblk->dev, msg);
            } // end if
        } // end if
    } // end if

    return ret;
} // kmp_thread_update_keyboard_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_thread_update_mouse_apdu_proc
//      Process a local cursor event in the background thread
//
//  Description:
//      This function processes a local cursor event in the background
//      thread. If there is an outstanding local cursor event waiting to be
//      injected back into the kernel, this function removes it from the
//      cursor event message queue and reports it to the kernel.
//
static INT kmp_thread_update_mouse_apdu_proc(
    sHDC_CBLK *cblk
    )
{
    INT                               ret = 0;
    sTERA_MSG                        *msg;
    sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT *cur_evt;
    INT                               delta_pix_x;
    INT                               delta_pix_y;
    INT                               per_disp_abs_x;
    INT                               per_disp_abs_y;
    INT                               client_pix_x;
    INT                               client_pix_y;
    UINT32                            cursor_disp;
    UINT16                            last_abs_x;
    UINT16                            last_abs_y;

    ASSERT(cblk != NULL);

    // Get a local cursor event, if any, from the message queue
    msg = tera_msg_q_dequeue(&cblk->kmp.cur_evt_msg_q);
    if (msg != NULL)
    {
        cur_evt = mTERA_MSG_BUF_GET(msg);

        // If the report indicates relative mode is preferred OR the driver is
        // configured to submit relative reports AND the report contains
        // relative data
        if ((cur_evt->flags & TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_FLAGS_REL_PREF) ||
            ((cblk->kmp.driver_cfg.lcur_mode == TERA_PCOIP_HDC_LCUR_REL_ALG) &&
             (cur_evt->flags & TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_FLAGS_BOTH_POS)))
        {
            // submit button & wheel data using the absolute cursor device
            input_report_key(cblk->kmp.lcur_abs, BTN_LEFT,   cur_evt->btn & 0x01);
            input_report_key(cblk->kmp.lcur_abs, BTN_MIDDLE, cur_evt->btn & 0x02);
            input_report_key(cblk->kmp.lcur_abs, BTN_RIGHT,  cur_evt->btn & 0x04);
            input_report_key(cblk->kmp.lcur_abs, BTN_SIDE,   cur_evt->btn & 0x10);
            input_report_key(cblk->kmp.lcur_abs, BTN_EXTRA,  cur_evt->btn & 0x20);
            input_report_rel(cblk->kmp.lcur_abs, REL_WHEEL,  cur_evt->wheel);
            input_sync(cblk->kmp.lcur_abs);

            // submit cursor data using the relative cursor device
            input_report_rel(cblk->kmp.lcur_rel, REL_X, cur_evt->rel_x);
            input_report_rel(cblk->kmp.lcur_rel, REL_Y, cur_evt->rel_y);
            input_sync(cblk->kmp.lcur_rel);

#ifdef CUR_REPORT_DBG
            mTERA_LOG_DBG("mode %s: evt abs (%d, %d) rel (%d, %d), t=%u, w=%hd, b=0x%02x, f=0x%x",
                          lcur_evt_alg_str_tbl[cblk->kmp.driver_cfg.lcur_mode],
                          cur_evt->abs_x, cur_evt->abs_y,
                          cur_evt->rel_x, cur_evt->rel_y,
                          cur_evt->tstamp,
                          cur_evt->wheel,
                          cur_evt->btn,
                          cur_evt->flags);
#endif
        }
        // Don't submit the report if the driver mode is relative and the
        // report indicates absolute mode is preferred AND the report does not
        // contain both types of data.
        else if (cblk->kmp.driver_cfg.lcur_mode == TERA_PCOIP_HDC_LCUR_REL_ALG)
        {
            mTERA_LOG_INFO("dropping report: mode=Relative, f=0x%x (only abs data provided)",
                           cur_evt->flags);
        }
        // if TwinView is disabled and more than one screen is active
        else if ((!cblk->kmp.disp_tbl.twinview_active) && (cblk->kmp.disp_tbl.num_active_displays > 1))
        {
            // if the screen changed generate a relative cursor report
            cursor_disp = kmp_get_disp_cursor_is_on(&cblk->kmp, cur_evt->abs_x, cur_evt->abs_y);
            if (cursor_disp != cblk->kmp.lcur_last_cursor_disp)
            {
                last_abs_x = cblk->kmp.lcur_last_abs_x[cblk->kmp.idx_last_client_pt];
                last_abs_y = cblk->kmp.lcur_last_abs_y[cblk->kmp.idx_last_client_pt];

                // if cursor moved right and crossed a screen boundary
                if ((cur_evt->abs_x > last_abs_x) &&
                    (last_abs_x < cblk->kmp.disp_tbl.displays[cursor_disp].min_abs_x) &&
                    (cur_evt->abs_x >= cblk->kmp.disp_tbl.displays[cursor_disp].min_abs_x))
                {
                    delta_pix_x = (((cur_evt->abs_x - last_abs_x) * cblk->kmp.disp_tbl.desktop_width) / 65536) + 1;
                }
                // if cursor moved left and crossed a screen boundary
                else if ((cur_evt->abs_x < last_abs_x) &&
                         (last_abs_x > cblk->kmp.disp_tbl.displays[cursor_disp].max_abs_x) &&
                         (cur_evt->abs_x <= cblk->kmp.disp_tbl.displays[cursor_disp].max_abs_x))
                {
                    delta_pix_x = (((cur_evt->abs_x - last_abs_x) * cblk->kmp.disp_tbl.desktop_width) / 65536) - 1;
                }
                else
                {
                    delta_pix_x = 0;
                }

                // if cursor moved down and crossed a screen boundary
                if ((cur_evt->abs_y > last_abs_y) &&
                    (last_abs_y < cblk->kmp.disp_tbl.displays[cursor_disp].min_abs_y) &&
                    (cur_evt->abs_y >= cblk->kmp.disp_tbl.displays[cursor_disp].min_abs_y))
                {
                    delta_pix_y = (((cur_evt->abs_y - last_abs_y) * cblk->kmp.disp_tbl.desktop_height) / 65536) + 1;
                }
                // if cursor moved up and crossed a screen boundary
                else if ((cur_evt->abs_y < last_abs_y) &&
                         (last_abs_y > cblk->kmp.disp_tbl.displays[cursor_disp].max_abs_y) &&
                         (cur_evt->abs_y <= cblk->kmp.disp_tbl.displays[cursor_disp].max_abs_y))
                {
                    delta_pix_y = (((cur_evt->abs_y - last_abs_y) * cblk->kmp.disp_tbl.desktop_height) / 65536) - 1;
                }
                else
                {
                    delta_pix_y = 0;
                }

                // submit button & wheel data using the absolute cursor device
                input_report_key(cblk->kmp.lcur_abs, BTN_LEFT,   cur_evt->btn & 0x01);
                input_report_key(cblk->kmp.lcur_abs, BTN_MIDDLE, cur_evt->btn & 0x02);
                input_report_key(cblk->kmp.lcur_abs, BTN_RIGHT,  cur_evt->btn & 0x04);
                input_report_key(cblk->kmp.lcur_abs, BTN_SIDE,   cur_evt->btn & 0x10);
                input_report_key(cblk->kmp.lcur_abs, BTN_EXTRA,  cur_evt->btn & 0x20);
                input_report_rel(cblk->kmp.lcur_abs, REL_WHEEL,  cur_evt->wheel);
                input_sync(cblk->kmp.lcur_abs);

                // submit cursor data using the relative cursor device
                input_report_rel(cblk->kmp.lcur_rel, REL_X, delta_pix_x);
                input_report_rel(cblk->kmp.lcur_rel, REL_Y, delta_pix_y);
                input_sync(cblk->kmp.lcur_rel);

#ifdef CUR_REPORT_DBG
                mTERA_LOG_DBG("rel rep (%d, %d) cur (%d, %d) last (%d, %d) r (%d, %d) d %d -> %d, t=%u, w=%hd, b=0x%02x",
                              delta_pix_x, delta_pix_y,
                              cur_evt->abs_x, cur_evt->abs_y,
                              last_abs_x, last_abs_y,
                              cur_evt->rel_x,
                              cur_evt->rel_y,
                              cblk->kmp.lcur_last_cursor_disp,
                              cursor_disp,
                              cur_evt->tstamp,
                              cur_evt->wheel,
                              cur_evt->btn);
#endif
            }
            // screen did not change
            else
            {
                // Convert the client absolute cursor data into pixel coordinates
                // and then convert that number to absolute coordinates for the
                // display the cursor is on
                client_pix_x = (cur_evt->abs_x * cblk->kmp.disp_tbl.desktop_width)  / 65536;
                client_pix_y = (cur_evt->abs_y * cblk->kmp.disp_tbl.desktop_height) / 65536;

                // The following IF condition might not be correct for all
                // Linux distributions. Testing shows the scaling factor differs
                // between CentOS 5.3 and CentOS 6.2. The code assumes the
                // scaling factor changed when CentOS 6.0 was released, which
                // is based on kernel version 2.6.32.
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32))
                per_disp_abs_x = ((client_pix_x - cblk->kmp.disp_tbl.displays[cursor_disp].min_pix_x) * 65536) / cblk->kmp.disp_tbl.displays[cursor_disp].width;
                per_disp_abs_y = ((client_pix_y - cblk->kmp.disp_tbl.displays[cursor_disp].min_pix_y) * 65536) / cblk->kmp.disp_tbl.displays[cursor_disp].height;
#else
                per_disp_abs_x = ((client_pix_x - cblk->kmp.disp_tbl.displays[cursor_disp].min_pix_x) * 65536) / cblk->kmp.disp_tbl.primary_width;
                per_disp_abs_y = ((client_pix_y - cblk->kmp.disp_tbl.displays[cursor_disp].min_pix_y) * 65536) / cblk->kmp.disp_tbl.primary_height;
#endif

                // report cursor, button and wheel data using the absolute
                // cursor device
                input_report_key(cblk->kmp.lcur_abs, BTN_LEFT,   cur_evt->btn & 0x01);
                input_report_key(cblk->kmp.lcur_abs, BTN_MIDDLE, cur_evt->btn & 0x02);
                input_report_key(cblk->kmp.lcur_abs, BTN_RIGHT,  cur_evt->btn & 0x04);
                input_report_key(cblk->kmp.lcur_abs, BTN_SIDE,   cur_evt->btn & 0x10);
                input_report_key(cblk->kmp.lcur_abs, BTN_EXTRA,  cur_evt->btn & 0x20);
                input_report_abs(cblk->kmp.lcur_abs, ABS_X,      per_disp_abs_x);
                input_report_abs(cblk->kmp.lcur_abs, ABS_Y,      per_disp_abs_y);
                input_report_rel(cblk->kmp.lcur_abs, REL_WHEEL,  cur_evt->wheel);
                input_sync(cblk->kmp.lcur_abs);
#ifdef CUR_REPORT_DBG
                mTERA_LOG_DBG("rep abs (%d, %d) from  abs (%d, %d) rel (%d, %d) pix (%d, %d) d %d, t=%u, w=%hd, b=0x%02x",
                              per_disp_abs_x,
                              per_disp_abs_y,
                              cur_evt->abs_x,
                              cur_evt->abs_y,
                              cur_evt->rel_x,
                              cur_evt->rel_y,
                              client_pix_x,
                              client_pix_y,
                              cursor_disp,
                              cur_evt->tstamp,
                              cur_evt->wheel,
                              cur_evt->btn);
#endif
            }

            // update information about the last event
            cblk->kmp.lcur_last_cursor_disp = cursor_disp;
        }
        // Report the client absolute cursor data without making any changes.
        // Note: this works when one screen is active or when TwinView is
        //       active for all screens on the desktop. This won't work if
        //       there are multiple X screens.
        else
        {
            // report cursor, button and wheel data using the absolute cursor
            // device
            input_report_key(cblk->kmp.lcur_abs, BTN_LEFT,   cur_evt->btn & 0x01);
            input_report_key(cblk->kmp.lcur_abs, BTN_MIDDLE, cur_evt->btn & 0x02);
            input_report_key(cblk->kmp.lcur_abs, BTN_RIGHT,  cur_evt->btn & 0x04);
            input_report_key(cblk->kmp.lcur_abs, BTN_SIDE,   cur_evt->btn & 0x10);
            input_report_key(cblk->kmp.lcur_abs, BTN_EXTRA,  cur_evt->btn & 0x20);
            input_report_abs(cblk->kmp.lcur_abs, ABS_X,      cur_evt->abs_x);
            input_report_abs(cblk->kmp.lcur_abs, ABS_Y,      cur_evt->abs_y);
            input_report_rel(cblk->kmp.lcur_abs, REL_WHEEL,  cur_evt->wheel);
            input_sync(cblk->kmp.lcur_abs);
#ifdef CUR_REPORT_DBG
            mTERA_LOG_DBG("abs (%d, %d), rel (%d, %d), t=%u, w=%hd, b=0x%02x",
                          cur_evt->abs_x,
                          cur_evt->abs_y,
                          cur_evt->rel_x,
                          cur_evt->rel_y,
                          cur_evt->tstamp,
                          cur_evt->wheel,
                          cur_evt->btn);
#endif
        }

        // update information about the last event
        if(!down_interruptible(&cblk->kmp.data_lock))
        {
            cblk->kmp.idx_last_client_pt = (cblk->kmp.idx_last_client_pt + 1) % MAX_CLIENT_CUR_POINTS;
            cblk->kmp.lcur_last_abs_x[cblk->kmp.idx_last_client_pt] = cur_evt->abs_x;
            cblk->kmp.lcur_last_abs_y[cblk->kmp.idx_last_client_pt] = cur_evt->abs_y;
            cblk->kmp.max_age_cur_rep[cblk->kmp.idx_last_client_pt] = jiffies + MAX_AGE_CURSOR_REPORT;
            cblk->kmp.lcur_last_btn_state   = cur_evt->btn;

            up(&cblk->kmp.data_lock);
        }
        else
        {
            ret = TERA_ERR_FAILURE;
            mTERA_LOG_ERR("error processing cursor update");
        }

        // We processed a local keyboard event
        ret = 1;

        // Free the message
        tera_msg_free(cblk->dev, msg);
    } // end if

    return ret;
} // kmp_thread_update_mouse_apdu_proc()

// --------------------------------------------------------------------------
//  kmp_output_kbd_event_proc
///     Process keyboard events from the input module
//
//  Description:
///     This function processes keyboard events received from the input event
///     handler. The event handler can send events to control the keyboard
///     LEDS along with keyclicks. The function and the firmware currently
///     only handle controlling the LEDs.
///
///     When a new LED event is received the function sets a flag that is
///     checked each time the dpc_entry function executes. When the flag is
///     set a message containing the LED settings is created and sent to the
///     firmware.
///
///     Note: the event handler asynchronoously updates the led field in the
///     keyboard input_dev structure when the state of any LEDs change.
//
static int kmp_output_kbd_event_proc(
    struct input_dev *dev,
    unsigned int type,
    unsigned int code,
    int value
    )
{
    int ret=0;

    // if a message needs to be sent to the FW
    if(type == EV_LED)
    {
        set_bit(1, &kbd_event_arrived);
        mTERA_LOG_DBG("keyboard LED event received");
    }
    //
    // Ignore EV_REP events. Sometimes this event is sent to the handler
    // when the user logs in. I don't know why, but it doesn't cause any
    // problems, so it seems safe to ignore it.
    //
    else if(type != EV_REP)
    {
        mTERA_LOG_ERR("ignored kbd event handler call with type %d", type);
        ret = -1;
    }

    return ret;
} // kmp_output_kbd_event_proc()

// --------------------------------------------------------------------------
//  kmp_lkbd_dev_init
///     Allocate and initialize an input device for keyboard input
//
//  Description:
///     This function allocates a kernel input device which the driver will
///     send/receive keyboard data to/from.
//
static INT32 kmp_lkbd_dev_init(
    sHDC_KMP_CBLK *cblk
    )
{
    INT32 i;
    INT32 err;

    // allocate an input device for the keyboard and initialize it
    cblk->lkbd = input_allocate_device();
    if (!cblk->lkbd)
    {
        mTERA_LOG_CRIT("failed to allocate keyboard input device");
        return TERA_ERR_FAILURE;
    }

    cblk->lkbd->name = "PCoIP Virtual Keyboard";
    cblk->lkbd->id.bustype = BUS_PCI;
    cblk->lkbd->id.vendor = TERA_PCI_VENDOR_ID_TERA1;
    cblk->lkbd->id.product = TERA_PCI_PRODUCT_ID_TERA1;

    cblk->lkbd->event = kmp_output_kbd_event_proc;
    set_bit(EV_KEY, cblk->lkbd->evbit);
    set_bit(EV_LED, cblk->lkbd->evbit);
    set_bit(EV_REP, cblk->lkbd->evbit);
    set_bit(LED_NUML, cblk->lkbd->ledbit);
    set_bit(LED_CAPSL, cblk->lkbd->ledbit);
    set_bit(LED_SCROLLL, cblk->lkbd->ledbit);
    set_bit(LED_COMPOSE, cblk->lkbd->ledbit);
    set_bit(LED_KANA, cblk->lkbd->ledbit);

    for (i=0; i<NUM_ENTRIES_KEYCODE_TABLE; i++)
    {
        set_bit(kbd_keycode[i], cblk->lkbd->keybit);
    }
    clear_bit(0, cblk->lkbd->keybit);

    __set_bit(MSC_SCAN, cblk->lkbd->mscbit);

    err = input_register_device(cblk->lkbd);
    if (err)
    {
        mTERA_LOG_CRIT("error registering keyboard, err: %d", err);
        input_free_device(cblk->lkbd);
        cblk->lkbd = NULL;
        return err;
    }
    else
    {
        return TERA_SUCCESS;
    }
} // kmp_lkbd_dev_init()

// --------------------------------------------------------------------------
//  kmp_lcur_dev_init
///     Allocate and initialize an input device for mouse input
//
//  Description:
///     This function allocates two kernel input devices which the driver will
///     send/receive cursor/mouse data to/from. One device is a mouse that
///     submits absolute cursor data. The second device is a mouse that submits
///     relative cursor data.
///
///     Two devices are needed to support systems with more than 1 display when
///     TwinView mode is disabled. Cursor position updates are processed by the
///     evdev driver. The current version of evdev assumes the system has only
///     one absolute cursor input device (for example a touchscreen). A PCoIP
///     client device treats the entire screen surface as a single absolute
///     surface. When the HDC driver processes cursor updates it checks to see
///     if the cursor changed screens. When the cursor changes screens the
///     driver generates a relative cursor position report. When the cursor
///     does not change screens the driver generates an absolute cursor
///     position report.
//
static INT32 kmp_lcur_dev_init(
    sHDC_KMP_CBLK *cblk
    )
{
    INT32 err;

    // absolute cursor device initialization
    cblk->lcur_abs = input_allocate_device();
    if (!cblk->lcur_abs)
    {
        mTERA_LOG_CRIT("failed to allocate mouse input device");
        return TERA_ERR_FAILURE;
    }

    cblk->lcur_abs->name = "PCoIP Virtual Mouse";
    cblk->lcur_abs->id.bustype = BUS_PCI;
    cblk->lcur_abs->id.vendor  = TERA_PCI_VENDOR_ID_TERA1;
    cblk->lcur_abs->id.product = TERA_PCI_PRODUCT_ID_TERA1;

    cblk->lcur_abs->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS) | BIT_MASK(EV_REL);
    cblk->lcur_abs->absbit[0] = BIT_MASK(ABS_X) | BIT_MASK(ABS_Y);
    cblk->lcur_abs->relbit[BIT_WORD(REL_WHEEL)] |= BIT_MASK(REL_WHEEL);
    cblk->lcur_abs->keybit[BIT_WORD(BTN_LEFT)]   = BIT_MASK(BTN_LEFT) |
                                                   BIT_MASK(BTN_RIGHT) |
                                                   BIT_MASK(BTN_MIDDLE) |
                                                   BIT_MASK(BTN_SIDE) |
                                                   BIT_MASK(BTN_EXTRA);
    input_set_abs_params(cblk->lcur_abs, ABS_X, 0, TERA_LCUR_MAX_WIDTH, 0, 0);
    input_set_abs_params(cblk->lcur_abs, ABS_Y, 0, TERA_LCUR_MAX_HEIGHT, 0, 0);

    err = input_register_device(cblk->lcur_abs);
    if (err)
    {
        mTERA_LOG_CRIT("error registering absolute cursor, err: %d", err);
        goto err_reg_lcur_abs;
    }

    // relative cursor device initialization
    cblk->lcur_rel = input_allocate_device();
    if (!cblk->lcur_rel)
    {
        mTERA_LOG_CRIT("failed to allocate relative local cursor input device");
        goto err_alloc_lcur_rel;
    }

    cblk->lcur_rel->name = "PCoIP Virtual Relative Mouse";
    cblk->lcur_rel->id.bustype = BUS_PCI;
    cblk->lcur_rel->id.vendor = TERA_PCI_VENDOR_ID_TERA1;
    cblk->lcur_rel->id.product = TERA_PCI_PRODUCT_ID_TERA1;

    //
    // The relative mouse only generates relative cursor position update
    // reports but when the device is registered the code specifies the
    // device reports button and wheel events. If this is not done the
    // kernel does not process the relative cursor reports properly.
    //
    // Use the command "cat /proc/bus/input/devices" to view the properties
    // of all the input devices.
    //
    cblk->lcur_rel->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_REL);
    cblk->lcur_rel->relbit[0] = BIT(REL_X) | BIT(REL_Y);
    cblk->lcur_rel->relbit[BIT_WORD(REL_WHEEL)] |= BIT_MASK(REL_WHEEL);
    cblk->lcur_rel->keybit[BIT_WORD(BTN_LEFT)]   = BIT_MASK(BTN_LEFT) |
                                                   BIT_MASK(BTN_RIGHT) |
                                                   BIT_MASK(BTN_MIDDLE) |
                                                   BIT_MASK(BTN_SIDE) |
                                                   BIT_MASK(BTN_EXTRA);

    err = input_register_device(cblk->lcur_rel);
    if (err)
    {
        mTERA_LOG_CRIT("error registering relative cursor, err: %d", err);
        goto err_reg_lcur_rel;
    }

    return TERA_SUCCESS;

err_reg_lcur_rel:
    input_free_device(cblk->lcur_rel);
    cblk->lcur_rel = NULL;
err_alloc_lcur_rel:
err_reg_lcur_abs:
    input_free_device(cblk->lcur_abs);
    cblk->lcur_abs = NULL;
    return err;
} // kmp_lcur_dev_init()

// --------------------------------------------------------------------------
//  kmp_get_disp_cursor_is_on
///     Get the index of the display the cursor is on
//
//  Description:
///     This function determines which display the cursor is on. The caller
///     passes in the new client cursor coordinates, which range from 0 to
///     0xFFFF in the X and Y directions. These coordinates cover the entire
///     desktop. The cursor is considered to be on a display when the X and
///     Y coordinates are between a displays min and max absolute coordinates.
///
///     Note: If this function encounters an error it returns zero.
//
static UINT32 kmp_get_disp_cursor_is_on(
    sHDC_KMP_CBLK *cblk_kmp,
    UINT16         abs_x,
    UINT16         abs_y
    )
{
    UINT32 idx = cblk_kmp->lcur_last_cursor_disp;
    int    i;

#ifdef DBG
    if (idx > TERA_PCOIP_HCC_APDU_MAX_NUM_DISP)
    {
        mTERA_LOG_ERR("Invalid cursor display index %d", idx);
    }
#endif

    // search through each display, starting with the entry the last cursor
    // was on
    for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
    {
        if ((cblk_kmp->disp_tbl.displays[idx].active) &&
            (abs_x >= cblk_kmp->disp_tbl.displays[idx].min_abs_x) &&
            (abs_x <= cblk_kmp->disp_tbl.displays[idx].max_abs_x) &&
            (abs_y >= cblk_kmp->disp_tbl.displays[idx].min_abs_y) &&
            (abs_y <= cblk_kmp->disp_tbl.displays[idx].max_abs_y))
        {
            return idx;
        }
        idx = (idx + 1) % TERA_PCOIP_HCC_APDU_MAX_NUM_DISP;
    }

    idx = cblk_kmp->lcur_last_cursor_disp;
    mTERA_LOG_ERR("unable to find display for cursor (%d, %d) (0x%X, 0x%X), last disp %d",
                  abs_x, abs_y, abs_x, abs_y, idx);
    for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
    {
        mTERA_LOG_ERR("disp[%d]: active %d, min_x %d, max_x %d, min_y %d, max_y %d",
                      idx,
                      cblk_kmp->disp_tbl.displays[idx].active,
                      cblk_kmp->disp_tbl.displays[idx].min_abs_x,
                      cblk_kmp->disp_tbl.displays[idx].max_abs_x,
                      cblk_kmp->disp_tbl.displays[idx].min_abs_y,
                      cblk_kmp->disp_tbl.displays[idx].max_abs_y);

        idx = (idx + 1) % TERA_PCOIP_HCC_APDU_MAX_NUM_DISP;
    }
    return 0;
} // kmp_get_disp_cursor_is_on()

// --------------------------------------------------------------------------
//  hdc_kmp_init
//      Initialize the KMP user
//
//  Description:
//      This function initializes the KMP user. This function is responsible
//      for initializing any and all user-specific objects.
//
INT32 hdc_kmp_init(     // TERA_SUCCESS, TERA_ERR_FAILURE
    sHDC_CBLK *cblk     // HDC control block
    )
{
    INT32   ret = TERA_ERR_FAILURE;
    INT     i;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // Create a spinlock that ensures only one request is sent to the FW
    init_MUTEX(&cblk->kmp.data_lock);

    ret = kmp_lkbd_dev_init(&cblk->kmp);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize keyboard input dev (ret=%d)", ret);
    }
    else
    {
        ret = kmp_lcur_dev_init(&cblk->kmp);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to initialize cursor input dev (ret=%d)", ret);
        }
        else
        {
            // Initialize the local cursor event message queue
            ret = tera_msg_q_init(&cblk->kmp.cur_evt_msg_q);
            if (ret != TERA_SUCCESS)
            {
                mTERA_LOG_ERR("failed to initialize local cursor event message "
                              "queue (ret=%d)", ret);
            }
            else
            {
                // Initialize the local keyboard event message queue
                ret = tera_msg_q_init(&cblk->kmp.kbd_evt_msg_q);
                if (ret != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("failed to initialize local keyboard event message "
                                  "queue (ret=%d)", ret);
                }
                else
                {
                    // Initialize the other state variables
                    cblk->kmp.lang = 0;
                    cblk->kmp.modifier = 0;

                    cblk->kmp.lkbd_last_key_id = 0;
                    cblk->kmp.lkbd_last_key_page = 0;
                    cblk->kmp.lkbd_last_key_state = 0;
                    cblk->kmp.lcur_last_btn_state = 0;
                    for (i=0; i<MAX_CLIENT_CUR_POINTS; i++)
                    {
                        cblk->kmp.lcur_last_abs_x[i] = 0;
                        cblk->kmp.lcur_last_abs_y[i] = 0;
                        cblk->kmp.max_age_cur_rep[i] = jiffies;
                    }
                    cblk->kmp.lcur_last_cursor_disp = 0;
                    cblk->kmp.disp_tbl.num_active_displays = 0;
                    for (i=0; i<TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
                    {
                        cblk->kmp.disp_tbl.displays[i].active = 0;
                    }
                    cblk->kmp.driver_cfg.lcur_mode = TERA_PCOIP_HDC_LCUR_ABS_ALG;
                }
            }
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // hdc_kmp_init()

// --------------------------------------------------------------------------
//  hdc_kmp_cleanup
///     HDC KMP user clean up routine
//
//  Description:
///     This function performs the HDC KMP user clean up routine.
///
/// \return
///     None
///
void hdc_kmp_cleanup(
    sHDC_CBLK *cblk
    )
{
    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    mTERA_LOG_DBG("end");
} // hdc_kmp_cleanup()

// --------------------------------------------------------------------------
//  hdc_kmp_start
//      Start the HDC KMP user portion
//
//  Description:
//      This function starts the HDC KMP user portion.
//
INT32 hdc_kmp_start(    // TERA_SUCCESS
    sHDC_CBLK *cblk     // HDC control block
    )
{
    INT32 ret = TERA_SUCCESS;

    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // reset state variables that must be reset each time this module starts
//    cblk->kmp.cur_systime_last_evt = 0;
//    cblk->kmp.cur_tstamp_last_evt = 0;
    cblk->kmp.kbd_last_key_press = 0;
    cblk->kmp.kbd_last_key_break_pend = 0;
    cblk->kmp.last_consumer_key = 0;
    cblk->kmp.kbd_consumer_key_break_pend = 0;
    cblk->kmp.last_gen_desk_key = 0;
    cblk->kmp.kbd_gen_desk_key_break_pend = 0;
    cblk->kmp.feat_stat.err = 0;
    cblk->kmp.feat_stat.cnt = 0;

    // TBD - can this go away (they're 1.4.5 variables)
    cblk->kmp.lkbd_last_key_id = 0;
    cblk->kmp.lkbd_last_key_page = 0;
    cblk->kmp.lkbd_last_key_state = 0;
    cblk->kmp.lcur_last_btn_state = 0;

    mTERA_LOG_DBG("end");

    return ret;
} // hdc_kmp_start()

// --------------------------------------------------------------------------
//  hdc_kmp_stop
//      Stop the HDC KMP user
//
//  Description:
//      This function stops the HDC KMP user.
//
void hdc_kmp_stop(
    sHDC_CBLK *cblk
    )
{
    mTERA_LOG_DBG("start");

    ASSERT(cblk != NULL);

    // Clear any keyboard or cursor presses
    kmp_key_lcur_press_reset(cblk);

    // unregister and free the input devices if they exist
    if (cblk->kmp.lkbd != NULL)
    {
        input_unregister_device(cblk->kmp.lkbd);
        cblk->kmp.lkbd = NULL;
    }
    if (cblk->kmp.lcur_abs != NULL)
    {
        input_unregister_device(cblk->kmp.lcur_abs);
        cblk->kmp.lcur_abs = NULL;
    }
    if (cblk->kmp.lcur_rel != NULL)
    {
        input_unregister_device(cblk->kmp.lcur_rel);
        cblk->kmp.lcur_rel = NULL;
    }

    //TODO purge all local cursor/keyboard events in the message queues?

    mTERA_LOG_DBG("end");
} // hdc_kmp_stop()

// --------------------------------------------------------------------------
//  hdc_kmp_scp_chan_open_proc
//      Process the KMP SCP channel open event
//
//  Description:
//      This function processes the KMP SCP channel open event. This function
//      is called whenever the channel has transitioned from init into ready
//      state. This function reads the MTU from one of the GFUN mailboxes and
//      tells the SCP about it. It also completes any HDC open ioctls in the
//      pending ioctl queues, thus notifying the user that the HDC is open.
//
void hdc_kmp_scp_chan_open_proc(
    sHDC_CBLK *cblk
    )
{
    INT32      err;
    UINT32     mtu;

    ASSERT(cblk != NULL);

    // Read the MTU from the GFUN mailbox
    err = tera_gfun_mbox_read(cblk->dev, TERA_GFUN_MBOX_SLOT_0, &mtu);
    if (err != TERA_SUCCESS)
    {
        mTERA_LOG_DBG("failed to read MTU from GFUN mailbox (ret=%d)",
                      err);

        //TODO What should we do if we ever run into this case?
    }
    else
    {
        mTERA_LOG_DBG("HDC MTU=%u", mtu);

        // Set the SCP MTU size
        err = tera_scp_mtu_set(cblk->scp_handle, (UINT16)mtu);
        if (err != TERA_SUCCESS)
        {
            mTERA_LOG_DBG("failed to set SCP MTU (ret=%d)", err);

            //TODO Again, what do we do here??
        }
    }

    // TBD - the Windows version implements a while loop where it checks for
    //       and completes multiple opens if more than one is pending. The
    //       current Linux architecture prevents this from happening. Is this
    //       a problem?

    // Complete the HDC open ioctl if one is pending
    if (cblk->pending_ioctl_flag == TERA_PCOIP_HDC_IOCTL_KMP_HDC_OPEN)
    {
        mTERA_LOG_DBG("completing HDC_IOCTL_KMP_HDC_OPEN ioctl");

        // set the flag that will wake up the waiting ioctl thread
        cblk->pending_ioctl_flag = 1;

        // wake up the waiting thread
        wake_up_interruptible(&cblk->pending_ioctl_q);
    } // end if

    // reset the KMP state variables
//    cblk->kmp.cur_systime_last_evt = 0;
//    cblk->kmp.cur_tstamp_last_evt = 0;
    cblk->kmp.kbd_last_key_press = 0;
    cblk->kmp.kbd_last_key_break_pend = 0;
    cblk->kmp.last_consumer_key = 0;
    cblk->kmp.kbd_consumer_key_break_pend = 0;
    cblk->kmp.last_gen_desk_key = 0;
    cblk->kmp.kbd_gen_desk_key_break_pend = 0;
    cblk->kmp.feat_stat.err = 0;
    cblk->kmp.feat_stat.cnt = 0;

    // TBD - can this go away (they're 1.4.5 variables)
    cblk->kmp.lkbd_last_key_id = 0;
    cblk->kmp.lkbd_last_key_page = 0;
    cblk->kmp.lkbd_last_key_state = 0;
    cblk->kmp.lcur_last_btn_state = 0;
} // hdc_kmp_scp_chan_open_proc()

// --------------------------------------------------------------------------
//  hdc_kmp_ioctl_proc
//      Process a KMP ioctl
//
//  Description:
//      This function processes a KMP ioctl. If the request is of a known
//      ioctl the request will either be completed or sent to the pending
//      ioctl queue, and the function returns TERA_SUCCESS. If the request
//      is of an unknown ioctl the function returns TERA_ERR_NOT_SUPPORTED
//      without completing the request.
//
INT32 hdc_kmp_ioctl_proc(
    sHDC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    )
{
    INT32 ret;

    ASSERT(cblk != NULL);

    switch (ioctl)
    {
        case TERA_PCOIP_HDC_IOCTL_KMP_HDC_OPEN:
        {
            ret = kmp_ioctl_hdc_open_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_HDC_CLOSE:
        {
            mTERA_LOG_DBG("HDC_IOCTL_KMP_HDC_CLOSE");

            ret = kmp_ioctl_hdc_close_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_FEAT_EN:
        {
            ret = kmp_ioctl_feature_activation_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_FEAT_STAT:
        {
            ret = kmp_ioctl_feature_status_apdu_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_LAST_FEAT_STAT:
        {
            ret = kmp_ioctl_last_feature_status_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_KBD_REPEAT:
        {
            ret = kmp_ioctl_keyboard_repeat_rate_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_SHAPE_BMP:
        {
            ret = kmp_ioctl_update_pointer_shape_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_SHAPE:
        {
            ret = kmp_ioctl_change_pointer_shape_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS:
        {
            ret = kmp_ioctl_set_pointer_location_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS_IF_NEW:
        {
            ret = kmp_ioctl_set_pointer_location_if_new_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_VIS:
        {
            ret = kmp_ioctl_set_pointer_visibility_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_IDLE_HIDE:
        {
            ret = kmp_ioctl_fade_out_control_apdu_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_SPD_GET:
        {
            ret = kmp_ioctl_mouse_speed_query_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_CUR_SPD_SET:
        {
            ret = kmp_ioctl_mouse_speed_update_apdu_set_proc(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_KBD_REPEAT_GET:
        {
            ret = kmp_ioctl_keyboard_repeat_rate_query_apdu_get_proc(cblk, arg, ioctl);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_DISP_TOPO_SET:
        {
            ret = kmp_ioctl_disp_topo_set(cblk, arg);
            break;
        }
        case TERA_PCOIP_HDC_IOCTL_KMP_DRIVER_CFG_SET:
        {
            ret = kmp_ioctl_driver_cfg_set(cblk, arg);
            break;
        }
        default:
        {
            mTERA_LOG_DBG("unsupported ioctl (0x%08x)", ioctl);
            ret = TERA_ERR_NOT_SUPPORTED;
            break;
        }
    }
    return ret;
} // hdc_kmp_ioctl_proc()

// --------------------------------------------------------------------------
//  hdc_kmp_apdu_proc
//      Process an KMP APDU
//
//  Description:
//      This function processes a KMP APDU from the ingress descriptor queue.
//      Only known APDUs are serviced.
//
//      NOTE: The APDU is freed by the calling function after this function
//      returns, so it should NOT be referenced for future use.
//
void hdc_kmp_apdu_proc(
    sHDC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    )
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE *apdu_type = apdu;

    ASSERT(cblk != NULL);
    ASSERT(apdu != NULL);

    if (apdu_len < sizeof(eTERA_PCOIP_HDC_APDU_KMP_TYPE))
    {
        mTERA_LOG_ERR("invalid APDU length");
    }
    else if (cblk->initialized)
    {
        // Need to byte swap the APDU type value
        *apdu_type = mTERA_UINT32_BYTE_SWAP(*apdu_type);

        switch (*apdu_type)
        {
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_EVT:
            {
//                mTERA_LOG_DBG("KMP_TYPE_KBD_EVT");

                kmp_update_keyboard_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_EVT:
            {
//                mTERA_LOG_DBG("KMP_TYPE_CUR_EVT");

                kmp_update_mouse_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_STAT:
            {
                mTERA_LOG_DBG("KMP_TYPE_FEAT_STAT");

                kmp_feature_status_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_LOCALE:
            {
                mTERA_LOG_DBG("KMP_TYPE_KBD_LOCALE");

                kmp_set_keyboard_locale_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD:
            {
                mTERA_LOG_DBG("KMP_TYPE_CUR_SPD");

                kmp_mouse_speed_update_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            case TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REPEAT:
            {
                mTERA_LOG_DBG("KMP_TYPE_KBD_REPEAT");

                kmp_set_keyboard_repeat_rate_apdu_proc(cblk, apdu, apdu_len);
                break;
            }
            default:
            {
                mTERA_LOG_ERR("unsupported KMP APDU type=%u", *apdu_type);
                break;
            }
        }
    } // end if
} // hdc_kmp_pkt_proc()

int test=0;

INT hdc_kmp_thread_proc(
    sHDC_CBLK *cblk
    )
{
    INT ret = 0;

    ASSERT(cblk != NULL);

#if 0 // TBD - Eventually delete this, the code delays the processing of cursor
      //       events by 1 second. Things seem to work okay, which means I don't
      //       need to inject the cursor events based on the timestamp in the
      //       cursor event generated by the client.
    test++;
    if (test > 1000)
    {
        if (test > 2000)
        {
            test = 0;
            mTERA_LOG_DBG("re-enable cursor input");
        }
    }
    else
    {
        ret |= kmp_thread_update_mouse_apdu_proc(cblk);
    }
#else
    //
    // Look for any oustanding keyboard and/or mouse events and inject them
    // into the kernel - mouse events have priority over keyboard events
    //
    ret |= kmp_thread_update_mouse_apdu_proc(cblk);
#endif

    ret |= kmp_thread_update_keyboard_apdu_proc(cblk);

    return ret;
} // hdc_kmp_thread_proc()

