////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host SW Driver Implementation
///
///     This file implements the driver for the PCoIP Host device. This
///     module services the TERA1 PCoIP hsot device which provides local
///     cursor and keyboard data, which is then relayed to the appropriate
///     input device drivers.
///
///     Currently the driver does not include support operating in low power
///     mode, where the system enters the suspend state.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_pcoip_driver.h"
#include "tera_hcc.h"
#include "tera_hdc.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_pcoip_hdc_ioctl.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// The following constant defines the version number of the driver. The driver
// writes a message to the file /var/log/messages containing the version
// string when the driver is loaded.
#define DRV_VERSION "4.1.6"

#define MAX_MSG_SIZE        20

// This constant defines the number of device interfaces that an application
// can use to interact with the driver. Currently the driver supports two
// (HCC and HDC).
#define NUM_DEVICES             MAX_NUM_GFUN

// This constant defines the minor device number for the first device inteface.
#define FIRST_MINOR_DEV_NUM     0

// ---- Local Variables -----------------------------------------------------

static const char pcoip_driver_version[] = DRV_VERSION;
static char pcoip_driver_name[] = "pcoip_host";
static const char pcoip_driver_string[] = "Teradici(R) PCoIP Host Driver";
static const char pcoip_copyright[] = "Copyright (c) 2013 Teradici Corporation.";
static sTERA_DEV_CBLK *cblk[MAX_NUM_GFUN] = {NULL};
static sTERA_DEV_CBLK *hcc_cblk = NULL;
static sTERA_DEV_CBLK *hdc_cblk = NULL;
static int gfun_interfaces_found=0;
static dev_t pcoip_dev;

//
// TERA_PCI_DEVICE - macro used to describe a specific pci device
// @vend: the 16 bit PCI Vendor ID
// @dev: the 16 bit PCI Device ID
// @subvend: the 16 bit PCI Subsystem Vendor ID
// @subdev: the 16 bit PCI Subsystem Device ID
//
// This macro is used to create a struct pci_device_id that matches a
// specific device. It is a modified version of the macro PCI_DEVICE()
// found in pci.h. This version of the macro allows the user to specify values
// for the subsystem vendor and subsystem device field. This must be done for
// the PCoIP host device, otherwise Linux tries to load this module to service
// one of the other functions presented by the TERA1 device on the PCI bus.
//
#define TERA_PCI_DEVICE(vend,dev,subvend,subdev) \
        .vendor = (vend), .device = (dev), \
        .subvendor = (subvend), .subdevice = (subdev)

// pcoip_pci_tbl - PCI Device ID Table
static struct pci_device_id pcoip_pci_tbl[] = {
    {TERA_PCI_DEVICE(TERA_PCI_VENDOR_ID_TERA1,
                     TERA_PCI_PRODUCT_ID_TERA1,
                     TERA_PCI_SUB_VENDOR_ID_TERA1,
                     TERA_PCI_SUB_DEV_ID_TERA1)},
    {TERA_PCI_DEVICE(TERA_PCI_VENDOR_ID_TERA2,
                     TERA_PCI_PRODUCT_ID_TERA2,
                     TERA_PCI_SUB_VENDOR_ID_TERA2,
                     TERA_PCI_SUB_DEV_ID_TERA2)},
    // required last entry
    {0,}
};

MODULE_DEVICE_TABLE(pci, pcoip_pci_tbl);
MODULE_AUTHOR("Teradici Corporation");
MODULE_DESCRIPTION("Teradici(R) Linux TERA1 PCoIP Host Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);

// ---- Local Function Prototypes -------------------------------------------

static INT32 modules_init_phase1(
    struct  pci_dev *dev
    );

static INT32 modules_init_phase2(
    struct  pci_dev *dev
    );

static void modules_cleanup(
    struct  pci_dev *dev
    );

static INT32 modules_start(
    struct  pci_dev *dev
    );

static void modules_stop(
    struct  pci_dev *dev
    );

static int pcoip_init_module(
    void
    );

static void pcoip_exit_module(
    void
    );

static int pcoip_probe(
    struct pci_dev *pdev,
    const struct pci_device_id *ent
    );

static void __devexit pcoip_remove(
    struct pci_dev *pdev
    );

static int pcoip_open(
    struct inode *inode,
    struct file *filp
    );

static int pcoip_release(
    struct inode *inode,
    struct file *filp
    );

#if HAVE_UNLOCKED_IOCTL
static long pcoip_ioctl(
    struct file *filp,
    unsigned int cmd,
    unsigned long arg
    );
#else
static int pcoip_ioctl(
    struct inode *inode,
    struct file *filp,
    unsigned int cmd,
    unsigned long arg
    );
#endif

static struct pci_driver pcoip_driver = {
    .name     = pcoip_driver_name,
    .id_table = pcoip_pci_tbl,
    .probe    = pcoip_probe,
    .remove   = __devexit_p(pcoip_remove),
};

/* pcoip_fops - PCoIP Device Driver File Operations Table */
#if HAVE_UNLOCKED_IOCTL
struct file_operations pcoip_fops = {
    .owner          =  THIS_MODULE,
    .unlocked_ioctl =  pcoip_ioctl,
    .open           =  pcoip_open,
    .release        =  pcoip_release,
};
#else
struct file_operations pcoip_fops = {
    .owner          =  THIS_MODULE,
    .ioctl          =  pcoip_ioctl,
    .open           =  pcoip_open,
    .release        =  pcoip_release,
};
#endif

// --------------------------------------------------------------------------
//  modules_init_phase1
//      Start initializing user modules
//
//  Description:
//      This function initializes the portions of the user modules that can be
//      initialized before the GFUN interface has been initialized. The modules
//      go through a two phase initialization sequence because a portion of the
//      initialization should only be performed when the driver knows whether
//      it is communicationg with the HCC or HDC module. Both user modules are
//      initialized, even though they may or may not be started.
//
static INT32 modules_init_phase1(
    struct  pci_dev *dev
    )
{
    INT32 ret;

    mTERA_LOG_DBG("start");

    // Initialize the HDC module
    ret = tera_hdc_init_phase1(dev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to initialize the HDC module (ret=%d)", ret);
    }
    else
    {
        // Initialize the HCC module
        ret = tera_hcc_init_phase1(dev);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to initialize the HCC module (ret=%d)", ret);
        }
    }

    if (ret != TERA_SUCCESS)
    {
        // Clean up all the modules
        tera_hdc_cleanup(dev);
        tera_hcc_cleanup(dev);
    }

    mTERA_LOG_DBG("end");

    return ret;
} // modules_init_phase1()

// --------------------------------------------------------------------------
//  modules_init_phase2
//      Finish initializing user modules
//
//  Description:
//      This function is called after the GFUN interface is initialized and
//      the driver knows which module (HCC or HDC) is active. This function
//      calls the modules 2nd phase initialization function which completes
//      the initialization phase. After the initialization phase is complete
//      the driver can start the module.
//
static INT32 modules_init_phase2(
    struct  pci_dev *dev
    )
{
    INT32 ret = TERA_ERR_INVALID_STATE;
    UINT32 gfun_cap;

    // Get the GFUN capabilities to determine which module to initialize
    gfun_cap = tera_gfun_cap_get(dev);

    if (gfun_cap & TERA_HDC_GFUN_FUNC_CAP_BITMASK)
    {
        mTERA_LOG_DBG("Finish initializing HDC");

        // Initialize the HDC module
        ret = tera_hdc_init_phase2(dev);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to initialize the HDC module (ret=%d)", ret);
        }
    }

    if (gfun_cap & TERA_HCC_GFUN_FUNC_CAP_BITMASK)
    {
        mTERA_LOG_DBG("Finish initializing HCC");

        // Initialize the HCC module
        ret = tera_hcc_init_phase2(dev);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to initialize the HCC module (ret=%d)", ret);
        }
    }

    if (ret != TERA_SUCCESS)
    {
        // Clean up all the modules
        tera_hdc_cleanup(dev);
        tera_hcc_cleanup(dev);
    }

    mTERA_LOG_DBG("end");

    return ret;
} // modules_init_phase2()

// --------------------------------------------------------------------------
//  modules_cleanup
//      Call the exit routines of the modules
//
//  Description:
//      This function calls the exit routines of the modules in this driver.
//
static void modules_cleanup(
    struct  pci_dev *dev
    )
{
    mTERA_LOG_DBG("start");

    tera_hdc_cleanup(dev);
    tera_hcc_cleanup(dev);

    mTERA_LOG_DBG("end");
} // modules_cleanup()

// --------------------------------------------------------------------------
//  modules_start
//      Call the start routines of the modules
//
//  Description:
//      This function calls the start routines of the modules in this driver.
//
//
static INT32 modules_start(    // TERA_SUCCESS, TERA_ERR_FAILURE
    struct  pci_dev *dev       // Device object
    )
{
    INT32  ret = TERA_SUCCESS;
    UINT32 gfun_cap;

    // Get the GFUN capabilities to determine which modules to start
    gfun_cap = tera_gfun_cap_get(dev);

    if (gfun_cap & TERA_HDC_GFUN_FUNC_CAP_BITMASK)
    {
        mTERA_LOG_DBG("start HDC");

        // Start the HDC module
        ret = tera_hdc_start(dev);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to start HDC module (ret=%d", ret);
        }
    }

    if (gfun_cap & TERA_HCC_GFUN_FUNC_CAP_BITMASK)
    {
        mTERA_LOG_DBG("start HCC");

        // Start the HCC module
        ret = tera_hcc_start(dev);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to start HCC module (ret=%d)", ret);
        }
    }

    return ret;
} // modules_start()

// --------------------------------------------------------------------------
//  modules_stop
//      Call the stop routines of the modules
//
//  Description:
//      This function calls the stop routines of the modules in this driver.
//
//  Notes:
//      The Windows version of the driver calls this function when the system
//      exits the D0 state, i.e. when the system is entering a low power state.
//      Currently this driver does not support entering/exiting low power
//      states, which is why this function is not referenced.
//
static void modules_stop(
    struct  pci_dev *dev
    )
{
    UINT32 gfun_cap;

    mTERA_LOG_DBG("start");

    // Get the GFUN capabilities to determine which modules to start
    gfun_cap = tera_gfun_cap_get(dev);

    if (gfun_cap & TERA_HDC_GFUN_FUNC_CAP_BITMASK)
    {
        // Stop the HDC module
        tera_hdc_stop(dev);
    }

    if (gfun_cap & TERA_HCC_GFUN_FUNC_CAP_BITMASK)
    {
        // Stop the HCC module
        tera_hcc_stop(dev);
    }

    mTERA_LOG_DBG("end");
} // modules_stop()

// --------------------------------------------------------------------------
//  pcoip_init_module
///     Driver Registration Routine
//
//  Description:
///     The pcoip_init_module is the first routine called when the driver is
///     loaded. All it does is register with the PCI subsystem.
///
///     This function is called once and only once during initialization.
//
static int pcoip_init_module(void)
{
    int ret;

    mTERA_LOG_INFO("%s - version %s",
                   pcoip_driver_string,
                   pcoip_driver_version);
    mTERA_LOG_INFO("%s", pcoip_copyright);

    ret = pci_register_driver(&pcoip_driver);
    return ret;
}

module_init(pcoip_init_module);

// --------------------------------------------------------------------------
//  pcoip_exit_module
///     Driver Exit Cleanup Routine
//
//  Description:
///     The pcoip_exit_module is called just before the driver is removed
///     from memory.
//
static void __exit pcoip_exit_module(void)
{
    mTERA_LOG_DBG("start");

    //
    // When this function is called pcoip_remove() is called twice, once for
    // each GFUN instance.
    //
    pci_unregister_driver(&pcoip_driver);

    mTERA_LOG_DBG("end");
}

module_exit(pcoip_exit_module);

// --------------------------------------------------------------------------
//  pcoip_probe
///     Device Initialization Routine
//
//  Description:
///     This function initializes the control block identified by a pci_dev
///     structure. The OS initialization, configuring of the control block
///     private structure, and a hardware reset occur.
///
///     The function returns 0 on success, negative on failure.
///
///     This function is called twice, once for the HCC and a 2nd time for
///     the HDC device.
//
//  Parameters:
///     pdev: PCI device information struct
///     ent: entry in pcoip_pci_tbl
//
static int __devinit pcoip_probe(
    struct pci_dev *pdev,
    const struct pci_device_id *ent
    )
{
    int     err = 0;
    INT32   ret;
    UINT32  dev_id;

    mTERA_LOG_DBG("start");

    //
    // The driver only supports the use of one PCoIP Host card (which has 2
    // device interfaces [HCC and HDC])
    //
    dev_id = gfun_interfaces_found;
    gfun_interfaces_found++;
    if (gfun_interfaces_found > MAX_NUM_GFUN)
    {
        mTERA_LOG_CRIT("The PCoIP Host driver only supports one PCoIP Host card with the Host Driver enabled");
        return -ENOSYS;
    }

    // allocate storage for the top control structure
    cblk[dev_id] = kzalloc(sizeof(sTERA_DEV_CBLK), GFP_KERNEL);
    if (!cblk[dev_id])
    {
        mTERA_LOG_CRIT("error allocating cblk[%d] structure", dev_id);
        return -ENOMEM;
    }
    cblk[dev_id]->pdev = pdev;
    pci_set_drvdata(pdev, cblk[dev_id]);

    // Initialize the control block device id
    cblk[dev_id]->dev_id = dev_id;

    // If this is the first device register a range of char devices
    if (dev_id == 0)
    {
        err = alloc_chrdev_region(&pcoip_dev,
                                  FIRST_MINOR_DEV_NUM,
                                  NUM_DEVICES,
                                  pcoip_driver_name);
        if (err < 0)
        {
            mTERA_LOG_CRIT("unable to allocate major number (err=%d)", err);
            goto err_alloc_major;
        }
        else
        {
            mTERA_LOG_DBG("pcoip_major=%d, pcoip_minor=%d",
                          MAJOR(pcoip_dev), MINOR(pcoip_dev));
        }
    }

    ret = tera_mem_init(pdev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_mem_init() failed (ret=%d)", ret);
        goto err_mem_init;
    }
    mTERA_LOG_DBG("tera_mem_init() okay");

    ret = tera_msg_init(&cblk[dev_id]->msg_cblk, MAX_MSG_SIZE, dev_id);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_msg_init() failed (ret=%d)", ret);
        goto err_msg_init;
    }
    mTERA_LOG_DBG("tera_msg_init() okay");

    ret = tera_pkt_init(pdev, dev_id);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_pkt_init() failed (ret=%d)", ret);
        goto err_pkt_init;
    }
    mTERA_LOG_DBG("tera_pkt_init() okay");

    ret = tera_scp_init(pdev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_scp_init() failed (ret=%d)", ret);
        goto err_scp_init;
    }
    mTERA_LOG_DBG("tera_scp_init() okay");

    ret = tera_gfun_init(pdev, pcoip_driver_name);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_gfun_init() failed (ret=%d)", ret);
        goto err_gfun_init;
    }
    mTERA_LOG_DBG("tera_gfun_init() okay");

    ret = modules_init_phase1(pdev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("modules_init_phase1() failed (ret=%d)", ret);
        goto err_modules_init_phase1;
    }
    mTERA_LOG_DBG("modules_init_phase1() okay");

    ret = tera_gfun_start(pdev,
                          TERA_HCC_GFUN_FUNC_CAP_BITMASK |
                          TERA_HDC_GFUN_FUNC_CAP_BITMASK);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("tera_gfun_start() failed (ret=%d)", ret);
        goto err_gfun_start;
    }
    mTERA_LOG_DBG("tera_gfun_start() okay");

    // TBD - possibly move phase2 init stuff into hcc/hdc_start
    ret = modules_init_phase2(pdev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("modules_init_phase2() failed (ret=%d)", ret);
        goto err_modules_init_phase2;
    }
    mTERA_LOG_DBG("modules_init_phase2() okay");

    ret = modules_start(pdev);
    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_CRIT("modules_start() failed (ret=%d)", ret);
        goto err_modules_start;
    }
    else if (cblk[dev_id]->hcc_cblk.started)
    {
        hcc_cblk = cblk[dev_id];
        mTERA_LOG_DBG("modules_start(dev_id=%d, HCC) okay", dev_id);
    }
    else if (cblk[dev_id]->hdc_cblk.started)
    {
        hdc_cblk = cblk[dev_id];
        mTERA_LOG_DBG("modules_start(dev_id=%d, HDC) okay", dev_id);
    }
    else
    {
        mTERA_LOG_CRIT("modules_start() ok but neither the HCC or HDC is started");
    }
    cblk[dev_id]->initialized = 1;

    //
    // Setup the character device structure used to control the PCoIP device.
    // This is the last things done because after the device is added its
    // operations can be called.
    //
    cdev_init(&cblk[dev_id]->cdev, &pcoip_fops);
    cblk[dev_id]->cdev.owner = THIS_MODULE;
    cblk[dev_id]->cdev.ops = &pcoip_fops;

    // If this is the last device register a range of char devices
    if (gfun_interfaces_found == NUM_DEVICES)
    {
        err = cdev_add(&cblk[dev_id]->cdev, pcoip_dev, NUM_DEVICES);
        if(err)
        {
            mTERA_LOG_CRIT("error adding character device - err %d", err);
            goto err_add_cdev;
        }
        else
        {
            mTERA_LOG_DBG("cdev_add() okay");
        }
    }

    return 0;

err_add_cdev:

#if 0
err_pcc_start:
    tera_pcc_stop(cblk);
err_pcc_init:
    pcc_lcur_exit(cblk);
err_init_lcur:
    pcc_lkbd_exit(cblk);
err_init_lkbd:
err_bus_start:
    tera_bus_exit(cblk);
err_bus_init:
#endif

    modules_stop(pdev);
err_modules_start:
err_modules_init_phase2:
    tera_gfun_stop(pdev);
err_gfun_start:
    modules_cleanup(pdev);
err_modules_init_phase1:
    tera_gfun_exit(pdev);
err_gfun_init:
    tera_scp_cleanup(pdev);
err_scp_init:
    tera_pkt_cleanup(pdev);
err_pkt_init:
    tera_msg_cleanup(&cblk[dev_id]->msg_cblk);
err_msg_init:
    tera_mem_exit(pdev);
err_mem_init:
    if (dev_id == 0)
    {
        unregister_chrdev_region(pcoip_dev, NUM_DEVICES);
    }
err_alloc_major:
    kfree(cblk[dev_id]);
    mTERA_LOG_ERR("end");
    return err;
}

// --------------------------------------------------------------------------
//  pcoip_remove
///     Device Removal Routine
//
//  Description:
///     pcoip_remove is called by the PCI subsystem to alert the driver
///     that it should release a PCI device.  The could be caused by a
///     Hot-Plug event, or because the driver is going to be removed from
///     memory.
//
//  Parameters:
///     pdev: PCI device information struct
//
// TBD - Fix problem where the the local cursor/keyboard feature is enabled
//       after the driver is unloaded. From what I can tell the host SW attempts
//       close the HDC channel by sending a COMM_CTRL command to the HCC module.
//       The code appears to do this but the ZC does not appear to receive the
//       message.
//
static void __devexit pcoip_remove(
    struct pci_dev *pdev
    )
{
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(pdev);

    mTERA_LOG_DBG("start - dev_id=%d", cblk->dev_id);

    cdev_del(&cblk->cdev);
    gfun_interfaces_found--;
    if(gfun_interfaces_found == 0)
    {
        unregister_chrdev_region(pcoip_dev, NUM_DEVICES);
    }

    // stop all HDC and HCC transfers
    modules_stop(pdev);

    // Free any resources called by the modules
    modules_cleanup(pdev);

    // Clean up the SCP module
    tera_scp_cleanup(pdev);

    // Clean up the message module
    tera_msg_cleanup(&cblk->msg_cblk);

    // Clean up the packet module
    tera_pkt_cleanup(pdev);

    // stop the GFUN interface
    // TBD - including this crashes the system (the function is not called on Windows platforms)
//    tera_gfun_stop(pdev);

    // Free any resources tied with the GFUN module
    tera_gfun_exit(pdev);

    // Close the DMA memory management module
    tera_mem_exit(pdev);

    mTERA_LOG_DBG("end");
}

// --------------------------------------------------------------------------
//  pcoip_open
///     Device Open Routine
//
//  Description:
///     Called when user space application opens the pcoip device. Note
///     the device must be opened in blocking mode.
//
static int pcoip_open(
    struct inode *inode,
    struct file *filp
    )
{
    sTERA_DEV_CBLK *cblk;

    mTERA_LOG_DBG("start");
    if(filp->f_flags & O_NONBLOCK)
    {
        mTERA_LOG_ERR("File must be opened in blocking mode");
        return -EPERM;
    }

#if 0
// TBD - update and include the following code (this is for error checking, so I'm putting it off for now)
    //
    // Verify the size of the message structures exchanged with the FW do not
    // exceed some constants. The C preprocessor does not support the
    // sizeof operator, so this check must be performed at run-time. See the
    // MSG_LEN_CHECK note for additional details.
    //
    if((TERA_MSG_PROT_FEAT_AVAIL_LEN < sizeof(sTERA_MSG_MAP)) ||
       (TERA_MSG_ENV_CMD_AVAIL_LEN < sizeof(sTERA_MSG_MAP)) ||
       (TERA_MSG_ENV_VAR_GET_REQ_LEN < sizeof(sTERA_MSG_ENV_VAR_GET_REQ)) ||
       (TERA_MSG_ENV_VAR_GET_REP_LEN < sizeof(sTERA_MSG_ENV_VAR_GET_REP)) ||
       (TERA_MSG_ENV_VAR_SET_REQ_LEN < sizeof(sTERA_MSG_ENV_VAR_SET_REQ)) ||
       (TERA_MSG_ENV_VAR_SET_REP_LEN < sizeof(sTERA_MSG_ENV_VAR_SET_REP)) ||
       (TERA_MSG_PROT_FEAT_AVAIL_LEN < sizeof(sTERA_MSG_MAP)) ||
       (TERA_MSG_LCUR_CMD_AVAIL_LEN < sizeof(sTERA_MSG_MAP)) ||
       (TERA_MSG_LCUR_EN_REQ_LEN < sizeof(sTERA_MSG_LCUR_EN_REQ)) ||
       (TERA_MSG_LCUR_EN_LEN < sizeof(sTERA_MSG_LCUR_EN)) ||
       (TERA_MSG_LCUR_POS_SET_LEN < sizeof(sTERA_MSG_LCUR_POS_SET)) ||
       (TERA_MSG_LCUR_DATA_LEN < sizeof(sTERA_MSG_LCUR_DATA)) ||
       (TERA_MSG_LCUR_SHAPE_LEN < sizeof(sTERA_MSG_LCUR_SHAPE)) ||
       (TERA_MSG_LCUR_STAT_LEN < sizeof(sTERA_MSG_LCUR_STAT)) ||
       (TERA_MSG_LKBD_CMD_AVAIL_LEN < sizeof(sTERA_MSG_MAP)) ||
       (TERA_MSG_LKBD_EN_REQ_LEN < sizeof(sTERA_MSG_LKBD_EN)) ||
       (TERA_MSG_LKBD_EN_REP_LEN < sizeof(sTERA_MSG_LKBD_EN)) ||
       (TERA_MSG_LKBD_OUT_DATA_LEN < sizeof(sTERA_MSG_LKBD_OUT_DATA)) ||
       (TERA_MSG_LKBD_IN_DATA_LEN < sizeof(sTERA_MSG_LKBD_IN_DATA)))
    {
        mTERA_LOG_CRIT("Message structure size > length expected by firmware");
        return -EPROTO;
    }
#endif

    cblk = container_of(inode->i_cdev, struct sTERA_DEV_CBLK_STRUCT, cdev);
    filp->private_data = cblk;

    mTERA_LOG_DBG("end dev_id=%d", cblk->dev_id);

    return 0;
}

// --------------------------------------------------------------------------
//  pcoip_release
///     Device Release Routine
//
//  Description:
///     called when user space application closes the pcoip device
//
static int pcoip_release(
    struct inode *inode,
    struct file *filp
    )
{
    mTERA_LOG_DBG("start - end");

    return 0;
}

// --------------------------------------------------------------------------
//  pcoip_ioctl
///     Device IOCTL Routine
//
//  Description:
///     This function is called when user space application issues an IOCTL
///     call to the pcoip_host device.
//
#if HAVE_UNLOCKED_IOCTL
static long pcoip_ioctl(
    struct file *filp,
    unsigned int cmd,
    unsigned long arg
    )
#else
static int pcoip_ioctl(
    struct inode *inode,
    struct file *filp,
    unsigned int cmd,
    unsigned long arg
    )
#endif
{
    int err = 0;

    //
    // Verify the HCC and HDC control blocks are setup.
    // Return ENOTTY (inappropriate ioctl) before access_ok()
    //
    if ((hcc_cblk == NULL) || (hdc_cblk == NULL))
    {
        mTERA_LOG_ERR("HCC and HDC control block not setup: cmd=0x%08X, arg-0x%08lX", cmd, arg);
        return -ENOTTY;
    }

    //
    // Extract the type and number bitfields, and don't decode wrong cmds.
    // Return ENOTTY (inappropriate ioctl) before access_ok()
    //
    if ((mTERA_PCOIP_DEV_IOCTL_DEVIF_ID_GET(cmd) != TERA_PCOIP_DEV_HCC_DEVIF) &&
        (mTERA_PCOIP_DEV_IOCTL_DEVIF_ID_GET(cmd) != TERA_PCOIP_DEV_HDC_DEVIF))
    {
        mTERA_LOG_ERR("Invalid IOCTL type: cmd=0x%08X, arg-0x%08lX", cmd, arg);
        return -ENOTTY;
    }

    //
    // The direction is a bitmask, and VERIFY_WRITE catches R/W transfers.
    // `Type' is user-oriented, while access_ok is kernel-oriented, so the
    // concept of "read" and "write" is reversed.
    //
    if (_IOC_DIR(cmd) & _IOC_READ)
    {
        err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
    }
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
    {
        err =  !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
    }
    if (err)
    {
        mTERA_LOG_ERR("Invalid IOCTL dir: cmd=0x%08X, arg-0x%08lX", cmd, arg);
        return -EFAULT;
    }

    if (mTERA_PCOIP_DEV_IOCTL_DEVIF_ID_GET(cmd) == TERA_PCOIP_DEV_HCC_DEVIF)
    {
        // process HCC IOCTL commands
        mTERA_LOG_DBG("calling tera_hcc_ioctl_proc: cmd=0x%08X, arg-0x%08lX", cmd, arg);
        err = tera_hcc_ioctl_proc(hcc_cblk->pdev, cmd, arg);
    }
    else
    {
        // process HDC IOCTL commands
        if (cmd != TERA_PCOIP_HDC_IOCTL_KMP_CUR_POS_IF_NEW)
        {
            mTERA_LOG_DBG("calling tera_hdc_ioctl_proc: cmd=0x%08X, arg-0x%08lX", cmd, arg);
        }
        err = tera_hdc_ioctl_proc(hdc_cblk->pdev, cmd, arg);
    }

    if(err != TERA_SUCCESS)
    {
        return -ENOTTY;
    }
    else
    {
        return 0;
    }
}
