////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Name
///
///     This file defines the top level prototypes and definitions for
///     the TERA1 PCoIP Host driver for Linux.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/input.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include "tera_types.h"
#include "tera_mem.h"
#include "tera_mem_internal.h"
#include "tera_log.h"
#include "tera_errors.h"
#include "tera_msg_internal.h"
#include "tera_pkt_internal.h"
#include "tera_scp_internal.h"
#include "tera_hcc_internal.h"
#include "tera_hdc_internal.h"
#include "tera_gfun_internal.h"

// ---- Constants and Types -------------------------------------------------
// Newer versions of the kernel no longer support the init_MUTEX() macro.
// The following code deals with kernels that do not support the macro.
#ifndef init_MUTEX
#define init_MUTEX(_m) sema_init((_m),1)
#endif

// The following macros are defined in a new version of bitops.h. They are
// defined here if they do not exist.
#ifndef BIT_MASK
#define BIT_MASK(nr)    (1UL << ((nr) % BITS_PER_LONG))
#endif
#ifndef BIT_WORD
#define BIT_WORD(nr)    ((nr) / BITS_PER_LONG)
#endif

// The following constants define various device ID values for the TERA1
// and TERA2 devices on the PCIe bus.
#define TERA_PCI_VENDOR_ID_TERADICI         0x6549
#define TERA_PCI_VENDOR_ID_TERA1            TERA_PCI_VENDOR_ID_TERADICI
#define TERA_PCI_PRODUCT_ID_TERA1           0x1200
#define TERA_PCI_SUB_VENDOR_ID_TERA1        TERA_PCI_VENDOR_ID_TERADICI
#define TERA_PCI_SUB_DEV_ID_TERA1           TERA_PCI_PRODUCT_ID_TERA1
#define TERA_PCI_VENDOR_ID_TERA2            TERA_PCI_VENDOR_ID_TERADICI
#define TERA_PCI_PRODUCT_ID_TERA2           0x2240
#define TERA_PCI_SUB_VENDOR_ID_TERA2        TERA_PCI_VENDOR_ID_TERADICI
#define TERA_PCI_SUB_DEV_ID_TERA2           TERA_PCI_PRODUCT_ID_TERA2

// Number of GFUN interfaces (HCC and HDC)
#define MAX_NUM_GFUN            2

// Constant that determines how often the HCC and HDC threads are scheduled.
// - The Linux Device Drivers book says not all Linux systems support timers
//   with a resolution of 1 ms.
// - Host SW releases prior to 3.0.3 set this constant equal to HZ/1000. This
//   causes problems for the standard SUSE release, which uses a 4 ms clock
//   rather than a 1 ms clock.
// - When this constant equals 10ms the Linux cursor is slower than the
//   local cursor when the round trip latency is < 5 ms.
//
#define IRQ_TIMER_PERIOD_MS        (4)

// Constant that determines how long the ioctls will wait before giving up.
// Currently this is equal to one second.
#define TERA_IOCTL_TIMEOUT_MS      (1000)

// --------------------------------------------------------------------------
//  sTERA_DEV_CBLK
///     Device control block
//
//  Description:
///     This structure defines the fields used to control the PCoIP device
///     on the PCIe bus which interacts with the PCoIP Host SW.
///
typedef struct sTERA_DEV_CBLK_STRUCT {
    INT       initialized;      ///< Initialization flag
    UINT32    mtu_size[GFUN_NUM_DMA_CHAN];  ///< GFUN DMA channel size
    UINT32    dev_id;           ///< PCoIP host interface device ID

    BYTE     *reg_base;         ///< Memory-mapped registers base address
    UINT32    reg_length;       ///< Memory-mapped registers length

    // Ingress descriptor queue and buffers
    sTERA_MEM_DMA  ingress_desc_q[GFUN_NUM_DMA_CHAN];
    sTERA_MEM_BUF *ingress_desc_q_buf[GFUN_NUM_DMA_CHAN][INGRESS_DESC_Q_LEN];

    // Egress descriptor queue and buffers
    sTERA_MEM_DMA  egress_desc_q[GFUN_NUM_DMA_CHAN];
    sTERA_MEM_BUF *egress_desc_q_buf[GFUN_NUM_DMA_CHAN][EGRESS_DESC_Q_LEN];

    // statistics (for debug purposes)
    UINT32    tx_buf_cnt;
    UINT32    rx_buf_cnt;

    // Control blocks
    sMSG_CBLK       msg_cblk;
    sTERA_MEM_CBLK  mem_cblk;
    sPKT_CBLK       pkt_cblk;
    sSCP_CBLK       scp_cblk;
    sHCC_CBLK       hcc_cblk;
    sHDC_CBLK       hdc_cblk;
    sTERA_GFUN_CBLK gfun_cblk;

    /* OS defined structs */
    struct pci_dev   *pdev;
    struct input_dev *lkbd;    ///< Keyboard input device
    struct input_dev *lcur;    ///< Cursor (mouse) input device
    struct input_dev *ljs;     ///< Joystick input device
    struct cdev       cdev;    ///< Character device accessed by user space app
} sTERA_DEV_CBLK;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------
