////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP HCC Device APDUs
///
///     This file defines the enumerations and structures for PCoIP Host
///     Control Channel (HCC) device APDUs. These APDU definitions
///     MUST match the definitions on the firmware.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

// ---- Constants and Types -------------------------------------------------

// Enforce a 4-byte alignment for all structures
#pragma pack(push, 4)

// Maximum miscellaneous string length
#define TERA_PCOIP_HCC_APDU_HMI_MISC_MAX_STR_LEN  (240)

// Maximum number of supported displays across all platforms (Tera1 and Tera2)
#define TERA_PCOIP_HCC_APDU_MAX_NUM_DISP          (4)

///////////////////////////////
// User settings FourCC codes
///////////////////////////////

#define MGMT_IMG_APDU_FOURCC_PREFERENCE_FRAME_RATE_VS_QUALITY   0x66767371  ///< 'fvsq'
#define MGMT_IMG_APDU_FOURCC_BUILD_TO_LOSSLESS_ENABLE           0x6c736265  ///< 'lsbe'

// Total number of user settings we can send to a peer
#define MGMT_IMG_APDU_NUM_USER_CONFIG               2

// Define the size of the largest value that can be sent as part of our
// T/L/V byte stream. Note that this size must be a multiple of 4
// to account for byte padding by the T/L/V library.
#define MGMT_IMG_APDU_USER_CONFIG_TLV_VAL_MAX_SIZE  sizeof( UINT32 )

// Combined size of the type and length fields for a T/L/V
#define MGMT_IMG_APDU_USER_CONFIG_TLV_TL_SIZE       ( 2 * sizeof( UINT32 ) )

// Number of bytes required to send all of our user setting TLV's to
// a peer endpoint. This is equal to the size of our largest TLV
// multiplied by the number of TLV's we send.
#define MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES                                                     \
    (  ( MGMT_IMG_APDU_USER_CONFIG_TLV_TL_SIZE + MGMT_IMG_APDU_USER_CONFIG_TLV_VAL_MAX_SIZE )   \
     * MGMT_IMG_APDU_NUM_USER_CONFIG )

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_USER
///     HCC users
//
//  Description:
///     This enumeration defines the HCC users. Each user has a different set
///     of APDUs.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_USER_HMI = 0,
    TERA_PCOIP_HCC_APDU_USER_KMP = 1,
    TERA_PCOIP_HCC_APDU_USER_IMG = 2,
    TERA_PCOIP_HCC_APDU_USER_MAX
} eTERA_PCOIP_HCC_APDU_USER;

/////////////////////////////////////////////////////////////////////////////
//
// HMI APDUs
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_HMI_CMD
///     HMI commands
//
//  Description:
///     This enumeration defines all HMI commands.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_HMI_CMD_AVAIL_CMD = 0,   ///< Available commands
    TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET   = 1,   ///< Get environment variable
    TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET   = 2,   ///< Set environment variable
    TERA_PCOIP_HCC_APDU_HMI_CMD_MAX
} eTERA_PCOIP_HCC_APDU_HMI_CMD;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_HMI_VAR
///     HMI variables
//
//  Description:
///     This enumeration defines the HMI variables that are available for get/
///     and/or set actions.
///
//  Notes:
///     The SESS_TAG2 enum is used to pass a newer version of the session tag
///     that contains additional security information. The agent uses the
///     existing SESS_TAG (4) set/get commands to write the newer tag with the
///     extra information. The tag field in the get/set commands is 240 bytes
///     long. The actual tag length equals 96 in the first version of the
///     session tag message and 106 in the newer version of the message.
/// 
///     The CACHED_TOPO field indicates the FW supports getting and setting
///     the cached topology. This is a capability provided by the imaging
///     manager, but the imaging manager does not provide an indication of
///     what commands it supports. HMI really is not the correct place to
///     advertise the fact that this capability is supported but it is
///     probably the easiest place to do this.
///
///     The UNICODE field is used by the Host SW to tell the FW the host SW
///     supports processing unicode keyboard events.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR        = 0,  ///< Available variables
    TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO        = 1,  ///< Network link information
    TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG         = 2,  ///< IPv4 configuration
    TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK        = 3,  ///< Session link information
    TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG         = 4,  ///< Session tag
    TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT        = 5,  ///< Connection port
    TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT        = 6,  ///< Session statistics
    TERA_PCOIP_HCC_APDU_HMI_VAR_DEPRECATED       = 7,  ///< Deprecated
    TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG          = 8,  ///< WOL configuration
    TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC        = 9,  ///< Session disconnect
    TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG2        = 10, ///< Session tag version 2
    TERA_PCOIP_HCC_APDU_HMI_VAR_CACHED_TOPO      = 11, ///< Cached topology
    TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE          = 12, ///< Unicode enabled
    TERA_PCOIP_HCC_APDU_HMI_VAR_IMG_USER_EXT_CFG = 13, ///< Imaging User Extended Cfg
    TERA_PCOIP_HCC_APDU_HMI_VAR_MAX
} eTERA_PCOIP_HCC_APDU_HMI_VAR;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_MAP
///     Command/variable map
//
//  Description:
///     This structure defines the format of the available commands/variables
///     map APDU. The map member is a bitmasked mapping of the
///     available commands/variables. To go from a command/variable to its
///     bitmask/array index, use the macros defined below.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_MAP_STRUCT
{
    UINT32 map[8];  ///< Support up to 256 commands/environment variables
} sTERA_PCOIP_HCC_APDU_HMI_MAP;

//
// Macros to get the bitmask and array index for a command/variable for the map
// array in a sTERA_PCOIP_HCC_APDU_HMI_MAP structure
//
// Usage example:
//      To check if command number 56 is available:
//
//      sTERA_PCOIP_HCC_APDU_HMI_MAP avail_cmd;
//
//      if (avail_cmd.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(56)] &
//          mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(56))
//      {
//          // Command number 56 is supported
//      }
//
#define mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(x)    (x / 32)
#define mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(x)  (1 << (x % 32))

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG
///     IPv4 network configuration environment variable structure
//
//  Description:
///     This structure defines the format of a network configration environment
///     variable. This structure is used for both setting and getting blade/
///     client network configurations. This structure is for IPv4 protocol.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG_STRUCT
{
    UINT8 flags;          ///< IPv4 and DHCP enable flags
    UINT8 unused[3];
    char  ip[16];         ///< IP address (ASCII)
    char  mask[16];       ///< Subnet mask (ASCII)
    char  gw[16];         ///< Default gateway (ASCII)
    char  pri_dns[16];    ///< Primary DNS server (ASCII)
    char  sec_dns[16];    ///< Secondary DNS server (ASCII)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG;

#define TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG_FLAGS_IPV4_EN  (0x01)
#define TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG_FLAGS_DHCP_EN  (0x02)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK
///     Session link HMI variable structure
//
//  Description:
///     This structure defines the format of the session link HMI variable.
///
///     When getting the session link, if a session exists, then the connected
///     flag will be set, and the IP and MAC address members will contain the
///     client's IP and MAC address, respectively. If a session does not exist
///     (connected flag is clear), then the IP and MAC address members are
///     invalid. The transition count is the number of session transitions,
///     i.e., connected->disconnected, or vice versa, that have occurred. The
///     count is tracked by the driver, so it is reset only when the driver is
///     unloaded. Applications can optionally detect transition changes by
///     comparing a local copy of the transition count to the one obtained in
///     this structure.
///
///     When setting the session link, the connected flag is used to either
///     connect (set) or disconnect (cleared) the session, and the rest of the
///     fields are ignored and should be zeroed. Currently only disconnecting
///     a session is supported.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_STRUCT
{
    UINT8 flags;             ///< Flags
    UINT8 transition_cnt;    ///< Transition count
    UINT8 unused[2];

    char  ip[16];            ///< Peer IP address (ASCII)
    char  mac[20];           ///< Peer MAC address (ASCII)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK;

#define TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED  (0x01)

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_HMI_LINK_STATE
///     Network link states
//
//  Description:
///     This enumeration defines the states (up, down, unknown) of a network
///     link.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_HMI_LINK_STATE_UNKNOWN = 0,
    TERA_PCOIP_HCC_APDU_HMI_LINK_STATE_UP      = 1,
    TERA_PCOIP_HCC_APDU_HMI_LINK_STATE_DOWN    = 2,
} eTERA_PCOIP_HCC_APDU_HMI_LINK_STATE;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_HMI_LINK_RATE
///     Network link rate
//
//  Description:
///     This enumeration defines the rate (10 Mbps, 100 Mbps, 1000 Mbps, none)
///     of a network link. The rate is unknown when there is no network link.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_NONE  = 0,
    TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_10M   = 1,
    TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_100M  = 2,
    TERA_PCOIP_HCC_APDU_HMI_LINK_RATE_1000M = 3,
} eTERA_PCOIP_HCC_APDU_HMI_LINK_RATE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO
///     Network link info environment variable structure
//
//  Description:
///     This structure defines the format of the network link info environment
///     variable. This variable contains the current link state, duplex mode,
///     and link rate.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO_STRUCT
{
    eTERA_PCOIP_HCC_APDU_HMI_LINK_STATE state;       ///< Link state
    eTERA_PCOIP_HCC_APDU_HMI_LINK_RATE  rate;        ///< Link rate (Mbps)
    UINT8                               flags;       ///< Flags
    UINT8                               unused[3];
    char                                mac[20];     ///< MAC address (ASCII)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO;

#define TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO_FLAGS_FULL_DUPLEX  (0x01)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG
///     Session tag HMI variable structure
//
//  Description:
///     This structure defines the format of the session tag HMI variable. The
///     session tag is used verify an incoming session to the blade. The blade
///     needs to reserve the tag prior to connecting to a peer. The peer must
///     have the same session tag set as the blade, otherwise the session gets
///     dropped. When a session is disconnected, the session tag must be
///     unreserved.
///
///     To reserve a session tag, set the reserve flag bit and fill out the tag
///     member. To unreserve the session tag, clear the reserve flag bit; the
///     tag member does not need to be filled out and should be zeroed out.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG_STRUCT
{
    UINT8 flags;                                        ///< Flags
    UINT8 unused[3];
    char  tag[TERA_PCOIP_HCC_APDU_HMI_MISC_MAX_STR_LEN]; ///< Session tag (ASCII)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG;

#define TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG_FLAGS_RESERVE (0x01)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT
///     Connection port HMI variable structure
//
//  Description:
///     This structure defines the format of the connection port HMI variable.
///     The connection port is the TCP port number to be used when initializing
///     a PCoIP session.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT_STRUCT
{
    UINT16 port;        ///< TCP port number
    UINT16 unused;
} sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT
///     Session statistics HMI variable
//
//  Description:
///     This structure defines the format of the session statistics HMI
///     variable. To get the session statistics, send a variable get APDU
///     with this structure zeroed out. The reply will contain the statistics
///     filled out. To reset the session statistics, send a variable set
///     APDU with this structure zeroed out.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT_STRUCT
{
    UINT64 tx_pkt_cnt;          ///< Tx packet count
    UINT64 rx_pkt_cnt;          ///< Rx packet count
    UINT64 rx_pkt_loss_cnt;     ///< Rx packet loss count

    UINT64 tx_byte_cnt;         ///< Tx byte count
    UINT64 rx_byte_cnt;         ///< Rx byte count

    UINT32 tx_bw_limit;         ///< Active tx bandwidth limit (kbps)
    UINT32 tx_bw_avg;           ///< Average tx bandwidth usage (kbps)
    UINT32 tx_bw_min;           ///< Minimum tx bandwidth usage (kbps)
    UINT32 tx_bw_max;           ///< Maximum tx bandwidth usage (kbps)

    UINT32 rx_bw_avg;           ///< Average rx bandwidth usage (kbps)
    UINT32 rx_bw_min;           ///< Minimum rx bandwidth usage (kbps)
    UINT32 rx_bw_max;           ///< Maximum rx bandwidth usage (kbps)

    UINT32 latency_cur;         ///< Current round trip latency (ms)
    UINT32 latency_min;         ///< Minimum round trip latency (ms)
    UINT32 latency_max;         ///< Maximum round trip latency (ms)

    UINT8  max_num_disp;        ///< Maximum number of supported displays (Tera1=2, Tera2=4)
    UINT8  unused[3];
    UINT32 frame_rate[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];  ///< Frame rates (fps)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG
///     Wake-on-LAN configuration variable
//
//  Description:
///     This structure defines the format of the WOL configuration HMI
///     variable. When used in a variable get APDU, the WOL configuration is
///     filled in by the client. When used in a variable set APDU, the WOL
///     configuration is saved into the client. The IP and MAC addresses are
///     only valid if the enable flag is set, else they should be zeroed out.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG_STRUCT
{
    UINT8 flags;        ///< Flags
    UINT8 unused[3];
    char  mac[20];      ///< MAC address (ASCII)
    char  ip[16];       ///< IP address (ASCII)
} sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG;

#define TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG_FLAGS_EN  (0x01)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC
///     Session disconnect HMI variable structure
//
//  Description:
///     This structure defines the format of the session disconnect HMI
///     variable.
///
///     This variable can only be set by the host SW, it cannot be read. The
///     reason_cause field describes why the session is being disconnected.
///
///     This variable should be set when disconnecting a session if the host
///     FW supports setting this variable. If the host FW does not support
///     setting this variable the upper level application should set the
///     session link variable, which does not include a reason.
///
///     Support for this variable was added in FW release 3.2.x.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC_STRUCT
{
    UINT32 reason_cause;    ///< Disconnect reason
} sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE
///     Unicode HMI variable structure
//
//  Description:
///     This structure defines the format of the unicode HMI variable.
///
///     This variable can only be set by the host SW, it cannot be read. The
///     enabled field describes whether the host SW supports processing unicode
///     key events.
///
///     This variable should be set when a session starts and cleared when the
///     host SW is uninstalled.
///
///     Support for this variable was added in FW release 3.5.0.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE_STRUCT
{
    UINT32 enabled;    ///< unicode enable
} sTERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR
///     Variable get reply/set request APDU
//
//  Description:
///     This structure defines the format of the get reply/set request
///     variable APDU. The data member format is dependent on the HMI variable.
///     Every get reply/set request APDU will be of the same size for all HMI
///     variables.
///
typedef struct sTERA_PCOIP_HCC_APDU_HMI_VAR_STRUCT
{
    eTERA_PCOIP_HCC_APDU_HMI_VAR               var;        ///< Variable

    union
    {
        UINT8                                  raw[248];     ///< Max buffer size
        sTERA_PCOIP_HCC_APDU_HMI_MAP           avail_var;    ///< Available variables
        sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO link_info;    ///< Network link info
        sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG  ipv4_cfg;     ///< IPv4 configuration
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;    ///< Session link info
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG  sess_tag;     ///< Session tag
        sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT conn_port;    ///< Connection port
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT sess_stat;    ///< Session statistics
        sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG   wol_cfg;      ///< WOL configuration
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC sess_disc;    ///< Session Disconnect
        sTERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE   unicode;      ///< Unicode enabled
    } data;
} sTERA_PCOIP_HCC_APDU_HMI_VAR;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HMI_LOCAL_VAR
///     HMI local variables
//
//  Description:
///     This enumeration defines the local HMI variables that are available for
///     get and/or set actions.
///
typedef enum
{
    TERA_PCOIP_HMI_LOCAL_VAR_HOST_SESS_STATE   = 0,    ///< Host session state
    TERA_PCOIP_HMI_LOCAL_VAR_LOCAL_INFO        = 1,    ///< Local information
    TERA_PCOIP_HMI_LOCAL_VAR_MAX
} eTERA_PCOIP_HMI_LOCAL_VAR;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HMI_HOST_SESSION_STATE
///     Host session state
//
//  Description:
///     This enumeration defines the possible values of the host session state.
///
typedef enum
{
    TERA_PCOIP_HMI_HOST_STATE_CONSOLE_CONNECT,
    TERA_PCOIP_HMI_HOST_STATE_CONSOLE_DISCONNECT,
    TERA_PCOIP_HMI_HOST_STATE_REMOTE_CONNECT,
    TERA_PCOIP_HMI_HOST_STATE_REMOTE_DISCONNECT,
    TERA_PCOIP_HMI_HOST_STATE_SESSION_LOGON,
    TERA_PCOIP_HMI_HOST_STATE_SESSION_LOGOFF,
    TERA_PCOIP_HMI_HOST_STATE_SESSION_LOCK,
    TERA_PCOIP_HMI_HOST_STATE_SESSION_UNLOCK,
    TERA_PCOIP_HMI_HOST_STATE_SESSION_REMOTE_CONTROL,
    TERA_PCOIP_HMI_HOST_STATE_INVALID,
} eTERA_PCOIP_HMI_HOST_SESSION_STATE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC
///     Host session state HMI variable structure
//
//  Description:
///     This structure defines the format of the local variable host session
///     state fields.
///
typedef struct sTERA_PCOIP_HMI_LOCAL_VAR_SESS_STATE_STRUCT
{
    eTERA_PCOIP_HMI_HOST_SESSION_STATE host_sess_state;      ///< host session state
} sTERA_PCOIP_HMI_LOCAL_VAR_SESS_STATE;

// --------------------------------------------------------------------------
// sTERA_PCOIP_HMI_LOCAL_VAR_INFO
///     Host software can request local information from the driver.
//
//  Description:
///     This structure defines the format of the local variable info
///     fields.
//
typedef struct sTERA_PCOIP_HMI_LOCAL_VAR_INFO_STRUCT
{
    UINT32                            dev_type;   ///< Host card TERA device type
} sTERA_PCOIP_HMI_LOCAL_VAR_INFO;

#define TERA_PCOIP_HCC_LOCAL_DEV_TYPE_TERA1x00  (0)
#define TERA_PCOIP_HCC_LOCAL_DEV_TYPE_TERA22x0  (1)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HMI_LOCAL_VAR
///     Variable get/set local driver variables
//
//  Description:
///     This structure defines the format of the get/set local variable
///     messages. The data member format is dependent on the HMI variable.
///
typedef struct sTERA_PCOIP_HMI_LOCAL_VAR_STRUCT
{
    eTERA_PCOIP_HMI_LOCAL_VAR                  var;        ///< Variable

    union
    {
        sTERA_PCOIP_HMI_LOCAL_VAR_SESS_STATE   sess_state;   ///< Session state
        sTERA_PCOIP_HMI_LOCAL_VAR_INFO         info;         ///< Driver info
    } data;
} sTERA_PCOIP_HMI_LOCAL_VAR;

/////////////////////////////////////////////////////////////////////////////
//
// KMP APDUs
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_KMP_APDU_TYPE
///     KMP APDU types
//
//  Description:
///     This enumeration defines all KMP APDU types.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_KMP_TYPE_COMM_CTRL = 1,   ///< Communication control
    TERA_PCOIP_HCC_APDU_KMP_TYPE_PRI_CTXT  = 2,
} eTERA_PCOIP_HCC_APDU_KMP_TYPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL
///     Communication control APDU
//
//  Description:
///     This structure defines the format of the communication control APDU.
///     This APDU is used to control the opening and closing of the HDC channel.
///
typedef struct sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL_STRUCT
{
    eTERA_PCOIP_HCC_APDU_KMP_TYPE apdu_type;      ///< APDU type
    UINT32                        open_channel;   ///< Open channel flag
} sTERA_PCOIP_HCC_APDU_KMP_COMM_CTRL;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT
///     Negotiated KMP PRI context APDU
//
//  Description:
///     This structure defines the format of the negotiated KMP PRI context
///     APDU. This APDU lists all the negotiated KMP capabilities for the
///     current PCoIP session.
///
typedef struct sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT_STRUCT
{
    eTERA_PCOIP_HCC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32 pri_num;                             ///< PRI number

    UINT8  kmp_supported;                       ///< KMP supported
    UINT8  kmp_enabled;                         ///< KMP enabled
    UINT8  encodings_supported;                 ///< KMP encodings supported

    // pcoip_kmp_1 encoding
    UINT8  pcoip_kmp_1_supported;               ///< pcoip_kmp_1 supported
    UINT8  cur_shape_caches_supported;          ///< Cursor shape caches supported
    UINT8  cur_shape_cache_size_supported;      ///< Cursor shape cache size supported
    UINT16 cur_shape_caches;                    ///< Number of cursor shape caches
    UINT32 cur_shape_cache_size;                ///< Cursor shape cache size (bytes)
    UINT8  cur_shape_bmp_supported;             ///< Cursor shape bitmaps supported
    UINT8  cur_shape_bmp_alpha;                 ///< Cursor shape alpha bitmaps supported
    UINT8  cur_shape_bmp_color;                 ///< Cursor shape color bitmaps supported
    UINT8  cur_shape_max_size_supported;        ///< kmp.encodings.pcoip_kmp_1.caps.cap_ptr_shape_max_size.supported
    UINT16 cur_shape_max_size_x;                ///< Cursor shape maximum width
    UINT16 cur_shape_max_size_y;                ///< Cursor shape maximum height
    UINT8  kbd_client_auto_repeat_supported;    ///< Keyboard client-side auto-repeat supported
    UINT8  kbd_client_auto_repeat_enable;       ///< Keyboard client-side auto-repeat enabled
    UINT8  cur_shape_fadeout_supported;         ///< Cursor fadeout supported
    UINT8  cur_shape_fadeout_enable;            ///< Cursor fadeout enabled
    UINT8  cur_speed_update_supported;          ///< Cursor speed update supported
    UINT8  cur_speed_update_enable;             ///< Cursor speed update enabled
    UINT8  usb_supported;						///< USB bridging supported
    UINT8  usb_enable;                          ///< USB bridging enabled

    UINT8  reserved[256];                       ///< Reserved for future expansion
} sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT;

/////////////////////////////////////////////////////////////////////////////
//
// IMG APDUs
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_IMG_TYPE
///     IMG APDU types
//
//  Description:
///     This enumeration defines all IMG APDU types used by the Host SW.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_IMG_TYPE_DISP_TOPO               = 0xC,  ///< Display topology
    TERA_PCOIP_HCC_APDU_IMG_TYPE_DISP_TOPO_REQ           = 0xD,  ///< Request for display topology
    TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_GET     = 0x10, ///< Get cached display topology
    TERA_PCOIP_HCC_APDU_IMG_TYPE_CACHE_DISP_TOPO_SET     = 0x11, ///< Set cached display topology
    TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG           = 0x16, ///< User config settings (bidirectional)
    TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_REQ       = 0x17, ///< Request for user config settings
    TERA_PCOIP_HCC_APDU_IMG_TYPE_HOST_USER_CFG_CACHE_SET = 0x18, ///< Set user config settings on client
} eTERA_PCOIP_HCC_APDU_IMG_TYPE;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HCC_APDU_IMG_ROTATION
///     Display rotation values
//
//  Description:
///     This enumeration defines the display rotation values. All degrees of
///     rotation are counter-clockwise, with 0 degrees of rotation being the
///     unrotated display.
///
typedef enum
{
    TERA_PCOIP_HCC_APDU_IMG_ROTATION_0       = 0,
    TERA_PCOIP_HCC_APDU_IMG_ROTATION_90      = 1,
    TERA_PCOIP_HCC_APDU_IMG_ROTATION_180     = 2,
    TERA_PCOIP_HCC_APDU_IMG_ROTATION_270     = 3,
    TERA_PCOIP_HCC_APDU_IMG_ROTATION_UNKNOWN = 0xFFFFFFFF
} eTERA_PCOIP_HCC_APDU_IMG_ROTATION;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_DISP
///     Display configuration
//
//  Description:
///     This structure defines the format of the display configuration. Each
///     structure only represents a single active display.
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_DISP_STRUCT
{
    INT32                             pos_x;       ///< Upper-left x-position
    INT32                             pos_y;       ///< Upper-left y-position
    UINT16                            width;       ///< Post-rotation width
    UINT16                            height;      ///< Post-rotation height
    eTERA_PCOIP_HCC_APDU_IMG_ROTATION rotation;    ///< Display rotation
} sTERA_PCOIP_HCC_APDU_IMG_DISP;

// --------------------------------------------------------------------------
// sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_BUF
///     Display topology buffer
//
//  Decription:
///     Defines the topology of the displays plus the refresh rate.
///
///     Note: The message includes a field that specifies which display
///     is the primary display. The host software ignores this field and
///     instead uses the position data in the topology to determine the
///     primary display. Whichever display has the origin equal to (0,0) is
///     considered the primary display.
///
//
typedef struct sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_BUF_STRUCT
{
    sTERA_PCOIP_HCC_APDU_IMG_DISP topology[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];
    UINT8                         refresh_rate[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];
    UINT32                        primary;
} sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_BUF;

//
// These values are put into the refresh_rate field of the cached topology
// response message over the HCC in the case where caching is not supported
// (this happens when talking to a soft client or a zero client that has
// disabled the display topology settings) or imaging is not yet active
// (this can happen at the start of a session or possibly when the GPU changes
// the output resolution) when the message is created/populated by the blade.
//
// If a client has disabled the topology and imaging is not active in the
// blade the FW will return the NO_CACHING value.
//
#define TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NO_CACHING               0xfe
#define TERA_PCOIP_HCC_HOST_TOPOLOGY_RATE_NOT_ACTIVE               0xff

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_IDEN
///     Display identification
//
//  Description:
///     This structure defines the format of the display identification
///     structure. Each structure contains information for a single display
///     attached to the client.
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_IDEN_STRUCT
{
    UINT8   vid[8];         ///< Vendor ID
    UINT8   pid[8];         ///< Product ID
    UINT8   serial[16];     ///< Serial number
    UINT8   name[16];       ///< Name
} sTERA_PCOIP_HCC_APDU_IMG_IDEN;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO
///     Display topology APDU
//
//  Description:
///     This structure defines the format of the display topology APDU. This
///     APDU may contain up to TERA_PCOIP_HCC_APDU_MAX_NUM_DISP displays. Only
///     active displays are listed in this APDU. The order of which the displays
///     are listed in the topology follows the rule: left -> right, top ->
///     bottom.
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_STRUCT
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE apdu_type;    ///< APDU type
    UINT32                        pri_num;      ///< PRI number
    sTERA_PCOIP_HCC_APDU_IMG_DISP disp[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];
} sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO
///     Cached display topology APDU
//
//  Description:
///     This structure defines the format of the cached display topology APDU.
///     This APDU may contain up to TERA_PCOIP_HCC_APDU_MAX_NUM_DISP displays.
///     Information for all displays attached to the client are listed in this
///     APDU. The displays are ordered by the DVI port the display is connected
///     to, where the first entry is DVI0, 2nd is DVI1... The iden structure
///     contains identification information for the monitor connected the DVI
///     port which is used to match the monitor with the host OS display
///     adapter that is driving the monitor.
///
///     When the client has disabled the topology settings the fields in the
///     topology structure will be filled with zeros, except for the
///     refresh_rate, which will contain some special values described in the
///     comments for the topology structure. The identification structure will
///     always contain EDID information for the displays connected to the
///     client's DVI ports.
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO_STRUCT
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE          apdu_type;    ///< APDU type
    UINT32                                 pri_num;      ///< PRI number
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_BUF topo;
    sTERA_PCOIP_HCC_APDU_IMG_IDEN          iden[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];
} sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO
///     Async Display topology APDU
//
//  Description:
///     This structure contains a copy of the last asynchronous display
///     topology APDU received from the firmware. Starting with FW 3.2.x the
///     client device sends this message at the start of a connection. The host
///     SW .
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO_STRUCT
{
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO apdu;    ///< APDU message
    UINT32                             msg_id;  ///< msg ID (incremented each
                                                ///  time a new msg is rcvd)
} sTERA_PCOIP_HCC_APDU_IMG_ASYNC_DISP_TOPO;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ
///     Display topology request APDU
//
//  Description:
///     This structure defines the format of the display topology request APDU.
///     This APDU is sent from the host to the blade to request the topology
///     that it currently has. This APDU should only be sent while in a session.
///
typedef struct sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ_STRUCT
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE apdu_type;    ///< APDU type
    UINT32                        pri_num;      ///< PRI number
} sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO_REQ;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HCC_IMG_MODE
//      Image engine mode settings.
//
//  Description:
//      This structure includes a subset of the values that determine the mode
//      the image engine operates in for any particular PRI. The settings
//      in this structure are the settings the host SW has access to.
//
typedef struct
{
    UINT32      image_qual_pref;           // frame rate versus quality
    UINT32      build_to_lossless_enable;
} sTERA_PCOIP_HCC_IMG_MODE;

#define TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MIN		(0)    // minimum valid value
#define TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MAX       (100)  // maximum valid value
#define TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_DEFAULT   (50)   // default value
#define TERA_PCOIP_HCC_APDU_BLD_2_LOSSLESS_DEFAULT  (1)    // default value

// --------------------------------------------------------------------------
// sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_PARAMS
///     User configuration notification sent to the host software
//
//  Description:
///     This message can be passed from the host FW to the host SW to
///     notify it of user configuration changes within a session. The
///     host SW can send this message back to the host FW to signal
///     that it wants to apply new settings to the local session.
///     This is the message that gets sent back if the user has
///     a "system tray" application that allows them to modify
///     any user settings within a session.
///
///     Two structures are defined for this message. One for the
///     receive direction (host SW -> host FW) and another for
///     the transmit direction (host FW -> host SW).
///
//
typedef struct
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE     apdu_type;  ///< Imaging APDU type
    UINT32                            pri_num;    ///< PRI number to be used by HCC only
} sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS;

typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS  host_user_cfg;                                              ///< Host user cfg message
    UINT8                                             user_cfg_tlv_bstream[MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES];  ///< Byte stream of user configuration settings stored as T/L/V's
} sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_TX_PARAMS;

// The following structure is exchanged between the host SW GUI and the driver.
// The driver creates outgoing APDUs from it or creates this structure when
// new APDUs are received that must be sent to the host SW GUI.
typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_RX_PARAMS  host_user_cfg_req;   ///< Standard use configuration message
    sTERA_PCOIP_HCC_IMG_MODE                          user_cfg_settings;   ///< unpacked data from APDU
} sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_DATA;

// --------------------------------------------------------------------------
// sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_PARAMS
///     User configuration request sent from the host software
//
//  Description:
///     Note that there is no need for a USER_CFG_CACHE_GET message
///     because the host is always aware of the client's user
///     configuration. If the client's configuration ever changes
///     inside of a session, the host is always notified with an
///     EXTENDED_CFG_UPDATE FCC APDU.
///
//
typedef struct
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE     apdu_type;  ///< Imaging APDU type
    UINT32                            pri_num;    ///< PRI number to be used by HCC only
} sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_RX_PARAMS;

typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_RX_PARAMS  host_user_cfg_req;                                          ///< Standard use configuration message
    UINT8                                                 user_cfg_tlv_bstream[MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES];  ///< Byte stream of user configuration settings stored as T/L/V's
} sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_TX_PARAMS;

// The following structure is exchanged between the host SW GUI and the driver.
// The driver creates outgoing APDUs from it or creates this structure when
// new APDUs are received that must be sent to the host SW GUI.
typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_HOST_USER_CFG_REQ_RX_PARAMS  host_user_cfg_req;   ///< Standard use configuration message
    sTERA_PCOIP_HCC_IMG_MODE                              user_cfg_settings;   ///< unpacked data from APDU
} sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA;

// --------------------------------------------------------------------------
// sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_PARAMS
///     Host software can request to set the user config values stored in
///     the client flash or GPO.
//
//  Description:
///     A host only version is also defined to be used by on the HCC on a
///     hard host. The host merely passes through the set for this
///     message type.
///
///     This is a fire and forget message - there is no response or
///     confirmation that the flash has been programmed.
//
typedef struct
{
    eTERA_PCOIP_HCC_APDU_IMG_TYPE     apdu_type;  ///< Imaging APDU type
    UINT32                            pri_num;    ///< PRI number to be used by HCC only
} sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_RX_PARAMS;

typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_RX_PARAMS  host_user_cfg_req;                                          ///< Standard use configuration message
    UINT8                                                  user_cfg_tlv_bstream[MGMT_IMG_APDU_USER_CONFIG_NUM_BYTES];  ///< Byte stream of user configuration settings stored as T/L/V's
} sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_TX_PARAMS;

// The following structure is exchanged between the host SW GUI and the driver.
// The driver creates outgoing APDUs from it or creates this structure when
// new APDUs are received that must be sent to the host SW GUI.
typedef struct
{
    sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_RX_PARAMS  host_user_cfg_req;   ///< Standard use configuration message
    UINT32                                                 image_qual_pref;     ///< frame rate versus quality
} sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_CACHE_SET_DATA;

#pragma pack(pop) // pack(push, 4)

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

