////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP HDC Device Messages
///
///     This file defines the enumerations and structures for PCoIP Host
///     Data Channel (HDC) device messages. These message definitions MUST
///     match the definitions on the firmware.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

// ---- Constants and Types -------------------------------------------------

// Enforce a 4-byte alignment for all structures
#pragma pack(push, 4)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_USER
///     HDC users
//
//  Description:
///     This enumeration defines the HDC users. Each user has a different set
///     of message commands and structures.
///
typedef enum
{
    TERA_PCOIP_HDC_APDU_USER_KMP = 0,
    TERA_PCOIP_HDC_APDU_USER_MAX
} eTERA_PCOIP_HDC_APDU_USER;

/////////////////////////////////////////////////////////////////////////////
//
// KMP APDUs
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HDC_APDU_KMP_APDU_TYPE
///     KMP APDU types
//
//  Description:
///     This enumeration defines all KMP APDU types.
///
typedef enum
{
    TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_EN       = 0x000A0000,    ///< Feature enable
    TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_EVT       = 0x000A0001,    ///< Local keyboard event
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_EVT       = 0x000A0002,    ///< Local cursor event
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SHAPE_BMP = 0x000A0003,    ///< Local cursor shape bitmap
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SHAPE     = 0x000A0004,    ///< Local cursor shape select
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_POS       = 0x000A0005,    ///< Local cursor position
    TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REPEAT    = 0x000A0006,    ///< Local keyboard repeat rate set
    TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_LOCK      = 0x000A0007,    ///< Local keyboard lock states
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_VIS       = 0x000A0008,    ///< Local cursor visibility
    TERA_PCOIP_HDC_APDU_KMP_TYPE_TOPOLOGY      = 0x000A0009,    ///< Host topology (deprecated)
    TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_STAT_REQ = 0x000A000A,    ///<
    TERA_PCOIP_HDC_APDU_KMP_TYPE_FEAT_STAT     = 0x000A000B,
    TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_LOCALE    = 0x000A000C,
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_IDLE_HIDE = 0x000A000D,
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ   = 0x000A000E,
    TERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD       = 0x000A000F,
    TERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ   = 0x000A0010,    ///< Local keyboard repeat rate query
    TERA_PCOIP_HDC_APDU_KMP_TYPE_MAX
} eTERA_PCOIP_HDC_APDU_KMP_TYPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN
///     KMP features enable APDU
//
//  Description:
///     This structure defines the format of the KMP features enable APDU,
///     which is sent by the host to enable KMP features (local keyboard/
///     cursor).
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        en;           ///< Enable flag
} sTERA_PCOIP_HDC_APDU_KMP_FEAT_EN;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT
///     Local keyboard event APDU
//
//  Description:
///     This structure defines the format of the local keyboard event APDU. This
///     APDU is sent from the client to the host whenever the keyboard on the
///     client has generated a keyboard event, i.e., key press.
///
///     Not all fields are always valid. The following table summarizes when
///     certain fields are valid:
///
///          Flag that is set       |  Fields that are valid
///         ------------------------|------------------------------------------
///          USB scancode           |  Scancode is USB format, else PS/2
///          Modifier keys valid    |  Modifier keys
///          Lock valid             |  Lock status
///          Locale valid           |  Language and Layout
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        tstamp;       ///< Timestamp
    UINT32                        scancode;     ///< Scancode (PS/2 or USB)
    UINT16                        lang;         ///< Language
    UINT16                        layout;       ///< Layout
    UINT8                         state;        ///< LSB Key state {0: up; 1: down}
                                                ///< 2nd LSB: {0: non-unicode; 1: unicode}
    UINT8                         flags;        ///< Flags (bitmapped)
    UINT8                         lock;         ///< Lock status (bitmapped)
    UINT8                         mod;          ///< Modifier keys (bitmapped)
} sTERA_PCOIP_HDC_APDU_KMP_KBD_EVT;

#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_PAGE   (0xFFFF0000)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_SCANCODE_USB_CODE   (0x0000FFFF)

#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_MAKE          (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_STATE_UNICODE       (0x02)

#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_FLAGS_USB_SCANCODE  (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_FLAGS_MOD_VALID     (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_FLAGS_LOCK_VALID    (0x04)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_FLAGS_LOCALE_VALID  (0x08)

#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_LOCK_NUM            (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_LOCK_CAPS           (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_LOCK_SCROLL         (0x04)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_LOCK_KANA           (0x08)

#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_LCTRL           (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_LSHIFT          (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_LALT            (0x04)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_LGUI            (0x08)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_RCTRL           (0x10)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_RSHIFT          (0x20)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_RALT            (0x40)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_EVT_MOD_RGUI            (0x80)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT
///     Local cursor event APDU
//
//  Description:
///     This structure defines the format of the local cursor event APDU. This
///     APDU is sent from the client to the host. The absolute positions are
///     in Windows native coordinates, i.e., 0-65k. If the absolute position
///     bit is set in the flags, then the APDU contains valid absolute
///     positions, otherwise the relative positions are valid.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
    UINT32                        tstamp;    ///< Timestamp
    UINT16                        abs_x;     ///< Absolute X position
    UINT16                        abs_y;     ///< Absolute Y position
    INT16                         rel_x;     ///< Relative X position
    INT16                         rel_y;     ///< Relative Y position
    INT16                         wheel;     ///< Wheel movement
    UINT8                         btn;       ///< Buttons (bitmapped)
    UINT8                         flags;     ///< Flags (bitmapped)
} sTERA_PCOIP_HDC_APDU_KMP_CUR_EVT;

#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_0          (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_1          (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_2          (0x04)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_3          (0x08)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_4          (0x10)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_BTN_5          (0x20)

// The first two masks are used to tell the host which type of data is
// preferred. The host submits the preferred type of data when possible. Below
// is a description of the exceptions to this rule.
//
// 1) The driver has a configuration setting that determines whether absolute
//    or relative data should be submitted. This setting can be changed by the
//    application.
// 2) The application configures the driver algorithm to submit relative
//    reports when the login screen is active OR when there is a topology
//    mismatch.
// 3) The client sets the REL_PREF bit when it wants the driver to submit
//    relative cursor reports.
// 4) The client sets the ABS_PREF bit when it wants the driver to submit
//    absolute cursor reports.
// 5) Relative reports have precedence over absolute reports. If the driver is
//    configured to submit relative reports or the client sets the REL_PREF bit
//    the driver will submit a relative cursor report.
//
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_FLAGS_ABS_PREF  (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_FLAGS_REL_PREF  (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_DATA_FLAGS_BOTH_POS  (0x04)

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE
///     Local cursor shape bitmap types
//
//  Description:
///     This enumeration defines the types of local cursor shape bitmaps.
///
typedef enum
{
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_ALPHA      = 0,  ///< Alpha
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_COLOR      = 1,  ///< Color
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_XOR        = 2,  ///< Xor
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_CMPR_ALPHA = 3,  ///< Compressed-alpha
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_CMPR_COLOR = 4,  ///< Compressed-color
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_CMPR_XOR   = 5,  ///< Compressed-xor
    TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE_MAX
} eTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_TYPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP
///     Local cursor shape bitmap APDU
//
//  Description:
///     This structure defines the format of a local cursor shape bitmap APDU.
///
/// TBD - the Linux user application must be modified to add the apdu_len field
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        tstamp;       ///< Timestamp
    UINT8                         bmp_type;     ///< Bitmap type
    UINT8                         cache;        ///< Caching {0-6: index; 7: enable}
    UINT8                         hot_x;        ///< X hotspot
    UINT8                         hot_y;        ///< Y hotspot
    UINT8                         width;        ///< Bitmap width
    UINT8                         height;       ///< Bitmap height
    UINT8                         and_depth;    ///< AND mask depth
    UINT8                         xor_depth;    ///< XOR mask depth
    UINT32                        apdu_len;     ///< Length of this structure +
                                                ///< length of bitmap
} sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP;

#define TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_CACHE_EN     (0x80)
#define TERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_BMP_CACHE_INDEX  (0x7F)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE
///     Local cursor shape set APDU
//
//  Description:
///     This structure defines the format of a local cursor shape set APDU. This
///     APDU is sent from the host to the client to set the local cursor to be
///     used for the local cursor overlay.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        tstamp;       ///< Timestamp
    UINT8                         index;        ///< Cache index
    UINT8                         unused[3];
} sTERA_PCOIP_HDC_APDU_KMP_CUR_SHAPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW
///     Local cursor position APDU
//
//  Description:
///     This structure defines the format of a local cursor position if new
///     message. This message is sent periodically from the userland
///     application (GUI) to the driver when the GUI detects the host cursor
///     moved.
///
///     Note: the X and Y position data are absolute coordinates that range
///     from 0-FFFF for the entire desktop (the format used by the client)
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW_STRUCT
{
    UINT16 abs_x;                           ///< X position
    UINT16 abs_y;                           ///< Y position
} sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_IF_NEW;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_POS
///     Local cursor position APDU
//
//  Description:
///     This structure defines the format of a local cursor position APDU. This
///     APDU is sent from the host to the client to set the lcaol cursor
///     position. The x-y coordinates are in Windows coordinates.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_POS_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
    UINT32                        tstamp;    ///< Timestamp
    UINT16                        x;         ///< X position
    UINT16                        y;         ///< Y position
} sTERA_PCOIP_HDC_APDU_KMP_CUR_POS;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT
///     Local keyboard repeat delay and rate APDU
//
//  Description:
///     This structure defines the format of a local keyboard repeat delay and
///     rate APDU. This APDU is sent from the host to the client to set the
///     local keyboard's repeat delay and rate.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
    UINT32                        tstamp;    ///< Timestamp
    UINT16                        interval;  ///< Interval between repeats (ms)
    UINT16                        delay;     ///< Repeat delay (ms)
} sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT;

#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE    (65535) // rate/interval value to disable auto-repeat (in ms)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY   (65535) // delay value to disable auto-repeat (in ms)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MIN		(250)   // minimum valid delay value
#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MAX       (1250)  // maximum valid delay value
#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MIN		(30)    // minimum valid rate value
#define TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MAX        (750)   // maximum valid rate value

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK
///     Local keyboard lock keys APDU
//
//  Description:
///     This structure defines the format of a local keyboard lock keys APDU.
///     This APDU is sent from the host to the client to set the keyboard's lock
///     keys states so the keyboard can correctly set the LED indicators.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
    UINT32                        tstamp;    ///< Timestamp
    UINT8                         state;     ///< Locks state {0: off; 1: on}
    UINT8                         unused[3];
} sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCK;

#define TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_NUM_LOCK     (0x01)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_CAPS_LOCK    (0x02)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_SCROLL_LOCK  (0x04)
#define TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_KANA_LOCK    (0x08)
// TBD - this mask isn't defined in the trunk code, so I'm not sure this is correct
#define TERA_PCOIP_HDC_APDU_KMP_KBD_LOCK_STATE_COMPOSE_LOCK (0x10)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS
///     Local cursor visibility APDU
//
//  Description:
///     This structure defines the format of a local cursor visibility APDU.
///     This APDU is sent from the host to the client to set the local cursor's
///     visibility.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
    UINT32                        tstamp;    ///< Timestamp
    UINT8                         visible;   ///< Visibile flag
    UINT8                         unused[3];
} sTERA_PCOIP_HDC_APDU_KMP_CUR_VIS;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_REQ
///     Feature status request APDU
//
//  Description:
///     This structure defines the format of a feature status request APDU. This
///     APDU is sent from the host to the client to request the status of the
///     KMP feature.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_REQ_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type; ///< APDU type
} sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_REQ;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT
///     Feature status APDU
//
//  Description:
///     This structure defines the format of a feature status APDU. This APDU
///     is sent from the client to host to report the status of the KMP feature.
///     The possible error codes are:
///         - Success (no error)
///         - No host monitors (invalid display topology)
///         - Host has more monitors than client can support
///         - Client has no monitors attached
///         - Host has more monitors than client has attached
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;  ///< APDU type
    UINT32                        err;        ///< Error code
} sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT;

#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_SUCCESS           (0x00000000)
#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_NOT_CONFIGURED        (0x0000FF00)
#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_NO_HOST_MON       (0x0000FF01)
#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_DMT_MISMATCH      (0x0000FF02)
#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_DISPLAY_GAP       (0x0000FF03)
#define TERA_PCOIP_HDC_APDU_KMP_FEAT_STAT_ERR_DISPLAY_OVERLAP   (0x0000FF04)

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT
///     Last Feature status APDU
//
//  Description:
///     This structure defines the format of the last feature status msg. The
///     firmware sends the FEATURE STATUS APDU message asynchronously to the
///     driver when the KMP feature status changes. The driver keeps a copy
///     of the last APDU along with a counter that indicates the number of
///     received APDUs. The counter can be used to check if new versions of
///     the message have been received since the last message was retrieved.
///
typedef struct sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT_STRUCT
{
    UINT32                        err;        ///< Error code
    UINT32                        cnt;        ///< Count of APDUs received
} sTERA_PCOIP_HDC_KMP_LAST_FEAT_STAT;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE
///     Local keyboard locale APDU
//
//  Description:
///     This structure defines the format of a local keyboard locale APDU. This
///     APDU is sent from the client to the host to inform the host of the
///     local keyboard's language and layout.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        tstamp;       ///< Timestamp
    UINT16                        lang;         ///< Keyboard language
    UINT16                        layout;       ///< Keyboard layout
} sTERA_PCOIP_HDC_APDU_KMP_KBD_LOCALE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE
///     Local cursor idle hide APDU
//
//  Description:
///     This structure defines the format of a local cursor idle hide APDU. This
///     APDU is sent from the host to the client to set the amount of time
///     the local cursor is idling before the overlay is hidden. A time of
///     zero means to never hide the overlay.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        time;         ///< Time (sec) before hiding idle cursor
} sTERA_PCOIP_HDC_APDU_KMP_CUR_IDLE_HIDE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ
///     Local cursor speed request APDU
//
//  Description:
///     This structure defines the format of a local cursor speed request APDU.
///     This APDU is sent from the host to the client to request the local
///     cursor speed be sent to it.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
} sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_REQ;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD
///     Local cursor speed APDU
//
//  Description:
///     This structure defines the format of a local cursor speed APDU. When
///     this APDU is sent from the host to the client, it is used to set the
///     local cursor speed. When this APDU is sent from the client to the host,
///     it is to report the local cursor speed.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
    UINT32                        spd;          ///< Local cursor speed
} sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD;

#define TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MIN		10
#define TERA_PCOIP_HDC_APDU_KMP_CUR_SPD_MAX     250

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ
///     Local keyboard repeat rate request APDU
//
//  Description:
///     This structure defines the format of a local keyboard repeat rate
///     request APDU. This APDU is sent from the host to the client to request
///     the local keyboard repeat rate APDU be sent to it.
///
typedef struct sTERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ_STRUCT
{
    eTERA_PCOIP_HDC_APDU_KMP_TYPE apdu_type;    ///< APDU type
} sTERA_PCOIP_HDC_APDU_KMP_TYPE_KBD_REP_REQ;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HDC_KMP_EVT_ALG_TYPE
///     KMP event processing algorithm types
//
//  Description:
///     This enumeration defines all KMP event processing algorithm types.
///
typedef enum
{
    TERA_PCOIP_HDC_KMP_TSTAMP_ALG = 0,    ///< Inject events based on the event timestamp
    TERA_PCOIP_HDC_KMP_MIN_DELAY_ALG = 1, ///< Inject events when they are received (minimal delay)
    TERA_PCOIP_HDC_KMP_MAX_ALG = 2,
} eTERA_PCOIP_HDC_KMP_EVT_ALG_TYPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_KMP_EVT_ALG_CTRL
///     KMP event algorithm control
//
//  Description:
///     This structure defines the format of the KMP event processing algorithm
///     control message. This message is sent to the driver to specify which
///     algorithm will be used to process the cursor data received from the
///     client device.
///
typedef struct sTERA_PCOIP_HDC_KMP_EVT_ALG_CTRL_STRUCT
{
    eTERA_PCOIP_HDC_KMP_EVT_ALG_TYPE alg_type;      ///< Algorithm type
} sTERA_PCOIP_HDC_KMP_EVT_ALG_CTRL;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_HDC_LCUR_EVT_ALG_TYPE
///     Local cursor event processing algorithm types
//
//  Description:
///     This enumeration defines all local cursor event processing algorithm
///     types.
///
///     Note: The driver might inject relative cursor reports while it is
///           configured to inject absolute reports. This happens when the
///           REL_PREF bit is set in the flag field of the mouse report APDU
//            received from the client.
///
typedef enum
{
    TERA_PCOIP_HDC_LCUR_ABS_ALG = 0,    ///< Inject cursor events as absolute reports
    TERA_PCOIP_HDC_LCUR_REL_ALG = 1,    ///< Inject cursor events as relative reports
    TERA_PCOIP_HDC_LCUR_MAX_ALG = 2,
} eTERA_PCOIP_HDC_LCUR_EVT_ALG_TYPE;

// --------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL
///     KMP driver config control
//
//  Description:
///     This structure defines the driver KMP configuration settings. This
///     message is sent to the driver to specify the mode of the driver.
///
typedef struct sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL_STRUCT
{
    eTERA_PCOIP_HDC_LCUR_EVT_ALG_TYPE lcur_mode;      ///< Local cursor mode
} sTERA_PCOIP_HDC_KMP_DRIVER_CFG_CTRL;

// --------------------------------------------------------------------------
//  sTERA_HDC_MSG_IOCTL_RSP
///     ioctl response message
//
//  Description:
///     This union includes all of the possible responses to ioctl commands.
///     When an ioctl command is received that requires a response from the
///     firmware the thread executing the ioctl command waits for the response
///     to be received by the dpc_entry from the firmware. The dpc_entry
///     copies the response into the control block and wakes the waiting
///     thread that is servicing the icotl commands. The ioctl thread then
///     copies the data from the control block into the ioctl buffer.
///
typedef struct sTERA_HDC_MSG_IOCTL_RSP_STRUCT
{
    union
    {
        sTERA_PCOIP_HDC_APDU_KMP_FEAT_STAT    feat_stat;  ///< Get reply
        sTERA_PCOIP_HDC_APDU_KMP_TYPE_CUR_SPD cur_speed;  ///< Get reply
        sTERA_PCOIP_HDC_APDU_KMP_KBD_REPEAT   kbd_repeat; ///< Get reply
    } data;
} sTERA_HDC_MSG_IOCTL_RSP;

// ---------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_INFO
///     Information for a display
//
//  Description:
///     This structure stores all of the information required to describe
///     a display
//
typedef struct
{
    INT                               active;     ///< Is there an active display at this index?
    INT32                             min_pix_x;  ///< The starting x pixel for the display
    INT32                             min_pix_y;  ///< The starting y pixel for the display
    INT32                             max_pix_x;  ///< The ending x pixel for the display
    INT32                             max_pix_y;  ///< The ending y pixel for the display
    INT32                             min_abs_x;  ///< The starting absolute x coordinate for the display
    INT32                             min_abs_y;  ///< The starting absolute y coordinate for the display
    INT32                             max_abs_x;  ///< The ending absolute x coordinate for the display
    INT32                             max_abs_y;  ///< The ending absolute y coordinate for the display
    INT32                             width;      ///< Total width
    INT32                             height;     ///< Total height
    eTERA_PCOIP_HCC_APDU_IMG_ROTATION rotation;   ///< Rotation of the display
} sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_INFO;

// ------------------------------------------------------------------------
//  sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE
//      Table used to store information regarding the active topology
//
//  Description:
///     This structure stores all of the information required to describe
///     the active topology
//
typedef struct
{
    INT                                       twinview_active;                             // Flag that indicates whether TwinView is active
    INT                                       num_active_displays;                         // Number of active displays
    INT32                                     min_x;                                       // Leftmost X display coordinate
    INT32                                     max_x;                                       // Rightmost X display coordinate
    INT32                                     min_y;                                       // Top y display coordinate
    INT32                                     max_y;                                       // Bottom y display coordinate
    INT32                                     desktop_width;                               // Total width (desktop - all screens combined)
    INT32                                     desktop_height;                              // Total height (desktop - all screens combined)
    INT32                                     primary_width;                               // Width of display at position (0,0), sometimes called the primary display
    INT32                                     primary_height;                              // Height of display at position (0,0), sometimes called the primary display
    UINT32                                    desktop_x_tolerance;                         // Tolerated difference on the X axis when comparing host and client coordinates
    UINT32                                    desktop_y_tolerance;                         // Tolerated difference on the Y axis when comparing host and client coordinates
    sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_INFO displays[TERA_PCOIP_HCC_APDU_MAX_NUM_DISP];  // Display info for each screen
} sTERA_PCOIP_HDC_LOCAL_CURSOR_DISPLAY_TABLE;

#pragma pack(pop) // pack(push, 4)

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

