////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Simple Communications Protocol Module Internal
///
///     This file defines the internal SCP module function prototypes and
///     definitions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_mem.h"
#include "tera_scp.h"

// ---- Constants and Types -------------------------------------------------

// --------------------------------------------------------------------------
//  sSCP_INSTANCE
//      SCP instance
//
//  Description:
//      This structure defines a SCP instance. Each SCP instance is responsible
//      for a communication path that has one or more channels. All channels
//      share the same MTU size, but need to be individually negotiated via
//      handshaking before it can be used.
//
typedef struct sSCP_INSTANCE_STRUCT
{
    struct sSCP_INSTANCE_STRUCT *next;              ///< Next pointer
    UINT8                        active;            ///< Channel active flags
    UINT8                        ready;             ///< Channel ready flags
    UINT16                       mtu;               ///< MTU size
    fTERA_SCP_CHAN_OPEN_CBACK    open_cback;        ///< Channel open callback
    void                        *open_cback_arg;    ///< Callback user argument
} sSCP_INSTANCE;

// --------------------------------------------------------------------------
//  sSCP_CBLK
//      SCP module control block
//
//  Description:
//      This structure defines the SCP module control block. The control block
//      is per device instance, and not per driver instance. The control block
//      is attached to the context area of each device instance.
//
typedef struct sSCP_CBLK_STRUCT
{
    INT            initialized;    ///< Initialized flag
    sSCP_INSTANCE *instances;      ///< SCP instances
} sSCP_CBLK;

// Accessor function to get the control block from the device context area
//WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(sSCP_CBLK, scp_cblk_get);

//
// Macros to get the pointer to the SCP and SAR packet headers from a packet
// buffer pointer. The SCP packet header is the first thing in the packet,
// followed by the SAR packet header. The data of the packet comes after the
// SAR header.
//
//      -------------------------
//      |   SCP packet header   |
//      |-----------------------|
//      |   SAR packet header   |
//      |-----------------------|
//      |         Data          |
//      |                       |
//
#define mSCP_PKT_HDR_PTR_GET(pkt_buf) \
    ((sSCP_PKT_HDR *)pkt_buf->virt_addr)

#define mSAR_PKT_HDR_PTR_GET(pkt_buf) \
    ((sSAR_PKT_HDR *)((UINT8 *)mSCP_PKT_HDR_PTR_GET(pkt_buf) + \
                     sizeof(sSCP_PKT_HDR)))

#define mDATA_PTR_GET(pkt_buf) \
    (void *)((UINT8 *)mSAR_PKT_HDR_PTR_GET(pkt_buf) + sizeof(sSAR_PKT_HDR))

// --------------------------------------------------------------------------
//  eSCP_PKT_TYPE
//      SCP packet types
//
//  Description:
//      This enumeration defines all the known SCP packet types. The INVITE and
//      OK packets are used for initial SCP channel handshaking only.
//
typedef enum
{
    SCP_PKT_TYPE_DATA   = 0,    ///< DATA
    SCP_PKT_TYPE_INVITE = 1,    ///< INVITE
    SCP_PKT_TYPE_OK     = 2,    ///< OK
    SCP_PKT_TYPE_MAX    = 3
} eSCP_PKT_TYPE;

typedef struct sSCP_PKT_HDR_STRUCT
{
    UINT8 chan_id;
    UINT8 pri_num;
    UINT8 pkt_type;
    UINT8 unused;
} sSCP_PKT_HDR;

// --------------------------------------------------------------------------
//  eAPDU_SEG_LOC
//      APDU segment locations
//
//  Description:
//      This enumeration defines the possible APDU segment location values. If
//      an APDU is not segmented, its segment location should be ONLY_ONE. For
//      segmented APDUs, the first and last APDUs should have locations of
//      FIRST and LAST, respectively; the middle segments should have locations
//      of MID.
//
typedef enum
{
    APDU_SEG_LOC_FIRST    = 1,    ///< First of segmented APDU
    APDU_SEG_LOC_MID      = 2,    ///< Middle segments of segmented APDU
    APDU_SEG_LOC_LAST     = 3,    ///< Last of segmented APDU
    APDU_SEG_LOC_ONLY_ONE = 4,    ///< Only one APDU (no segmentation)
} eAPDU_SEG_LOC;

typedef struct sSAR_PKT_HDR_STRUCT
{
    UINT32 apdu_size;
    UINT16 apdu_seg_size;
    UINT8  apdu_seg_loc;
    UINT8  unused;
} sSAR_PKT_HDR;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

