////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Monitor Class
///
///     This file defines the monitor class.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

// ---- Constants and Types -------------------------------------------------

// bit masks defining errors detected by the push_topology_to_os() function
#define TERA_DISP_HOST_EDID_NOT_FOUND       0x00000001
#define TERA_DISP_EDID_PARSE_ERR            0x00000002
#define TERA_DISP_EDID_MATCH_NOT_FOUND      0x00000004
#define TERA_DISP_ERR_CHANGING_OS_SETTINGS  0x00000008
#define TERA_DISP_ERR_TOO_MANY_SYS_DISP     0x00000010
#define TERA_DISP_ERR_NO_SYS_DISP           0x00000020

// Max string lengths
#define MAX_DEVICE_KEY_LENGTH 128
#define MAX_DEVICE_STR_LENGTH 128

#define TERA_MON_EDID_VID_LEN     4
#define TERA_MON_EDID_PID_LEN     5
#define TERA_MON_EDID_SERIAL_LEN  14
#define TERA_MON_EDID_NAME_LEN    14

// Maximum number of displays this class can detect
#define TERA_DISP_MAX_DISP  (4)

typedef enum
{
    TERA_DISP_ROTATION_0   = 0,    ///< 0 degrees
    TERA_DISP_ROTATION_90  = 1,    ///< 90 degrees
    TERA_DISP_ROTATION_180 = 2,    ///< 180 degrees
    TERA_DISP_ROTATION_270 = 3,    ///< 270 degrees
    TERA_DISP_ROTATION_MAX
} eTERA_DISP_ROTATION;

typedef struct sTERA_DISP_STRUCT
{
    INT32               x_origin;       ///< X-origin
    INT32               y_origin;       ///< Y-origin
    UINT16              width;          ///< Width (pixels)
    UINT16              height;         ///< Height (pixels)
    eTERA_DISP_ROTATION rotation;       ///< Rotation
    UINT8               refresh_rate;   ///< Refresh rate
} sTERA_DISP;

typedef struct sTERA_DISP_TOPO_STRUCT
{
    UINT32     num_disp;
    sTERA_DISP disp[TERA_DISP_MAX_DISP];
} sTERA_DISP_TOPO;

typedef struct sTERA_DISP_EDID_ID_STRUCT
{
    char vid[TERA_MON_EDID_VID_LEN];        ///< Vendor ID
    char pid[TERA_MON_EDID_PID_LEN];        ///< Product ID
    char serial[TERA_MON_EDID_SERIAL_LEN];  ///< Serial number
    char name[TERA_MON_EDID_NAME_LEN];      ///< Name
} sTERA_DISP_EDID_ID;

typedef struct sTERA_DISP_EDID_STRUCT
{
    UINT32              num_disp;
    sTERA_DISP_EDID_ID  edid[TERA_DISP_MAX_DISP];
} sTERA_DISP_EDID;

// --------------------------------------------------------------------------
//  cTERA_DISP
///     Monitor class
//
//  Description:
///     This is the monitor class, which is used for detecting active monitor
///     configurations. This class only supports detecting of physical monitors
///     and ignores any virtual or mirrored display devices.
///
class cTERA_DISP
{

public:

    // --- Public Functions -------------------------------------------------

    cTERA_DISP(void);

    ~cTERA_DISP(void);

    INT32 disp_topology_get(
        sTERA_DISP_TOPO *disp_topo,
        sTERA_DISP_EDID *disp_edid,
        sTERA_DISP_EDID *client_edid);

    INT32 disp_cnt_get(void) const;

    INT32 check_client_topo_active(
        sTERA_DISP_EDID_ID *client_disp_edid,
        sTERA_DISP_TOPO    *client_disp_topo,
        INT                *active,
        UINT32             *num_sys_disp);

    INT32 push_topology_to_os(
        sTERA_DISP_TOPO    *client_disp_topo,
        sTERA_DISP_EDID_ID *client_disp_edid,
        UINT32             *err_mask);

    INT edid_data_unique(
        sTERA_DISP_EDID_ID *client_disp_edid,
        INT32               num_disp);

    INT cached_topo_is_valid(
        sTERA_DISP_TOPO    *client_disp_topo);

    void monitor_rescan(void);

protected:

    // --- Protected Functions ----------------------------------------------

#if 0
    INT32 cTERA_DISP::update_display_setting(
        INT             cfg_mon,
        DWORD           flags,
        DEVMODE        *mode,
        DISPLAY_DEVICE *adapter,
        UINT32         *err_mask,
        LONG           *err);

    INT32 host_topology_get(
        INT                 active_disp,
        DISPLAY_DEVICE     *adapter,
        DISPLAY_DEVICE     *mon,
        DEVMODE            *mode,
        UINT32             *err_mask,
        UINT32             *num_sys_disp);

    void change_settings_debug(
        LONG ret);

    BOOL edid_find(
        WCHAR *mon_device_key,
        BYTE  *edid,
        DWORD  len_edid);
#endif

    void vid_extract(
        char *vid,
        BYTE *edid);

    void pid_extract(
        char *pid,
        BYTE *edid);

    void serial_extract(
        char *serial,
        BYTE *edid);

    void edid_desc_block_parse(
        char  serial[14],
        char  name[14],
        BYTE *edid);

#if 0
    BOOL edid_parse(
        WCHAR *mon_device_key,
        sTERA_DISP_EDID_ID *mon_edid_id);
#endif

    INT edid_match_found(
        sTERA_DISP_EDID_ID  *sys_disp_edid,
        sTERA_DISP_EDID_ID  *client_disp_edid,
        INT32               *client_topo_idx);

#if 0
    void dump_display_mode(
        DEVMODE *mode);

    void dump_display_device(
        DISPLAY_DEVICE *dev);

    INT is_client_mapping_valid(
        INT32   *mapping,
        DEVMODE *mode);

    INT is_monitor_attached(
        DISPLAY_DEVICE  *mon);

    INT is_adapter_active_no_monitor(
        DISPLAY_DEVICE  *mon);
#endif

    // --- Protected Variables ----------------------------------------------

    UINT32 m_disp_cnt;          ///< Display count

}; // cTERA_DISP

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

