////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Logging Module
///
///     This file defines the logging module API prototypes and macros. The
///     macros should be used to give a uniform logging output. All output
///     is sent to the window the agent executable was started from.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include <time.h>
#include <stdio.h>

// ---- Constants and Types -------------------------------------------------

// --------------------------------------------------------------------------
//  eTERA_LOG_LEVEL
///     Logging levels
//
//  Description:
///     This enumeration defines the supported logging levels.
///
typedef enum
{
    TERA_LOG_NONE     = 0,
    TERA_LOG_CRITICAL = 1,
    TERA_LOG_ERROR    = 2,
    TERA_LOG_INFO     = 3,
    TERA_LOG_DEBUG    = 4,
    TERA_LOG_MAX
} eTERA_LOG_LEVEL;

// This macro configures the value of the logging level
#define mSET_LOG_LEVEL(level) (m_log_level = (level))

// This macro determines where the log messages are written. The daemon and
// the agent share this file but they write messages to different
// destinations.
#if DAEMON
#define OUTPUT_LOG_MSG  fprintf(stderr,
#else
#define OUTPUT_LOG_MSG  printf(
#endif

#if __WXDEBUG__

//
// Using printf instead of wxLogFatalError() and wxLogDebug() because the wx
// functions don't work on systems with Xinerama enabled.
//
#define mTERA_LOG_CRIT(x, arg...)                                           \
do {                                                                        \
        if (m_log_level >= TERA_LOG_CRITICAL)                               \
        {                                                                   \
            time_t tera_log_raw_time;                                       \
            struct tm * tera_log_timeinfo;                                  \
            time (&tera_log_raw_time);                                      \
            tera_log_timeinfo = localtime ( &tera_log_raw_time );           \
            OUTPUT_LOG_MSG "%02d/%02d %02d:%02d:%02d [C] %36s | " x "\n",   \
                   tera_log_timeinfo->tm_mon+1, tera_log_timeinfo->tm_mday, \
                   tera_log_timeinfo->tm_hour, tera_log_timeinfo->tm_min,   \
                   tera_log_timeinfo->tm_sec, __FUNCTION__, ##arg);         \
        }                                                                   \
} while (0)

#define mTERA_LOG_ERR(x, arg...)                                            \
do {                                                                        \
        if (m_log_level >= TERA_LOG_ERROR)                                  \
        {                                                                   \
            time_t tera_log_raw_time;                                       \
            struct tm * tera_log_timeinfo;                                  \
            time ( &tera_log_raw_time );                                    \
            tera_log_timeinfo = localtime ( &tera_log_raw_time );           \
            OUTPUT_LOG_MSG "%02d/%02d %02d:%02d:%02d [E] %36s | " x "\n",   \
                   tera_log_timeinfo->tm_mon+1, tera_log_timeinfo->tm_mday, \
                   tera_log_timeinfo->tm_hour, tera_log_timeinfo->tm_min,   \
                   tera_log_timeinfo->tm_sec, __FUNCTION__, ##arg);         \
        }                                                                   \
} while (0)

#define mTERA_LOG_INFO(x, arg...)                                           \
do {                                                                        \
        if (m_log_level >= TERA_LOG_INFO)                                   \
        {                                                                   \
            time_t tera_log_raw_time;                                       \
            struct tm * tera_log_timeinfo;                                  \
            time ( &tera_log_raw_time );                                    \
            tera_log_timeinfo = localtime ( &tera_log_raw_time );           \
            OUTPUT_LOG_MSG "%02d/%02d %02d:%02d:%02d [I] %36s | " x "\n",   \
                   tera_log_timeinfo->tm_mon+1, tera_log_timeinfo->tm_mday, \
                   tera_log_timeinfo->tm_hour, tera_log_timeinfo->tm_min,   \
                   tera_log_timeinfo->tm_sec, __FUNCTION__, ##arg);         \
        }                                                                   \
} while (0)

#define mTERA_LOG_DBG(x, arg...)                                            \
do {                                                                        \
        if (m_log_level >= TERA_LOG_DEBUG)                                  \
        {                                                                   \
            time_t tera_log_raw_time;                                       \
            struct tm * tera_log_timeinfo;                                  \
            time ( &tera_log_raw_time );                                    \
            tera_log_timeinfo = localtime ( &tera_log_raw_time );           \
            OUTPUT_LOG_MSG "%02d/%02d %02d:%02d:%02d [D] %36s | " x "\n",   \
                   tera_log_timeinfo->tm_mon+1, tera_log_timeinfo->tm_mday, \
                   tera_log_timeinfo->tm_hour, tera_log_timeinfo->tm_min,   \
                   tera_log_timeinfo->tm_sec, __FUNCTION__, ##arg);         \
        }                                                                   \
} while (0)

#define mTERA_ASSERT(cond)                                           \
do {                                                                 \
        if (!(cond)) {                                               \
        mTERA_LOG_CRIT("assertion in file %s, function %s, line %d", \
                       __FILE__, __FUNCTION__, __LINE__);            \
        }                                                            \
} while (0)

#else

#define mTERA_LOG_CRIT(x, ...)
#define mTERA_LOG_ERR(x, ...)
#define mTERA_LOG_INFO(x, ...)
#define mTERA_LOG_DBG(x, ...)
#define mTERA_ASSERT(cond)

#endif

// Define the following macro to enable additional logging for the feature
// that send cursor position updates from the host OS to the client
#undef UPDATE_CLIENT_CUR_POS_DBG
//#define UPDATE_CLIENT_CUR_POS_DBG       1

// Define the following macro to enable additional logging for the feature
// that send cursor bitmaps to the client
#undef HOST_CUR_BMP_DBG
//#define HOST_CUR_BMP_DBG                1

// Define the following macro to enable additional logging for the code that
// reads/modifies the host SW configuration data
#undef CFG_DBG
//#define CFG_DBG                         1

// ---- Variable Externs ----------------------------------------------------

// Storage for this variable is defined in tera_pcoip_agent.cpp
extern eTERA_LOG_LEVEL     m_log_level;

// ---- Function Prototypes -------------------------------------------------

