////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Application Class Implementation
///
///     This file implements the PCoIP Host Software application class. This
///     class is the main entry point of this application. Depending on the
///     command line argument, this application can run as either a Windows
///     service, or an user application with a GUI.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include <wx/display.h>
#include <wx/cmdline.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>

#include "tera_errors.h"
#include "tera_log.h"
#include "tera_mem.h"
#include "tera_types.h"
#include "tera_pcoip_agent.h"
#include "tera_pcoip_agent_gui.h"
#include "tera_sinst.h"

// ---- Global Variables ----------------------------------------------------

#if !wxUSE_EXCEPTIONS
    #error "The code must be compiled with wxUSE_EXCEPTIONS == 1"
#endif

// Declare storage for global variable that determines the event logging level
eTERA_LOG_LEVEL     m_log_level;

// ---- Private Constants and Types -----------------------------------------


BEGIN_EVENT_TABLE(cTERA_PCOIP_AGENT, wxApp)
    EVT_END_SESSION(cTERA_PCOIP_AGENT::evt_end_session)
    EVT_QUERY_END_SESSION(cTERA_PCOIP_AGENT::evt_end_session)
END_EVENT_TABLE()

// The following table is used to parse the command line.
static const wxCmdLineEntryDesc g_cmdLineDesc[] =
{
    {
        wxCMD_LINE_SWITCH,
        _T("dde"),
        _T("dis_popup_disc_exit"),
        _T("Disable Host SW UI right-click disconnect & exit"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("disable_lcur"),
        _T("disable_lcur"),
        _T("Disable local cursor & keyboard"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("drcs"),
        _T("dis_remote_cursor_sync"),
        _T("Disable sending remote cursor changes to client"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("dt"),
        _T("disable_twinview"),
        _T("Disable twinview"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("disconnect"),
        _T("disconnect"),
        _T("Disconnect PCoIP session"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_OPTION,
        _T("gc"),
        _T("global_config"),
        _T("Specify name and path of global configuration file (default = /etc/pcoip_agent.conf)"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("h"),
        _T("help"),
        _T("Display help"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_OPTION_HELP
    },
    {
        wxCMD_LINE_SWITCH,
        _T("info"),
        _T("info"),
        _T("Query host card information"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_OPTION,
        _T("lc"),
        _T("local_config"),
        _T("Specify name and path of local configuration file (default = ~/.pcoip_agent)"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
#if __WXDEBUG__
    {
        wxCMD_LINE_OPTION,
        _T("ll"),
        _T("log_level"),
        _T("Specify logging level, supported values: 0=none (default), 1=Critical, 2=Error, 3=Info, 4=Debug"),
        wxCMD_LINE_VAL_NUMBER,
        wxCMD_LINE_PARAM_OPTIONAL
    },
#endif
    {
        wxCMD_LINE_OPTION,
        _T("rcsp"),
        _T("remote_cursor_sync_period"),
        _T("Specify remote cursor update period (in ms, default = 250). The processing load of the host SW increases as this setting is lowered."),
        wxCMD_LINE_VAL_NUMBER,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    {
        wxCMD_LINE_SWITCH,
        _T("t"),
        _T("twinview"),
        _T("Enable twinview"),
        wxCMD_LINE_VAL_STRING,
        wxCMD_LINE_PARAM_OPTIONAL
    },
    // don't delete or add any entries after the following entry, this must
    // appear at the end of the array
    { wxCMD_LINE_NONE,NULL, NULL, NULL, wxCMD_LINE_VAL_NONE, 0 }
};

// ---- Local Variables -----------------------------------------------------

// This constant determines the path and name of the lock file used by the
// host SW to ensure only one instance is running. Normally lock files are
// stored in /var/run or possibly a . file in the users home directory. The
// /var/run directory is only accessible by root and the host SW is run by
// root at the login screen or the logged in user. This means the /var/run
// directory cannot be used. A . file in the users home directory does not
// ensure only one instance of the host SW is running, which is necessary.
// The only common directory that is accessile to all users is /tmp, which
// is why the lock file resides there.
#define HOSTSW_LOCK_FILENAME    "/tmp/pcoiphostsw.pid"

#define MAX_PID_STR_LEN         50

// These constants determine how often the code checks to see if the host
// cursor location differs from the client cursor location. If the locations
// differ and the Remote Cursor Synchronization feature is enabled the host
// SW sends the host cursor position to the client.
//
// Note: Currently the default setting for this constant is 250 ms, which
// minimizes the processing done by the host SW GUI. If this is changed to 20
// ms the CPU usage of the host SW GUI goes from ~0.3% to ~5%. When the agent
// runs more frequently, mouse devices connected directly to the host or
// bridged mouse devices, including tablets, can be used. When this constant
// is 250 ms, mouse devices connected to the host or bridged mouse devices
// work but the client cursor movement is very choppy because it is updated
// only 4 times per second.
#define PCOIP_AGENT_REM_CUR_SYNC_PERIOD_DEFAULT     250
#define PCOIP_AGENT_REM_CUR_SYNC_PERIOD_MINIMUM     5
#define PCOIP_AGENT_REM_CUR_SYNC_PERIOD_MAXIMUM     250

// These constants specify the location of the Host SW configuration files
const wxString PCOIP_AGENT_DEF_GLOBAL_CFG_FILENAME = _T("/etc/pcoiphostswglobal.conf");
#define PCOIP_AGENT_DEF_LOCAL_CFG_FILENAME           _T("%s/.pcoip_agent")

// ---- Local Function Prototypes -------------------------------------------

IMPLEMENT_APP(cTERA_PCOIP_AGENT)

cTERA_PCOIP_AGENT::cTERA_PCOIP_AGENT(void)
{
    m_gui = NULL;
    m_lock_fp = -1;

    // enable processing of fatal exceptions
    wxHandleFatalExceptions(true);

} // cTERA_PCOIP_AGENT::cTERA_PCOIP_AGENT()

cTERA_PCOIP_AGENT::~cTERA_PCOIP_AGENT(void)
{
    // unlock the lock file if it's open
    if (m_lock_fp != -1)
    {
        if (lockf(m_lock_fp, F_ULOCK, 0) == -1)
        {
            mTERA_LOG_ERR("Unable to unlock lockfile %s, code %d (%s)",
                          HOSTSW_LOCK_FILENAME, errno, strerror(errno));
        }
        if (close(m_lock_fp) == -1)
        {
            mTERA_LOG_ERR("Error closing lockfile %s, code %d (%s)",
                          HOSTSW_LOCK_FILENAME, errno, strerror(errno));
        }
    }
    mTERA_LOG_DBG("End");

} // cTERA_PCOIP_AGENT::cTERA_PCOIP_AGENT()

bool cTERA_PCOIP_AGENT::OnInit(void)
{
    INT twinview_mode;
    INT enable_disc_exit = 1;     // enable this feature by default
    INT enable_rem_cur_sync = 1;  // enable this feature by default
    INT done = 1;
    char pid_str[MAX_PID_STR_LEN];
    long rcs_period;
    eTERA_LOG_LEVEL log_level = TERA_LOG_NONE; // disable logging by default
    long new_log_level;
    wxString global_config_filename;
    wxString local_config_filename;

    // initialize the default settings
    mSET_LOG_LEVEL(log_level);
    twinview_mode = is_twinview_enabled();
    rcs_period = PCOIP_AGENT_REM_CUR_SYNC_PERIOD_DEFAULT;
    global_config_filename = PCOIP_AGENT_DEF_GLOBAL_CFG_FILENAME;
    init_def_local_filename(&local_config_filename);

    if (argc > 1)
    {
        // Parse the command line
        wxCmdLineParser cmdParser(g_cmdLineDesc, argc, argv);
        int res;

        // Pass false to supress auto usage message if help was requested
        wxLogNull log;
        res = cmdParser.Parse(true);

#if __WXDEBUG__
        // Override default logging level (only supported by the debug build)
        if (cmdParser.Found(wxT("log_level"), &new_log_level))
        {
            if ((new_log_level >= TERA_LOG_NONE) &&
                (new_log_level <= TERA_LOG_DEBUG))
            {
                log_level = (eTERA_LOG_LEVEL)new_log_level;
                mSET_LOG_LEVEL(log_level);
                mTERA_LOG_DBG("User set logging level = %d",
                              new_log_level);
            }
            else
            {
                mTERA_LOG_ERR("Bad value %d specified for logging level",
                              new_log_level);
                wxMessageBox(_T("Invalid logging level. The value must be >= 0 "
                                "and <= 4. Enter a valid value and try again."),
                             _T("PCoIP Host Software: Error"),
                             wxCANCEL);
                return false;
            }
            done = 0;
        }
#endif

        // disconnect the session and return
        if (cmdParser.Found(wxT("disconnect")))
        {
            mTERA_LOG_DBG("Caller disconnecting session");
            disconnect(TERA_DISCONNECT_CAUSE_HOST_DRIVER_MANUAL_USER_DISCONNECT);
            done = 1;
        }
        // disable the local cursor and keyboard and return
        else if (cmdParser.Found(wxT("disable_lcur")))
        {
            mTERA_LOG_DBG("Caller disabled local cursor");
            disable_lcur();
            done = 1;
        }
        // run the "info" action and return
        else if (cmdParser.Found(wxT("info")))
        {
            mTERA_LOG_DBG("Caller dumping host SW info");
            dump_network_cfg();
            done = 1;
        }

        // Force disable TwinView mode (this overrides the value set by the
        // is_twinview_enabled() function)
        if (cmdParser.Found(wxT("disable_twinview")))
        {
            mTERA_LOG_DBG("Caller disabled Twinview mode");
            twinview_mode = 0;

            // don't exit when done processing the arguments
            done = 0;
        }

        // Force enable TwinView mode (this overrides the value set by the
        // is_twinview_enabled() function)
        if (cmdParser.Found(wxT("twinview")))
        {
            mTERA_LOG_DBG("Caller enabled Twinview mode");
            twinview_mode = 1;

            // don't exit when done processing the arguments
            done = 0;
        }

        // disable disconnect and exit on the right-click popup menu
        if (cmdParser.Found(wxT("dis_popup_disc_exit")))
        {
            mTERA_LOG_DBG("Caller disabled disconnect and exit options on agent right-click menu");
            enable_disc_exit = 0;

            // don't exit when done processing the arguments
            done = 0;
        }

        // Disable sending remote cursor changes to client
        if (cmdParser.Found(wxT("dis_remote_cursor_sync")))
        {
            mTERA_LOG_DBG("Caller disabled sending remote cursor changes to client");
            enable_rem_cur_sync = 0;

            // don't exit when done processing the arguments
            done = 0;
        }

        // Override the default remote cursor synchronization period
        if (cmdParser.Found(wxT("remote_cursor_sync_period"), &rcs_period))
        {
            if ((rcs_period <= PCOIP_AGENT_REM_CUR_SYNC_PERIOD_MAXIMUM) &&
                (rcs_period >= PCOIP_AGENT_REM_CUR_SYNC_PERIOD_MINIMUM))
            {
                mTERA_LOG_DBG("User set remote_cursor_sync_period = %d",
                              rcs_period);
            }
            else
            {
                mTERA_LOG_ERR("Bad value %d specified for remote_cursor_sync_period",
                              rcs_period);
                wxMessageBox(_T("Invalid remote cursor synchronization period. The "
                                "value must be >= 5 and <= 250. Enter a valid value "
                                "and try again."),
                             _T("PCoIP Host Software: Error"),
                             wxCANCEL);
                return false;
            }

            // don't exit when done processing the arguments
            done = 0;
        }

        // Override the default global configuration filename
        if (cmdParser.Found(wxT("global_config"), &global_config_filename))
        {
            if (global_config_filename == wxEmptyString)
            {
                mTERA_LOG_ERR("global config option specified without filename");
                wxMessageBox(_T("Global config option specified without filename. "
                                "Please enter a valid filename and try again."),
                             _T("PCoIP Host Software: Error"),
                             wxCANCEL);
                return false;
            }
            else if(!is_global_config_file_ok(&global_config_filename))
            {
                mTERA_LOG_ERR("global config option specified with invalid filename '%S'",
                              global_config_filename.c_str());
                wxMessageBox(_T("Global config option specified with invalid filename. "
                                "Please enter a valid filename and try again."),
                             _T("PCoIP Host Software: Error"),
                             wxCANCEL);
                return false;
            }

            // don't exit when done processing the arguments
            done = 0;
        }

        // Override the default local configuration filename
        if (cmdParser.Found(wxT("local_config"), &local_config_filename))
        {
            if (local_config_filename == wxEmptyString)
            {
                mTERA_LOG_ERR("local config option specified without filename");
                wxMessageBox(_T("Local config option specified without filename. "
                                "Please enter a valid filename and try again."),
                             _T("PCoIP Host Software: Error"),
                             wxCANCEL);
                return false;
            }

            // don't exit when done processing the arguments
            done = 0;
        }

        if (done)
        {
            return false;
        }
    }

    // verify local config file is ok
    if(!is_local_config_file_ok(&local_config_filename))
    {
        mTERA_LOG_ERR("Error accessing local config file '%S', verify"
                      " process running host SW has write permission",
                      local_config_filename.c_str());
        wxMessageBox(_T("Local config option specified with invalid file. "
                        "Verify the path exists and the process running "
                        "the host SW has write permission to the directory."),
                     _T("PCoIP Host Software: Error"),
                     wxCANCEL);
        return false;
    }

    // exit if the agent was invoked by the root
    wxArrayString output;
    wxString label;
    wxExecute(_T("id -u"), output, wxEXEC_SYNC);
    mTERA_LOG_DBG("id -u returned '%s'", output.Item(0).c_str());
    label.Printf(_T("0"));
    if (output.Item(0).data() == label)
    {
        wxMessageBox(_T("The agent was started by the root user, which is not "
                        "supported. Login as a different user and try again."),
                     _T("PCoIP Host Software: Error"),
                     wxCANCEL);

        return false;
    }

    // Set the application info for registry access
    SetVendorName(_T("Teradici Corporation"));
    SetAppName(_T("PCoIP Host Software"));

    // Suppress log since it pops up a dialog when there is a stale lock file
    wxLogNull *suppress_log = new wxLogNull;

    // Create a lock file. If this fails the host SW exits because another
    // instance is running. The file will remain locked until the host SW
    // exits.
    m_lock_fp = open(HOSTSW_LOCK_FILENAME, O_RDWR | O_CREAT, 0666);
    if (m_lock_fp == -1)
    {
        mTERA_LOG_ERR("Unable to open lockfile %s, code %d (%s)",
                      HOSTSW_LOCK_FILENAME, errno, strerror(errno));
        wxMessageBox(_T("Could not create lock file. Please ensure the directory "
                        "'/tmp' exists and the process running the host SW has "
                        "write access and try again."),
                     _T("PCoIP Host Software: Error"),
                     wxCANCEL);
        return false;
    }
    if (lockf(m_lock_fp, F_TLOCK, 0) == -1)
    {
        mTERA_LOG_ERR("Unable to lock lockfile %s, code %d (%s)",
                      HOSTSW_LOCK_FILENAME, errno, strerror(errno));
        wxMessageBox(_T("Could not start the agent. Another instance of the "
                        "agent is already running or the last instance of the "
                        "agent did not exit properly. Close the other "
                        "instance or delete the file '" HOSTSW_LOCK_FILENAME "' "
                        "and try again."),
                     _T("PCoIP Host Software: Error"),
                     wxCANCEL);
        return false;
    }

    // store the process PID in the lock file
    snprintf(pid_str, sizeof(pid_str), "%d\n", getpid());
    write(m_lock_fp, pid_str, strlen(pid_str));

    if (suppress_log)
    {
        delete suppress_log;
    }

    mTERA_LOG_DBG("Twinview mode = %d", twinview_mode);
    mTERA_LOG_DBG("Global config file = %S", global_config_filename.c_str());
    mTERA_LOG_DBG("Local config file  = %S", local_config_filename.c_str());

    // set the remote cursor syncronization period = to the default value if
    // the feature is disabled
    if ((!enable_rem_cur_sync) && (rcs_period != PCOIP_AGENT_REM_CUR_SYNC_PERIOD_DEFAULT))
    {
        mTERA_LOG_DBG("Resetting remote cursor sync period to default because the feature is disabled");
        rcs_period = PCOIP_AGENT_REM_CUR_SYNC_PERIOD_DEFAULT;
    }

    // Create the top level GUI object
    m_gui = new cPCOIP_AGENT_GUI(NULL, _T("PCoIP Host Software"));

    // initialize startup settings
    m_gui->Show(false);
    m_gui->set_twinview_mode(twinview_mode);
    m_gui->set_disc_exit_mode(enable_disc_exit);
    m_gui->set_remote_cur_sync_settings(enable_rem_cur_sync, rcs_period);
    wxGetApp().gui().host_cfg().init(&global_config_filename,
                                     &local_config_filename);

    // Make the GUI object the top level window of this application
    SetTopWindow(m_gui);

    mTERA_LOG_DBG("end");

    return true;
} // cTERA_PCOIP_AGENT::OnInit()

int cTERA_PCOIP_AGENT::OnExit(void)
{
    mTERA_LOG_DBG("start");

    return wxApp::OnExit();
} // cTERA_PCOIP_AGENT::OnExit()

void cTERA_PCOIP_AGENT::OnFatalException(void)
{
    mTERA_LOG_DBG("start");
    m_gui->end_session();

    wxApp::OnExit();
} // cTERA_PCOIP_AGENT::OnFatalException()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::disconnect
//      Disconnect the current PCoIP session, if any
//
//  Description:
//      This function disconnects the current PCoIP session, if any.
//
void cTERA_PCOIP_AGENT::disconnect(
    eTERA_PCOIP_DISCONNECT_CAUSE reason
    )
{
    cTERA_PCOIP_LINK                pcoip_link;

    // Start the PCoIP link object
    pcoip_link.start();

    // Disconnect the current PCoIP session
	pcoip_link.sess_link_disc(reason);

    // Stop the PCoIP link object
    pcoip_link.stop();
} // cTERA_PCOIP_AGENT::disconnect()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::disable_lcur
//      Disable the local cursor and keyboard
//
//  Description:
//      This function opens a link to the FW and sends a message to disable
//      the local cursor and keyboard.
//
void cTERA_PCOIP_AGENT::disable_lcur(void)
{
    cTERA_PCOIP_LINK    pcoip_link;

    // Start the PCoIP link object
    pcoip_link.start();

    // Blindly disable local cursor and keyboard
    pcoip_link.lcur_lkbd_disable();

    // Stop the PCoIP link object
    pcoip_link.stop();
} // cTERA_PCOIP_AGENT::disable_lcur()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::dump_network_cfg
//      Dump the network config settings
//
//  Description:
//      This function reads the workstation and PCoIP host card network
//      settings and writes the information to the command line.
//
INT32 cTERA_PCOIP_AGENT::dump_network_cfg(void)
{
    INT32                                   ret = TERA_SUCCESS;
    UINT32                                  i;
    cTERA_NETIF                             host_netif;
    cTERA_PCOIP_LINK                        pcoip_link;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG   ip4_cfg;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO  link_info;
    char                                    name[TERA_NETIF_NAME_MAX_LEN];
    char                                    ip[TERA_NETIF_IP_MAX_LEN];
    UINT8                                   mac[6];
    char                                    hostname[100];

    // read the computer hostname
    if (!gethostname(hostname, 100))
    {
        printf("Host Name: %s\n", hostname);
    }
    else
    {
        printf("Unable to read Host Name - gethostname() failed with error: %d\n", errno);
        ret = TERA_ERR_FAILURE;
    }

    // Get the IP and MAC for each network interface on the workstation
    for (i = 0; i < host_netif.if_count(); i++)
    {
        ret = host_netif.if_by_index(i, name, ip, mac);
        if (ret != TERA_SUCCESS)
        {
            printf("Unable to get workstation IP and MAC");
            break;
        }
        else
        {
            printf("Host IP: %s\n", ip);
            printf("Host MAC: %02X-%02X-%02X-%02X-%02X-%02X\n",
                   mac[0], mac[1], mac[2],
                   mac[3], mac[4], mac[5]);
        }
    } // end for

    // Write the host SW build information
    printf("PCoIP Host SW Version: %s\n", HOST_SW_VERSION_STRING);
    printf("PCoIP Host SW Build Date: %s %s\n", __DATE__, __TIME__);

    // read the link network config settings
    ret = pcoip_link.link_info_get(&link_info);

    if (ret == TERA_SUCCESS)
    {
        // read the host card network config settings
        ret = pcoip_link.ipv4_cfg_get(&ip4_cfg);

        if (ret == TERA_SUCCESS)
        {
            printf("PCoIP Host card IP: %s\n", ip4_cfg.ip);
            printf("PCoIP Host card MAC: %s\n", link_info.mac);
        }
    }
    if (ret != TERA_SUCCESS)
    {
        printf("Unable to get PCoIP host card IP and MAC\n");
    }

    return ret;

} // cTERA_PCOIP_AGENT::dump_network_cfg()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::is_twinview_enabled
//      Check if TwinView is enabled
//
//  Description:
//      This function attempts to check if TwinView is enabled. It does this
//      using the nvidia-settings command line utility. The function returns
//      zero if TwinView is not enabled, if the nvidia-settings command is
//      not installed or if the nvidia-settings command does not support
//      querying the value of the TwinView setting. The function returns a
//      non-zero value if nvidia-settings supports querying the state of the
//      TwinView setting AND TwinView is enabled.
//
int cTERA_PCOIP_AGENT::is_twinview_enabled(void)
{
    int twinview_mode = 0;

    // Query the twinview setting
    wxArrayString outStr;
    wxExecute(_T("/usr/bin/nvidia-settings --query TwinView"), outStr, wxEXEC_SYNC);
    for (unsigned i=0; i<outStr.GetCount(); i++)
    {
        // Inspect each line of output to determine whether TwinView is enabled.
        // The code looks for a line containing the words "Attribute" and
        // "'TwinView'". If both words are found the code considers TwinView is
        // enabled if the string ") 1." is present in the line.
        if ((outStr.Item(i).Contains(L"Attribute")) &&
            (outStr.Item(i).Contains(L"'TwinView'")))
        {
            mTERA_LOG_DBG("nvidia-settings supports querying TwinView");
            if (outStr.Item(i).Contains(L"): 1."))
            {
                mTERA_LOG_DBG("TwinView enabled: %S", outStr.Item(i).c_str());
                twinview_mode = 1;
            }
            else
            {
                mTERA_LOG_DBG("TwinView disabled: %S", outStr.Item(i).c_str());
            }

            // Exit the for loop when the code parses the line containing the
            // TwinView setting
            break;
        }
    }

    return twinview_mode;
} // cTERA_PCOIP_AGENT::is_twinview_enabled()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::init_def_local_filename
//      Initialize the local config filename
//
//  Description:
///     This function initializes the local configuration file filename, which
///     is located in the home directory of the user that started the Host SW.
//
void cTERA_PCOIP_AGENT::init_def_local_filename(
    wxString *cfg_filename)
{
    struct passwd *pw;

    cfg_filename->Printf(PCOIP_AGENT_DEF_LOCAL_CFG_FILENAME, wxGetHomeDir().wc_str());

#if CFG_DBG
    mTERA_LOG_DBG("homedir %S", wxGetHomeDir().c_str());
#endif
} // cTERA_PCOIP_AGENT::init_def_local_filename()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::is_global_config_file_ok
//      Check if the specified filename is a valid global config file
//
//  Description:
//      This function returns a non-zero value if the specified filename is a
//      valid global configuration file. The file is considered valid if it
//      exists and it is readable.
//
bool cTERA_PCOIP_AGENT::is_global_config_file_ok(
    const wxString *cfg_filename)
{
    // return 0 if the file is not readable
    if (access(cfg_filename->fn_str(), R_OK) != 0)
    {
        mTERA_LOG_ERR("Unable to read from file %S", cfg_filename->c_str());
        return 0;
    }
    else
    {
#if CFG_DBG
        mTERA_LOG_DBG("Global config file %S is readable", cfg_filename->c_str());
#endif
        return 1;
    }
} // cTERA_PCOIP_AGENT::is_global_config_file_ok()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_AGENT::is_local_config_file_ok
//      Check if the specified filename is a valid local config file
//
//  Description:
//      This function returns a non-zero value if the specified filename is a
//      valid local configuration file. The file is considered valid for a
//      couple reasons:
//
//      1) the file exists and it is writable
//      2) the file is successfully created if it does not exist
//
bool cTERA_PCOIP_AGENT::is_local_config_file_ok(
    const wxString *cfg_filename)
{
    int  cfg_fp;
    bool ret = true;

    // If the file exists
    if (access(cfg_filename->fn_str(), F_OK) == 0)
    {
        // return 0 if the file is not writable
        if (access(cfg_filename->fn_str(), W_OK) != 0)
        {
            mTERA_LOG_ERR("Unable to write the local config file %S", cfg_filename->c_str());
            ret = false;
        }
    }
    else
    {
        // Create the file and set the permissions appropriately
        cfg_fp = open(cfg_filename->fn_str(), O_RDWR | O_CREAT, 0666);
        if (cfg_fp == -1)
        {
            mTERA_LOG_ERR("Unable to open local config file %S, code %d (%s)",
                          cfg_filename->c_str(), errno, strerror(errno));
            ret = false;
        }
        else
        {
            // close the file
            if (close(cfg_fp))
            {
                mTERA_LOG_ERR("Error closing newly created local config file %S, code %d (%s)",
                              cfg_filename->c_str(), errno, strerror(errno));
            }
        }
    }

#if CFG_DBG
    if (ret)
    {
        mTERA_LOG_DBG("Local config file %S is readable", cfg_filename->c_str());
    }
#endif

    return ret;
} // cTERA_PCOIP_AGENT::is_local_config_file_ok()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI::evt_end_session
//      Do what is needed when a session ends
//
//  Description:
//      This function disables the local cursor feature.
//
//void cTERA_PCOIP_AGENT::evt_end_session(void)
void cTERA_PCOIP_AGENT::evt_end_session(
    wxCloseEvent& evt)
{
    cTERA_PCOIP_LINK                        pcoip_link;

    mTERA_LOG_DBG("start");

    // Blindly disable local cursor and keyboard
    pcoip_link.lcur_lkbd_disable();

    // Stop the auto-updating of the link object
    pcoip_link.stop();

    mTERA_LOG_DBG("end");

} // cPCOIP_AGENT_GUI::evt_end_session()
