////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software GUI Class
///
///     This file implements the PCoIP Host Software GUI class. This class is
///     the main GUI class. All other GUI related classes and objects hang off
///     of this class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#ifdef WX_PRECOMP
#include "wx_pch.h"
#endif

#include <wx/menu.h>

#include "tera_log.h"
#include "tera_mem.h"
#include "tera_types.h"
#include "tera_pcoip_agent_internal.h"
#include "tera_pcoip_agent_gui_tray.h"
#include "tera_pcoip_agent_gui.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// System tray menu event IDs
typedef enum
{
    PCOIP_AGENT_TRAY_EVT_ID_EXIT,
    PCOIP_AGENT_TRAY_EVT_ID_PROP,
    PCOIP_AGENT_TRAY_EVT_ID_DISC
} ePCOIP_AGENT_TRAY_EVT_ID;

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_TRAY, wxTaskBarIcon)
    EVT_MENU(PCOIP_AGENT_TRAY_EVT_ID_PROP, cPCOIP_AGENT_GUI_TRAY::evt_menu_prop)
    EVT_MENU(PCOIP_AGENT_TRAY_EVT_ID_EXIT, cPCOIP_AGENT_GUI_TRAY::evt_menu_exit)
    EVT_MENU(PCOIP_AGENT_TRAY_EVT_ID_DISC, cPCOIP_AGENT_GUI_TRAY::evt_menu_disc)
    EVT_TASKBAR_LEFT_DCLICK(cPCOIP_AGENT_GUI_TRAY::evt_left_dclick)
END_EVENT_TABLE()

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

void cPCOIP_AGENT_GUI_TRAY::evt_menu_prop(
    wxCommandEvent& evt
    )
{
    mTERA_LOG_DBG("showing property sheet");

    wxGetApp().gui().prop_sheet_show();
} // cPCOIP_AGENT_GUI_TRAY::evt_menu_prop()

void cPCOIP_AGENT_GUI_TRAY::evt_menu_exit(
    wxCommandEvent& evt
    )
{
    mTERA_LOG_DBG("Host SW exiting...");

    wxGetApp().gui().Close();
} // cPCOIP_AGENT_GUI_TRAY::evt_menu_exit()

void cPCOIP_AGENT_GUI_TRAY::evt_menu_disc(
    wxCommandEvent& evt
    )
{
    mTERA_LOG_DBG("Disconecting session");

    // Disconnect the current PCoIP session
    wxGetApp().gui().pcoip_link().sess_link_disc(TERA_DISCONNECT_CAUSE_HOST_DRIVER_MANUAL_USER_DISCONNECT);

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_TRAY::evt_menu_disc()

void cPCOIP_AGENT_GUI_TRAY::evt_left_dclick(
    wxTaskBarIconEvent& evt
    )
{
    mTERA_LOG_DBG("showing property sheet");

    wxGetApp().gui().prop_sheet_show();
} // cPCOIP_AGENT_GUI_TRAY::evt_left_dclick()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_TRAY::CreatePopupMenu
///     Create tray popup menu
//
//  Description:
///     Called by the library when the user requests popup menu.
///
wxMenu *cPCOIP_AGENT_GUI_TRAY::CreatePopupMenu(void)
{
    wxMenu *menu = new wxMenu;

    menu->Append(PCOIP_AGENT_TRAY_EVT_ID_PROP, _("Properties..."));

    // if the GUI disconnect and exit features are enabled create them
    if (wxGetApp().gui().m_disc_exit_popup_enabled)
    {
        menu->AppendSeparator();
        menu->Append(PCOIP_AGENT_TRAY_EVT_ID_DISC, _("Disconnect PCoIP Session"));

        // If the daemon is active and no one is logged in locally the option
        // is named "Restart Host SW" because the daemon will restart the host
        // SW after it detects it has stopped
        if (is_daemon_active() && !wxGetApp().gui().is_user_logged_in_locally())
        {
            menu->Append(PCOIP_AGENT_TRAY_EVT_ID_EXIT, _("Restart Host SW"));
        }
        else
        {
            menu->Append(PCOIP_AGENT_TRAY_EVT_ID_EXIT, _("Exit"));
        }
    }

    mTERA_LOG_DBG("end");
    return menu;
} // cPCOIP_AGENT_GUI_TRAY::CreatePopupMenu()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_TRAY::is_daemon_active
///     Check if the host SW daemon is running
//
//  Description:
///     This function returns 1 if the host SW daemon is running, 0 if the
///     daemon is not running and -1 if an error occurred and the function
///     could determine if the daemon is running.
///
int cPCOIP_AGENT_GUI_TRAY::is_daemon_active(void)
{
    FILE *fp;
    int ret;
    int rv=-1;

    fp = popen("ps h -C pcoiphostswd", "r");
    if (fp == NULL)
    {
        mTERA_LOG_ERR("popen() failed, code=%d (%s)", errno, strerror(errno));
    }
    else
    {
        rv = 0;
        char line[500];
        while(fgets(line,sizeof line,fp)!= NULL) /* read a line from a file */
        {
            if (strstr(line, "pcoiphostswd") != NULL)
            {
                // exit while loop if an instance of the daemon is running
                mTERA_LOG_DBG("pcoiphostswd active");
                rv = 1;
                break;
            }
        }

        ret = pclose(fp);

        // Log error message if pclose failed. Note: pclose returns -1
        // and sets errno = ECHILD when the child process created by the
        // popen() call above has finished, which is expected. This is why
        // the code below ignores this error.
        if ((ret == -1) && (errno != ECHILD))
        {
            mTERA_LOG_ERR("pclose() failed, code=%d (%s)", errno, strerror(errno));
        }
    }

    if (rv == 0)
    {
        mTERA_LOG_DBG("pcoiphostswd not active");
    }
    else if (rv == -1)
    {
        mTERA_LOG_ERR("Unable to determine if the daemon is active!");
    }
    return rv;
} // cPCOIP_AGENT_GUI_TRAY::is_daemon_active()
